
const char *sffs_guts_c_version="$Id: sffs_guts.c,SFFS 2007/11/16";

#include "yportenv.h"

#include "sffsinterface.h"
#include "sffs_guts.h"

#define SFFS_PASSIVE_GC_CHUNKS 2

#if 0
// Use Steven Hill's ECC struff instead
// External functions for ECC on data
void nand_calculate_ecc (const u_char *dat, u_char *ecc_code);
int nand_correct_data (u_char *dat, u_char *read_ecc, u_char *calc_ecc);
#define sffs_ECCCalculate(data,ecc) nand_calculate_ecc(data,ecc)
#define sffs_ECCCorrect(data,read_ecc,calc_ecc) nand_correct_ecc(data,read_ecc,calc_ecc)
#else
#include "sffs_ecc.h"
#endif

// countBits is a quick way of counting the number of bits in a byte.
// ie. countBits[n] holds the number of 1 bits in a byte with the value n.

static const char sffs_countBitsTable[256] =
{
0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,
1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8
};

static int sffs_CountBits(__u8 x)
{
	int retVal;
	retVal = sffs_countBitsTable[x];
	return retVal;
}



static int sffs_TranactionCreateStart(sffs_Device *dev,int m_nChunk);
static int sffs_WriteWearLeveling(sffs_Device *dev);
static int sffs_ReadWearLeveling(sffs_Device *dev);
static int sffs_IsCrash(sffs_Device *dev);
static int sffs_TransactionStart(sffs_Device *dev,int m_nChunk);
static int sffs_TransactionEnd(sffs_Device *dev);
static int sffs_ChunkCacheCp(sffs_ChunkCache *src, sffs_ChunkCache *dest);
static int sffs_InitFileSystem(sffs_Device *dev);
static int sffs_InitMountReigon(sffs_Device *dev);
static int sffs_ReadLoginfo(sffs_Device *dev);
static int sffs_UnMount(sffs_Device *dev);
static int sffs_WriteMountRegion(sffs_Device *dev);
static int sffs_WriteMetaRegion(sffs_Device *dev);
static int sffs_WriteLogRegion(sffs_Device *dev);
static int sffs_unMountScanFree(sffs_Device *dev);
static int sffs_Scan(sffs_Device *dev);

static void js_spare_dump(sffs_Spare *spare);
static int sffs_CheckObjectHashSanity(sffs_Device *dev);
static void sffs_LoadTagsIntoSpare(sffs_Spare *sparePtr, sffs_Tags *tagsPtr);
static void sffs_GetTagsFromSpare(sffs_Device *dev, sffs_Spare *sparePtr,sffs_Tags *tagsPtr);
static int sffs_PutChunkIntoFile(sffs_Object *in,int chunkInInode, int chunkInNAND, int inScan);

static sffs_Object *sffs_CreateNewObject(sffs_Device *dev,int number,sffs_ObjectType type);
static void sffs_AddObjectToDirectory(sffs_Object *directory, sffs_Object *obj);
static int sffs_UpdateObjectHeader(sffs_Object *in,const char *name, int force);
static void sffs_DeleteChunk(sffs_Device *dev,int chunkId,int markNAND);
static void sffs_RemoveObjectFromDirectory(sffs_Object *obj);
static int sffs_CheckStructures(void);
static int sffs_DeleteWorker(sffs_Object *in, sffs_Tnode *tn, __u32 level, int chunkOffset,int *limit);
static int sffs_DoGenericObjectDeletion(sffs_Object *in);

static sffs_BlockInfo *sffs_GetBlockInfo(sffs_Device *dev,int blockNo);

// Robustification
static void sffs_RetireBlock(sffs_Device *dev,int blockInNAND);
static void sffs_HandleReadDataError(sffs_Device *dev,int chunkInNAND);
static void sffs_HandleWriteChunkError(sffs_Device *dev,int chunkInNAND);
static void sffs_HandleWriteChunkOk(sffs_Device *dev,int chunkInNAND,const __u8 *data, const sffs_Spare *spare);
static void sffs_HandleUpdateChunk(sffs_Device *dev,int chunkInNAND, const sffs_Spare *spare);

static int  sffs_CheckChunkErased(struct sffs_DeviceStruct *dev,int chunkInNAND);

static int sffs_UnlinkWorker(sffs_Object *obj);
static void sffs_AbortHalfCreatedObject(sffs_Object *obj);

static int sffs_VerifyCompare(const __u8 *d0, const __u8 * d1, const sffs_Spare *s0, const sffs_Spare *s1);



loff_t sffs_GetFileSize(sffs_Object *obj);

static int sffs_ReadChunkTagsFromNAND(sffs_Device *dev,int chunkInNAND, sffs_Tags *tags, int *chunkDeleted);
static int sffs_TagsMatch(const sffs_Tags *tags, int objectId, int chunkInObject, int chunkDeleted);


static int sffs_AllocateChunk(sffs_Device *dev,int useReserve);

#ifdef SFFS_PARANOID
static int sffs_CheckFileSanity(sffs_Object *in);
#else
#define sffs_CheckFileSanity(in)
#endif

static void sffs_InvalidateWholeChunkCache(sffs_Object *in);
static void sffs_InvalidateChunkCache(sffs_Object *object, int chunkId);

// Chunk bitmap manipulations

static __inline __u8 *sffs_BlockBits(sffs_Device *dev, int blk)
{
	if(blk < dev->startBlock || blk > dev->endBlock)
	{
		T(SFFS_TRACE_ERROR,(TSTR("**>> sffs: BlockBits block %d is not valid" TENDSTR),blk));
		YBUG();
	}
	return dev->chunkBits + (dev->chunkBitmapStride * (blk - dev->startBlock));
}

static __inline__ void sffs_ClearChunkBits(sffs_Device *dev,int blk)
{
	__u8 *blkBits = sffs_BlockBits(dev,blk);

	 memset(blkBits,0,dev->chunkBitmapStride);
}

static __inline__ void sffs_ClearChunkBit(sffs_Device *dev,int blk,int chunk)
{
	__u8 *blkBits = sffs_BlockBits(dev,blk);
	
	blkBits[chunk/8] &=  ~ (1<<(chunk & 7));
}

static __inline__ void sffs_SetChunkBit(sffs_Device *dev,int blk,int chunk)
{
	__u8 *blkBits = sffs_BlockBits(dev,blk);

	blkBits[chunk/8] |=   (1<<(chunk & 7));
}

static __inline__ int sffs_CheckChunkBit(sffs_Device *dev,int blk,int chunk)
{
	__u8 *blkBits = sffs_BlockBits(dev,blk);
	return (blkBits[chunk/8] &  (1<<(chunk & 7))) ? 1 :0;
}

static __inline__ int sffs_StillSomeChunkBits(sffs_Device *dev,int blk)
{
	__u8 *blkBits = sffs_BlockBits(dev,blk);
	int i;
	for(i = 0; i < dev->chunkBitmapStride; i++)
	{
		if(*blkBits) return 1;
		blkBits++;
	}
	return 0;
}

// Function to manipulate block info


static  __inline__ int sffs_HashFunction(int n)
{
	return (n % SFFS_NOBJECT_BUCKETS);
}


sffs_Object *sffs_Root(sffs_Device *dev)
{
	return dev->rootDir;
}

sffs_Object *sffs_LostNFound(sffs_Device *dev)
{
	return dev->lostNFoundDir;
}

int sffs_InitialiseNAND(struct sffs_DeviceStruct *dev)
{
		T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_InitialiseNAND\n"));        
	return dev->initialiseNAND(dev);
}

static int sffs_WriteNewChunkToNAND(struct sffs_DeviceStruct *dev, const __u8 *data, sffs_Spare *spare,int useReserve)
{
	int chunk;
	
	int writeOk = 1;
	int attempts = 0;
	
	unsigned char rbData[SFFS_BYTES_PER_CHUNK];
	sffs_Spare rbSpare;

	do{
		chunk = sffs_AllocateChunk(dev,useReserve);
       
		if(chunk >= 0)
		{

			// First check this chunk is erased...
#ifndef CONFIG_SFFS_DISABLE_CHUNK_ERASED_CHECK
			writeOk = sffs_CheckChunkErased(dev,chunk);
#endif		
			if(!writeOk)
			{
				T(SFFS_TRACE_ERROR,(TSTR("**>> sffs chunk %d was not erased" TENDSTR),chunk));
			}
			else
			{
				writeOk =  sffs_WriteChunkToNAND(dev,chunk,data,spare);
			}
			attempts++;
			if(writeOk)
			{
				// Readback & verify
				// If verify fails, then delete this chunk and try again
				// To verify we compare everything except the block and 
				// page status bytes.
				// NB We check a raw read without ECC correction applied
				sffs_ReadChunkFromNAND(dev,chunk,rbData,&rbSpare,0);
				
#ifndef CONFIG_SFFS_DISABLE_WRITE_VERIFY
				if(!sffs_VerifyCompare(data,rbData,spare,&rbSpare))
				{
					// Didn't verify
					T(SFFS_TRACE_ERROR,(TSTR("**>> sffs write verify failed on chunk %d" TENDSTR), chunk));

					writeOk = 0;
				}	
#endif				
				
			}
			if(writeOk)
			{
				// Copy the data into the write buffer.
				// NB We do this at the end to prevent duplicates in the case of a write error.
				//Todo
				sffs_HandleWriteChunkOk(dev,chunk,data,spare);
			}
			else
			{
				sffs_HandleWriteChunkError(dev,chunk);
			}
		}
		
	} while(chunk >= 0 && ! writeOk);
	
	if(attempts > 1)
	{
		T(SFFS_TRACE_ERROR,(TSTR("**>> sffs write required %d attempts" TENDSTR),attempts));
		dev->nRetriedWrites+= (attempts - 1);	
	}
	
	return chunk;
}

///
// Functions for robustisizing
//
//

static void sffs_RetireBlock(sffs_Device *dev,int blockInNAND)
{
	// Ding the blockStatus in the first two pages of the block.
	
	sffs_Spare spare;
		T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_RetireBlock\n"));        

	memset(&spare, 0xff,sizeof(sffs_Spare));

	spare.blockStatus = 0;
	
	// TODO change this retirement marking for other NAND types
	sffs_WriteChunkToNAND(dev, blockInNAND * dev->nChunksPerBlock, NULL , &spare);
	sffs_WriteChunkToNAND(dev, blockInNAND * dev->nChunksPerBlock + 1, NULL , &spare);
	
	sffs_GetBlockInfo(dev,blockInNAND)->blockState = SFFS_BLOCK_STATE_DEAD;
	dev->nRetiredBlocks++;
}


static int sffs_RewriteBufferedBlock(sffs_Device *dev)
{
	dev->doingBufferedBlockRewrite = 1;
	//
	//	Remove erased chunks
	//  Rewrite existing chunks to a new block
	//	Set current write block to the new block
	
	dev->doingBufferedBlockRewrite = 0;
	
	return 1;
}


static void sffs_HandleReadDataError(sffs_Device *dev,int chunkInNAND)
{
	int blockInNAND = chunkInNAND/dev->nChunksPerBlock;
//		T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_HandleReadDataError\n"));        
	// Mark the block for retirement
	sffs_GetBlockInfo(dev,blockInNAND)->needsRetiring = 1;
	T(SFFS_TRACE_ERROR | SFFS_TRACE_BAD_BLOCKS,(TSTR("**>>Block %d marked for retirement" TENDSTR),blockInNAND));


	//TODO	
	// Just do a garbage collection on the affected block then retire the block
	// NB recursion
}

static void sffs_CheckWrittenBlock(sffs_Device *dev,int chunkInNAND)
{
}

static void sffs_HandleWriteChunkOk(sffs_Device *dev,int chunkInNAND,const __u8 *data, const sffs_Spare *spare)
{
}

static void sffs_HandleUpdateChunk(sffs_Device *dev,int chunkInNAND, const sffs_Spare *spare)
{
}

static void sffs_HandleWriteChunkError(sffs_Device *dev,int chunkInNAND)
{
	int blockInNAND = chunkInNAND/dev->nChunksPerBlock;

	// Mark the block for retirement
	sffs_GetBlockInfo(dev,blockInNAND)->needsRetiring = 1;
	// Delete the chunk
	sffs_DeleteChunk(dev,chunkInNAND,1);
}




static int sffs_VerifyCompare(const __u8 *d0, const __u8 * d1, const sffs_Spare *s0, const sffs_Spare *s1)
{

		T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_VerifyCompare\n"));        
	if( memcmp(d0,d1,SFFS_BYTES_PER_CHUNK) != 0 ||
		s0->tagByte0 != s1->tagByte0 ||
		s0->tagByte1 != s1->tagByte1 ||
		s0->tagByte2 != s1->tagByte2 ||
		s0->tagByte3 != s1->tagByte3 ||
		s0->tagByte4 != s1->tagByte4 ||
		s0->tagByte5 != s1->tagByte5 ||
		s0->tagByte6 != s1->tagByte6 ||
		s0->tagByte7 != s1->tagByte7 ||
		s0->ecc1[0]  != s1->ecc1[0]  ||
		s0->ecc1[1]  != s1->ecc1[1]  ||
		s0->ecc1[2]  != s1->ecc1[2]  ||
		s0->ecc2[0]  != s1->ecc2[0]  ||
		s0->ecc2[1]  != s1->ecc2[1]  ||
		s0->ecc2[2]  != s1->ecc2[2] )
		{
			return 0;
		}
	
	return 1;
}


///////////////////////// Object management //////////////////
// List of spare objects
// The list is hooked together using the first pointer
// in the object

// static sffs_Object *sffs_freeObjects = NULL;

// static int sffs_nFreeObjects;

// static sffs_ObjectList *sffs_allocatedObjectList = NULL;

// static sffs_ObjectBucket sffs_objectBucket[SFFS_NOBJECT_BUCKETS];


static __u16 sffs_CalcNameSum(const char *name)
{
	__u16 sum = 0;
	__u16 i = 1;
	
	__u8 *bname = (__u8 *)name;
	if(bname)
	{
		while ((*bname) && (i <=SFFS_MAX_NAME_LENGTH))
		{

#ifdef CONFIG_SFFS_CASE_INSENSITIVE
			sum += toupper(*bname) * i;
#else
			sum += (*bname) * i;
#endif
			i++;
			bname++;
		}
	}
	return sum;
}

void sffs_SetObjectName(sffs_Object *obj, const char *name)
{
#ifdef CONFIG_SFFS_SHORT_NAMES_IN_RAM
					if(name && strlen(name) <= SFFS_SHORT_NAME_LENGTH)
					{
						strcpy(obj->shortName,name);
					}
					else
					{
						obj->shortName[0]='\0';
					}
#endif

					obj->sum = sffs_CalcNameSum(name);
}

void sffs_CalcECC(const __u8 *data, sffs_Spare *spare)
{
		T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_CalcECC\n"));            
	sffs_ECCCalculate(data , spare->ecc1);
	sffs_ECCCalculate(&data[256] , spare->ecc2);
}

void sffs_CalcTagsECC(sffs_Tags *tags)
{
	// Calculate an ecc
	
	unsigned char *b = ((sffs_TagsUnion *)tags)->asBytes;
	unsigned  i,j;
	unsigned  ecc = 0;
	unsigned bit = 0;
		T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_CalcTagsECC\n"));            
	tags->ecc = 0;
	
	for(i = 0; i < 8; i++)
	{
		for(j = 1; j &0xff; j<<=1)
		{
			bit++;
			if(b[i] & j)
			{
				ecc ^= bit;
			}
		}
	}
	
	tags->ecc = ecc;
	
	
}

int  sffs_CheckECCOnTags(sffs_Tags *tags)
{
	unsigned ecc = tags->ecc;
	
	sffs_CalcTagsECC(tags);
	
	ecc ^= tags->ecc;
			T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_CheckECCOnTags\n"));            
	if(ecc && ecc <= 64)
	{
		// TODO: Handle the failure better. Retire?
		unsigned char *b = ((sffs_TagsUnion *)tags)->asBytes;

		ecc--;
				
		b[ecc / 8] ^= (1 << (ecc & 7));
		
		// Now recvalc the ecc
		sffs_CalcTagsECC(tags);
		
		return 1; // recovered error
	}
	else if(ecc)
	{
		// Wierd ecc failure value
		// TODO Need to do somethiong here
		return -1; //unrecovered error
	}
	
	return 0;
}


///////////////////////// TNODES ///////////////////////

// List of spare tnodes
// The list is hooked together using the first pointer
// in the tnode.

//static sffs_Tnode *sffs_freeTnodes = NULL;

// static int sffs_nFreeTnodes;

//static sffs_TnodeList *sffs_allocatedTnodeList = NULL;



// sffs_CreateTnodes creates a bunch more tnodes and
// adds them to the tnode free list.
// Don't use this function directly

static int sffs_CreateTnodes(sffs_Device *dev,int nTnodes)
{
    int i;
    sffs_Tnode *newTnodes;
    sffs_TnodeList *tnl;
    
    if(nTnodes < 1) return SFFS_OK;
   
	// make these things
	
    newTnodes = YMALLOC(nTnodes * sizeof(sffs_Tnode));
    
    if (!newTnodes)
    {
		T(SFFS_TRACE_ERROR,(TSTR("sffs: Could not allocate Tnodes"TENDSTR)));
		return SFFS_FAIL;
    }
    
    // Hook them into the free list
    for(i = 0; i < nTnodes - 1; i++)
    {
    	newTnodes[i].internal[0] = &newTnodes[i+1];
#ifdef CONFIG_SFFS_TNODE_LIST_DEBUG
    	newTnodes[i].internal[SFFS_NTNODES_INTERNAL] = 1;
#endif
    }
    	
	newTnodes[nTnodes - 1].internal[0] = dev->freeTnodes;
#ifdef CONFIG_SFFS_TNODE_LIST_DEBUG
   	newTnodes[nTnodes - 1].internal[SFFS_NTNODES_INTERNAL] = 1;
#endif
	dev->freeTnodes = newTnodes;
	dev->nFreeTnodes+= nTnodes;
	dev->nTnodesCreated += nTnodes;

	// Now add this bunch of tnodes to a list for freeing up.
	// NB If we can't add this to the management list it isn't fatal
	// but it just means we can't free this bunch of tnodes later.
	tnl = YMALLOC(sizeof(sffs_TnodeList));
	if(!tnl)
	{
		T(SFFS_TRACE_ERROR,(TSTR("sffs: Could not add tnodes to management list" TENDSTR)));
		
	}
	else
	{
		tnl->tnodes = newTnodes;
		tnl->next = dev->allocatedTnodeList;
		dev->allocatedTnodeList = tnl;
	}


	T(SFFS_TRACE_ALLOCATE,(TSTR("sffs: Tnodes added" TENDSTR)));


	return SFFS_OK;
}


// GetTnode gets us a clean tnode. Tries to make allocate more if we run out
static sffs_Tnode *sffs_GetTnode(sffs_Device *dev)
{
	sffs_Tnode *tn = NULL;

	if(!dev->freeTnodes)
	{
		sffs_CreateTnodes(dev,SFFS_ALLOCATION_NTNODES);
	}
	
	if(dev->freeTnodes)
	{
		tn = dev->freeTnodes;
#ifdef CONFIG_SFFS_TNODE_LIST_DEBUG
    	if(tn->internal[SFFS_NTNODES_INTERNAL] != 1)
		{
			// Hoosterman, this thing looks like it isn't in the list
				T(SFFS_TRACE_ALWAYS,(TSTR("sffs: Tnode list bug 1" TENDSTR)));
		}
#endif
		dev->freeTnodes = dev->freeTnodes->internal[0];
		dev->nFreeTnodes--;
		// zero out
		memset(tn,0,sizeof(sffs_Tnode));
	}
	

	return tn;
}


// FreeTnode frees up a tnode and puts it back on the free list
static void sffs_FreeTnode(sffs_Device*dev, sffs_Tnode *tn)
{
	if(tn)
	{
#ifdef CONFIG_SFFS_TNODE_LIST_DEBUG
    	if(tn->internal[SFFS_NTNODES_INTERNAL] != 0)
		{
			// Hoosterman, this thing looks like it is already in the list
				T(SFFS_TRACE_ALWAYS,(TSTR("sffs: Tnode list bug 2" TENDSTR)));
		}
		tn->internal[SFFS_NTNODES_INTERNAL] = 1;
#endif
		tn->internal[0] = dev->freeTnodes;
		dev->freeTnodes = tn;
		dev->nFreeTnodes++;
	}
}


static void sffs_DeinitialiseTnodes(sffs_Device*dev)
{
	// Free the list of allocated tnodes
	while(dev->allocatedTnodeList)
	{
		YFREE(dev->allocatedTnodeList->tnodes);
		dev->allocatedTnodeList =  dev->allocatedTnodeList->next;
	}
	
	dev->freeTnodes = NULL;
	dev->nFreeTnodes = 0;
}

static void sffs_InitialiseTnodes(sffs_Device*dev)
{

	dev->allocatedTnodeList = NULL;
	dev->freeTnodes = NULL;
	dev->nFreeTnodes = 0;
	dev->nTnodesCreated = 0;

}

#if 0
void sffs_TnodeTest(sffs_Device *dev)
{

	int i;
	int j;
	sffs_Tnode *tn[1000];
	
	YINFO("Testing TNodes");
	
	for(j = 0; j < 50; j++)
	{
		for(i = 0; i < 1000; i++)
		{
			tn[i] = sffs_GetTnode(dev);
			if(!tn[i])
			{
				YALERT("Getting tnode failed");
			}
		}
		for(i = 0; i < 1000; i+=3)
		{
			sffs_FreeTnode(dev,tn[i]);
			tn[i] = NULL;
		}
		
	}
}
#endif


////////////////// END OF TNODE MANIPULATION ///////////////////////////

/////////////// Functions to manipulate the look-up tree (made up of tnodes)
// The look up tree is represented by the top tnode and the number of topLevel
// in the tree. 0 means only the level 0 tnode is in the tree.


// FindLevel0Tnode finds the level 0 tnode, if one exists.
// Used when reading.....
static sffs_Tnode *sffs_FindLevel0Tnode(sffs_Device *dev,sffs_FileStructure *fStruct, __u32 chunkId)
{
	
	sffs_Tnode *tn = fStruct->top;
	__u32 i;
	int requiredTallness;	
	int level = fStruct->topLevel;
    
	// Check sane level and chunk Id
	if(level < 0 || level > SFFS_TNODES_MAX_LEVEL)
	{
		return NULL;
	}
	
	if(chunkId > SFFS_MAX_CHUNK_ID)
	{
		return NULL;
	}

	// First check we're tall enough (ie enough topLevel)
	
	i = chunkId >> (/*dev->chunkGroupBits  + */SFFS_TNODES_LEVEL0_BITS);
	requiredTallness = 0;
	while(i)
	{
		i >>= SFFS_TNODES_INTERNAL_BITS;
		requiredTallness++;
	}
	
	
	if(requiredTallness > fStruct->topLevel)
	{
		// Not tall enough, so we can't find it, return NULL.
		return NULL;
	}
		
	
	// Traverse down to level 0
	while (level > 0 && tn)
	{
	    tn = tn->internal[(chunkId >>(/* dev->chunkGroupBits + */ SFFS_TNODES_LEVEL0_BITS + (level-1) * SFFS_TNODES_INTERNAL_BITS)) & 
	                       SFFS_TNODES_INTERNAL_MASK]; 
		level--;
	
	}


	return tn;		
}

// AddOrFindLevel0Tnode finds the level 0 tnode if it exists, otherwise first expands the tree.
// This happens in two steps:
//  1. If the tree isn't tall enough, then make it taller.
//  2. Scan down the tree towards the level 0 tnode adding tnodes if required.
//
// Used when modifying the tree.
//
static sffs_Tnode *sffs_AddOrFindLevel0Tnode(sffs_Device *dev, sffs_FileStructure *fStruct, __u32 chunkId)
{
	
	sffs_Tnode *tn; 
	
	int requiredTallness;
	int i;
	int l;
	
	__u32 x;
		
	
	// Check sane level and page Id
	if(fStruct->topLevel < 0 || fStruct->topLevel > SFFS_TNODES_MAX_LEVEL)
	{
		return NULL;
	}
	
	if(chunkId > SFFS_MAX_CHUNK_ID)
	{
		return NULL;
	}
	
	// First check we're tall enough (ie enough topLevel)
	
	x = chunkId >> (/*dev->chunkGroupBits + */SFFS_TNODES_LEVEL0_BITS);
	requiredTallness = 0;
	while(x)
	{
		x >>= SFFS_TNODES_INTERNAL_BITS;
		requiredTallness++;
	}
	
	
	
	if(requiredTallness > fStruct->topLevel)
	{
		for(i = fStruct->topLevel; i < requiredTallness; i++)
		{
			
			tn = sffs_GetTnode(dev);
			
			if(tn)
			{
				tn->internal[0] = fStruct->top;
				fStruct->top = tn;
			}
			else
			{
					T(SFFS_TRACE_ERROR,(TSTR("sffs: no more tnodes" TENDSTR)));
			}
		}
		
		fStruct->topLevel = requiredTallness;
	}
	


	// Traverse down to level 0, adding anything we need
	
	l = fStruct->topLevel;
	tn = fStruct->top;


    
	while (l > 0 && tn)
	{

		x = (chunkId >> (/*dev->chunkGroupBits + */SFFS_TNODES_LEVEL0_BITS + (l-1) * SFFS_TNODES_INTERNAL_BITS)) & 
	                       SFFS_TNODES_INTERNAL_MASK;
		
	    if(!tn->internal[x])
	    {
	    	tn->internal[x] = sffs_GetTnode(dev);

	    }
	    
	    tn = 	tn->internal[x];
		l--;
	
	}
	
	return tn;		
}

// DeleteWorker scans backwards through the tnode tree and deletes all the
// chunks and tnodes in the file
// Returns 1 if the tree was deleted. Returns 0 if it stopped early due to hitting the limit and the delete is incomplete.

static int sffs_DeleteWorker(sffs_Object *in, sffs_Tnode *tn, __u32 level, int chunkOffset,int *limit)
{
	int i;
	int chunkInInode;
	int theChunk;
	sffs_Tags tags;
	int found;
	int chunkDeleted;
	int allDone = 1;
	
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_DeleteWorker\n"));            
	if(tn)
	{
		if(level > 0)
		{
		
			for(i = SFFS_NTNODES_INTERNAL -1; allDone && i >= 0; i--)
			{
			    if(tn->internal[i])
		    	{
					if(limit && (*limit) < 0)
					{
						allDone = 0;
					}
					else
					{
						allDone = sffs_DeleteWorker(in,tn->internal[i],level - 1,
										(chunkOffset << SFFS_TNODES_INTERNAL_BITS ) + i ,limit);
					}
					if(allDone)
					{
						sffs_FreeTnode(in->myDev,tn->internal[i]);
			    		tn->internal[i] = NULL;
					}
			    }
		    
			}
			return (allDone) ? 1 : 0;
		}
		else if(level == 0)
		{
			int hitLimit = 0;
			
			for(i = SFFS_NTNODES_LEVEL0 -1; i >= 0 && !hitLimit; i--)
			{
			    if(tn->level0[i])
		    	{
					int j;
					
					chunkInInode = (chunkOffset << SFFS_TNODES_LEVEL0_BITS ) + i;
					
					theChunk =  tn->level0[i] << in->myDev->chunkGroupBits;

					// Now we need to search for it
					for(j = 0,found = 0; theChunk && j < in->myDev->chunkGroupSize && !found; j++)
					{
						sffs_ReadChunkTagsFromNAND(in->myDev,theChunk,&tags,&chunkDeleted);
						if(sffs_TagsMatch(&tags,in->objectId,chunkInInode,chunkDeleted))
						{
							// found it;
							found = 1;
					
						}
						else
						{
							theChunk++;
						}
					}
					
					if(found)
					{
						sffs_DeleteChunk(in->myDev,theChunk,1);
						in->nDataChunks--;
						if(limit)
						{ 
							*limit = *limit-1;
							if(limit <= 0) 
							{ 
								hitLimit = 1;
							}
						}
					
					}
					
			    	tn->level0[i] = 0;
			    }
		    
			}
			return (i < 0) ? 1 : 0;

			
		}
		
	}
	
	return 1;
	
}

// SoftDeleteWorker scans backwards through the tnode tree and soft deletes all the chunks in the file.
// All soft deleting does is increment the block's softdelete count and pulls the chunk out
// of the tnode.
// THus, essentially this is the same as DeleteWorker except that the chunks are soft deleted.
//
static int sffs_SoftDeleteWorker(sffs_Object *in, sffs_Tnode *tn, __u32 level, int chunkOffset)
{
	int i;
	int chunkInInode;
	int theChunk;
	sffs_BlockInfo *theBlock;
//	sffs_Tags tags;
//	int found;
//	int chunkDeleted;
	int allDone = 1;
	
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_SoftDeleteWorker\n"));            	
	if(tn)
	{
		if(level > 0)
		{
		
			for(i = SFFS_NTNODES_INTERNAL -1; allDone && i >= 0; i--)
			{
			    if(tn->internal[i])
		    	{
						allDone = sffs_SoftDeleteWorker(in,tn->internal[i],level - 1,
										(chunkOffset << SFFS_TNODES_INTERNAL_BITS ) + i);
					if(allDone)
					{
						sffs_FreeTnode(in->myDev,tn->internal[i]);
			    		tn->internal[i] = NULL;
					}
					else
					{
						//Hoosterman... how could this happen.
					}			    
				}		    
			}
			return (allDone) ? 1 : 0;
		}
		else if(level == 0)
		{
			
			for(i = SFFS_NTNODES_LEVEL0 -1; i >=0; i--)
			{
			    if(tn->level0[i])
		    	{
					theChunk =  (tn->level0[i] << in->myDev->chunkGroupBits);
					T(SFFS_TRACE_SCAN,(TSTR("soft delete tch %d cgb %d chunk %d" TENDSTR),
						tn->level0[i],in->myDev->chunkGroupBits,theChunk));
						
					theBlock =	sffs_GetBlockInfo(in->myDev,  theChunk/in->myDev->nChunksPerBlock);
					if(theBlock)
					{
						theBlock->softDeletions++;
					}
			    	tn->level0[i] = 0;
			    }
		    
			}
			return 1;
			
		}
		
	}
	
	return 1;
		
}



static void sffs_SoftDeleteFile(sffs_Object *obj)
{
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_SoftDeleteFile\n"));            	    
	if(obj->deleted &&
	   obj->variantType == SFFS_OBJECT_TYPE_FILE &&
	   !obj->softDeleted)
	{
		if(obj->nDataChunks <= 0)
		{
				// Empty file, just delete it immediately
				sffs_FreeTnode(obj->myDev,obj->variant.fileVariant.top);
				obj->variant.fileVariant.top = NULL;
				T(SFFS_TRACE_TRACING,(TSTR("sffs: Deleting empty file %d" TENDSTR),obj->objectId));
				sffs_DoGenericObjectDeletion(obj);	
		}
		else
		{
			sffs_SoftDeleteWorker(obj, obj->variant.fileVariant.top, obj->variant.fileVariant.topLevel, 0);
			obj->softDeleted = 1;
		}
	}
}





// Pruning removes any part of the file structure tree that is beyond the
// bounds of the file (ie that does not point to chunks).
//
// A file should only get pruned when its size is reduced.
//
// Before pruning, the chunks must be pulled from the tree and the
// level 0 tnode entries must be zeroed out.
// Could also use this for file deletion, but that's probably better handled
// by a special case.

// sffs_PruneWorker should only be called by sffs_PruneFileStructure()

static sffs_Tnode *sffs_PruneWorker(sffs_Device *dev, sffs_Tnode *tn, __u32 level, int del0)
{
	int i;
	int hasData;
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_PruneWorker\n"));            	    	
	if(tn)
	{
		hasData = 0;
		
		for(i = 0; i < SFFS_NTNODES_INTERNAL; i++)
		{
		    if(tn->internal[i] && level > 0)
		    {
		    	tn->internal[i] = sffs_PruneWorker(dev,tn->internal[i],level - 1, ( i == 0) ? del0 : 1);
		    }
		    
		    if(tn->internal[i])
		    {
		    	hasData++;
			}
		}
		
		if(hasData == 0 && del0)
		{
			// Free and return NULL
			
			sffs_FreeTnode(dev,tn);
			tn = NULL;
		}
		
	}

	return tn;
	
}

static int sffs_PruneFileStructure(sffs_Device *dev, sffs_FileStructure *fStruct)
{
	int i;
	int hasData;
	int done = 0;
	sffs_Tnode *tn;
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_PruneFileStructure\n"));            	    		
	if(fStruct->topLevel > 0)
	{
		fStruct->top = sffs_PruneWorker(dev,fStruct->top, fStruct->topLevel,0);
		
		// Now we have a tree with all the non-zero branches NULL but the height
		// is the same as it was.
		// Let's see if we can trim internal tnodes to shorten the tree.
		// We can do this if only the 0th element in the tnode is in use 
		// (ie all the non-zero are NULL)
		
		while(fStruct->topLevel && !done)
		{
			tn = fStruct->top;
			
			hasData = 0;
			for(i = 1; i <SFFS_NTNODES_INTERNAL; i++)
			{
				if(tn->internal[i])
		    	{
		    		hasData++;
				}
			}
			
			if(!hasData)
			{
				fStruct->top = tn->internal[0];
				fStruct->topLevel--;
				sffs_FreeTnode(dev,tn);
			}
			else
			{
				done = 1;
			}
		}
	}
	
	return SFFS_OK;
}





/////////////////////// End of File Structure functions. /////////////////

// sffs_CreateFreeObjects creates a bunch more objects and
// adds them to the object free list.
static int sffs_CreateFreeObjects(sffs_Device *dev, int nObjects)
{
    int i;
    sffs_Object *newObjects;
    sffs_ObjectList *list;
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_CreateFreeObjects\n"));            	    		    
    if(nObjects < 1) return SFFS_OK;
   
	// make these things

	T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_CreateFreeObjects : before YMALLOC\n"));
    newObjects = YMALLOC(nObjects * sizeof(sffs_Object));
   
    if (!newObjects)
    {
		T(SFFS_TRACE_ALLOCATE,(TSTR("sffs: Could not allocate more objects" TENDSTR)));
		return SFFS_FAIL;
    }
    
    // Hook them into the free list
    for(i = 0; i < nObjects - 1; i++)
    {
    	newObjects[i].siblings.next = (struct list_head *)(&newObjects[i+1]);
    }
    	
	newObjects[nObjects - 1].siblings.next = (void *)dev->freeObjects;
	dev->freeObjects = newObjects;
	dev->nFreeObjects+= nObjects;
	dev->nObjectsCreated+= nObjects;
	
	// Now add this bunch of Objects to a list for freeing up.
	
	list = YMALLOC(sizeof(sffs_ObjectList));
	if(!list)
	{
		T(SFFS_TRACE_ALLOCATE,(TSTR("Could not add objects to management list" TENDSTR)));
	}
	else
	{
		list->objects = newObjects;
		list->next = dev->allocatedObjectList;
		dev->allocatedObjectList = list;
	}
	
	
	
	return SFFS_OK;
}


// AllocateEmptyObject gets us a clean Object. Tries to make allocate more if we run out
static sffs_Object *sffs_AllocateEmptyObject(sffs_Device *dev)
{
	sffs_Object *tn = NULL;
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_AllocateEmptyObject\n"));            	    		    	
	// If there are none left make more
	if(!dev->freeObjects)
	{
		sffs_CreateFreeObjects(dev,SFFS_ALLOCATION_NOBJECTS);
	}
	
	if(dev->freeObjects)
	{
       	T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_AllocateEmptyObject\n")); 
		tn = dev->freeObjects;
		dev->freeObjects = (sffs_Object *)(dev->freeObjects->siblings.next);
		dev->nFreeObjects--;
		
		// Now sweeten it up...
	
		memset(tn,0,sizeof(sffs_Object));
		tn->myDev = dev;
		tn->chunkId = -1;
		tn->variantType = SFFS_OBJECT_TYPE_UNKNOWN;
		INIT_LIST_HEAD(&(tn->hardLinks));
		INIT_LIST_HEAD(&(tn->hashLink));
		INIT_LIST_HEAD(&tn->siblings);
		
		// Add it to the lost and found directory.
		// NB Can't put root or lostNFound in lostNFound so
		// check if lostNFound exists first
		if(dev->lostNFoundDir)
		{
			sffs_AddObjectToDirectory(dev->lostNFoundDir,tn);	
		}
	}
	

	return tn;
}

static sffs_Object *sffs_CreateFakeDirectory(sffs_Device *dev,int number,__u32 mode)
{

	sffs_Object *obj = sffs_CreateNewObject(dev,number,SFFS_OBJECT_TYPE_DIRECTORY);		

	if(obj)
	{
		obj->fake = 1;			// it is fake so it has no NAND presence...
		obj->renameAllowed= 0;	// ... and we're not allowed to rename it...
		obj->unlinkAllowed= 0;	// ... or unlink it
		obj->deleted = 0;
		obj->unlinked = 0;
		obj->st_mode = mode;
		obj->myDev = dev;
		obj->chunkId = 0; // Not a valid chunk.
	}
	
	return obj;
	
}


static void sffs_UnhashObject(sffs_Object *tn)
{
	int bucket;
	sffs_Device *dev = tn->myDev;
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_UnhashObject\n"));            	    		    		
	
	// If it is still linked into the bucket list, free from the list
	if(!list_empty(&tn->hashLink))
	{
		list_del_init(&tn->hashLink);
		bucket =  sffs_HashFunction(tn->objectId);
		dev->objectBucket[bucket].count--;
	}
	
}


// FreeObject frees up a Object and puts it back on the free list
static void sffs_FreeObject(sffs_Object *tn)
{
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_FreeObject\n"));            	    		    		
	sffs_Device *dev = tn->myDev;
	
	sffs_UnhashObject(tn);
	
	// Link into the free list.
	tn->siblings.next = (struct list_head *)(dev->freeObjects);
	dev->freeObjects = tn;
	dev->nFreeObjects++;
}




static void sffs_DeinitialiseObjects(sffs_Device *dev)
{
	// Free the list of allocated Objects
//				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_DeinitialiseObjects\n"));            	    		    			
	while( dev->allocatedObjectList)
	{
		YFREE(dev->allocatedObjectList->objects);
		dev->allocatedObjectList =  dev->allocatedObjectList->next;
	}
	
	dev->freeObjects = NULL;
	dev->nFreeObjects = 0;
}

static void sffs_InitialiseObjects(sffs_Device *dev)
{
	int i;
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_InitialiseObjects\n"));            	    		    				
	dev->allocatedObjectList = NULL;
	dev->freeObjects = NULL;
	dev->nFreeObjects = 0;
	
	for(i = 0; i < SFFS_NOBJECT_BUCKETS; i++)
	{
		INIT_LIST_HEAD(&dev->objectBucket[i].list);
		dev->objectBucket[i].count = 0;	
	}

}






int sffs_FindNiceObjectBucket(sffs_Device *dev)
{
	static int x = 0;
	int i;
	int l = 999;
	int lowest = 999999;

				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_FindNiceObjectBucket\n"));            	    		    						
	// First let's see if we can find one that's empty.
	
	for(i = 0; i < 10 && lowest > 0; i++)
	 {
		x++;
		x %=  SFFS_NOBJECT_BUCKETS;
		if(dev->objectBucket[x].count < lowest)
		{
			lowest = dev->objectBucket[x].count;
			l = x;
		}
		
	}
	
	// If we didn't find an empty list, then try
	// looking a bit further for a short one
	
	for(i = 0; i < 10 && lowest > 3; i++)
	 {
		x++;
		x %=  SFFS_NOBJECT_BUCKETS;
		if(dev->objectBucket[x].count < lowest)
		{
			lowest = dev->objectBucket[x].count;
			l = x;
		}
		
	}
	
	return l;
}

static int sffs_CreateNewObjectNumber(sffs_Device *dev)
{
	int bucket = sffs_FindNiceObjectBucket(dev);
	
	// Now find an object value that has not already been taken
	// by scanning the list.
	
	int found = 0;
	struct list_head *i;
	
	__u32 n = (__u32)bucket;

	//sffs_CheckObjectHashSanity();	
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_CreateNewObjectNumber\n"));            	    		    							
	while(!found)
	{
		found = 1;
		n +=  SFFS_NOBJECT_BUCKETS;
		if(1 ||dev->objectBucket[bucket].count > 0)
		{
			list_for_each(i,&dev->objectBucket[bucket].list)
			{
				// If there is already one in the list
				if(i && list_entry(i, sffs_Object,hashLink)->objectId == n)
				{
					found = 0;
				}
			}
		}
	}
	

	return n;	
}

void sffs_HashObject(sffs_Object *in)
{
	int bucket = sffs_HashFunction(in->objectId);
	sffs_Device *dev = in->myDev;
	
	if(!list_empty(&in->hashLink))
	{

	}

	
	list_add(&in->hashLink,&dev->objectBucket[bucket].list);
	dev->objectBucket[bucket].count++;

}

sffs_Object *sffs_FindObjectByNumber(sffs_Device *dev,__u32 number)
{
	int bucket = sffs_HashFunction(number);
	struct list_head *i;
	sffs_Object *in;

//sffs_InodeToObject(dir)->objectId,
//dentry->d_name.name

	list_for_each(i,&dev->objectBucket[bucket].list)
	{
		// Look if it is in the list
		if(i)
		{
			in = list_entry(i, sffs_Object,hashLink);
			if(in->objectId == number)
			{
				return in;
			}
		}
	}
	
	return NULL;
}


//chunk о object .  file ûũ ׳ а ٸ  ûũ  .
sffs_Object *sffs_CreateNewObject(sffs_Device *dev,int number,sffs_ObjectType type)
{
		
	sffs_Object *theObject;

	if(number < 0)
	{
		number = sffs_CreateNewObjectNumber(dev);
	}
	
	theObject = sffs_AllocateEmptyObject(dev);
	
	if(theObject)
	{
		theObject->fake = 0;
		theObject->renameAllowed = 1;
		theObject->unlinkAllowed = 1;
		theObject->objectId = number;
		sffs_HashObject(theObject);
		theObject->variantType = type;
#ifdef CONFIG_SFFS_WINCE
		yfsd_WinFileTimeNow(theObject->win_atime);
		theObject->win_ctime[0] = theObject->win_mtime[0] = theObject->win_atime[0];
		theObject->win_ctime[1] = theObject->win_mtime[1] = theObject->win_atime[1];

#else
		theObject->st_atime = theObject->st_mtime =	theObject->st_ctime = CURRENT_TIME;
#endif
		switch(type)
		{
			case SFFS_OBJECT_TYPE_FILE: 
				theObject->variant.fileVariant.fileSize = 0;
				theObject->variant.fileVariant.scannedFileSize = 0;
				theObject->variant.fileVariant.topLevel = 0;
				theObject->variant.fileVariant.top  = sffs_GetTnode(dev);
				break;
			case SFFS_OBJECT_TYPE_DIRECTORY:
				INIT_LIST_HEAD(&theObject->variant.directoryVariant.children);
				break;
			case SFFS_OBJECT_TYPE_SYMLINK:
				// No action required
				break;
			case SFFS_OBJECT_TYPE_HARDLINK:
				// No action required
				break;
			case SFFS_OBJECT_TYPE_SPECIAL:
				// No action required
				break;
			case SFFS_OBJECT_TYPE_UNKNOWN:
				// todo this should not happen
				break;
		}
	}
    
	
	return theObject;
}

sffs_Object *sffs_FindOrCreateObjectByNumber(sffs_Device *dev, int number,sffs_ObjectType type)
{
	sffs_Object *theObject = NULL;

	if(number > 0)
	{
		theObject = sffs_FindObjectByNumber(dev,number);
	}
	
	if(!theObject)
	{
		theObject = sffs_CreateNewObject(dev,number,type);
	}
	
	return theObject;

}

char *sffs_CloneString(const char *str)
{
	char *newStr = NULL;
	
	if(str && *str)
	{
		newStr = YMALLOC(strlen(str) + 1);
		strcpy(newStr,str);
	}

	return newStr;
	
}

//
// Mknod (create) a new object.
// equivalentObject only has meaning for a hard link;
// aliasString only has meaning for a sumlink.
// rdev only has meaning for devices (a subset of special objects)
sffs_Object *sffs_MknodObject( sffs_ObjectType type,
								 sffs_Object *parent,
								 const char *name, 
								 __u32 mode,
								 __u32 uid,
								 __u32 gid,
								 sffs_Object *equivalentObject,
								 const char *aliasString,
								 __u32 rdev)
{
	sffs_Object *in;

	sffs_Device *dev = parent->myDev;


    // writeϷµ ̹ Ұ.
	if(sffs_FindObjectByName(parent,name))
	{
		return NULL;
	}

    //type : , 丮...
	in = sffs_CreateNewObject(dev,-1,type);
	
	if(in)
	{
		in->chunkId = -1;
		in->valid = 1;
		in->variantType = type;

		in->st_mode  = mode;
		
#ifdef CONFIG_SFFS_WINCE
		yfsd_WinFileTimeNow(in->win_atime);
		in->win_ctime[0] = in->win_mtime[0] = in->win_atime[0];
		in->win_ctime[1] = in->win_mtime[1] = in->win_atime[0];
		
#else
		in->st_atime = in->st_mtime = in->st_ctime = CURRENT_TIME;
		in->st_rdev  = rdev;
		in->st_uid   = uid;
		in->st_gid   = gid;
#endif		
		in->nDataChunks = 0;

		sffs_SetObjectName(in,name);

        //Ƽ Ʈ 1 ؼ 忡 writeϰ Ѵ.
		in->dirty = 1;
		
		sffs_AddObjectToDirectory(parent,in);
		
		in->myDev = parent->myDev;
		
				
		switch(type)
		{
			case SFFS_OBJECT_TYPE_SYMLINK:
				in->variant.symLinkVariant.alias = sffs_CloneString(aliasString);
				break;
			case SFFS_OBJECT_TYPE_HARDLINK:
				in->variant.hardLinkVariant.equivalentObject = equivalentObject;
				in->variant.hardLinkVariant.equivalentObjectId = equivalentObject->objectId;
				list_add(&in->hardLinks,&equivalentObject->hardLinks);
				break;
			case SFFS_OBJECT_TYPE_FILE: // do nothing
			case SFFS_OBJECT_TYPE_DIRECTORY: // do nothing
			case SFFS_OBJECT_TYPE_SPECIAL: // do nothing
			case SFFS_OBJECT_TYPE_UNKNOWN:
				break;
		}

        //  忡 writeѴ.
		if(/*sffs_GetNumberOfFreeChunks(dev) <= 0 || */
		   sffs_UpdateObjectHeader(in,name,0) < 0)
		{
			// Could not create the object header, fail the creation
			sffs_AbortHalfCreatedObject(in);
			in = NULL;
		}

	}
	
	return in;
}

sffs_Object *sffs_MknodFile(sffs_Object *parent,const char *name, __u32 mode, __u32 uid, __u32 gid)
{
	return sffs_MknodObject(SFFS_OBJECT_TYPE_FILE,parent,name,mode,uid,gid,NULL,NULL,0);
}

sffs_Object *sffs_MknodDirectory(sffs_Object *parent,const char *name, __u32 mode, __u32 uid, __u32 gid)
{
	return sffs_MknodObject(SFFS_OBJECT_TYPE_DIRECTORY,parent,name,mode,uid,gid,NULL,NULL,0);
}

sffs_Object *sffs_MknodSpecial(sffs_Object *parent,const char *name, __u32 mode, __u32 uid, __u32 gid, __u32 rdev)
{
	return sffs_MknodObject(SFFS_OBJECT_TYPE_DIRECTORY,parent,name,mode,uid,gid,NULL,NULL,rdev);
}

sffs_Object *sffs_MknodSymLink(sffs_Object *parent,const char *name, __u32 mode, __u32 uid, __u32 gid,const char *alias)
{
	return sffs_MknodObject(SFFS_OBJECT_TYPE_SYMLINK,parent,name,mode,uid,gid,NULL,alias,0);
}

// NB sffs_Link returns the object id of the equivalent object.
sffs_Object *sffs_Link(sffs_Object *parent, const char *name, sffs_Object *equivalentObject)
{
	// Get the real object in case we were fed a hard link as an equivalent object
	equivalentObject = sffs_GetEquivalentObject(equivalentObject);
	
	if(sffs_MknodObject(SFFS_OBJECT_TYPE_HARDLINK,parent,name,0,0,0,equivalentObject,NULL,0))
	{
		return equivalentObject;
	}
	else
	{
		return NULL;
	}
	
}


static int sffs_ChangeObjectName(sffs_Object *obj, sffs_Object *newDir, const char *newName,int force)
{
	int unlinkOp;

	if(newDir == NULL)
	{
		newDir = obj->parent; // use the old directory
	}

	unlinkOp = (newDir == obj->myDev->unlinkedDir && obj->variantType == SFFS_OBJECT_TYPE_FILE);
	
	// If the object is a file going into the unlinked directory, then it is OK to just stuff it in since
	// duplicate names are allowed.
	// Otherwise only proceed if the new name does not exist and if we're putting it into a directory.
	if( (unlinkOp|| 
		 force || 
		 !sffs_FindObjectByName(newDir,newName))  &&
	     newDir->variantType == SFFS_OBJECT_TYPE_DIRECTORY)
	{
		sffs_SetObjectName(obj,newName);
		obj->dirty = 1;
		
		sffs_AddObjectToDirectory(newDir,obj);
		
		if(unlinkOp) obj->unlinked = 1;
		
		
		if(sffs_UpdateObjectHeader(obj,newName,0) >= 0)
		{
			return SFFS_OK;
		}
	}
	
	return SFFS_FAIL;
}



int sffs_RenameObject(sffs_Object *oldDir, const char *oldName, sffs_Object *newDir, const char *newName)
{
	sffs_Object *obj;
	int force = 0;
	
#ifdef CONFIG_SFFS_CASE_INSENSITIVE
	// Special case for WinCE.
	// While look-up is case insensitive, the name isn't.
	// THerefore we might want to change x.txt to X.txt
	if(oldDir == newDir && _stricmp(oldName,newName) == 0)
	{
		force = 1;
	}	
#endif
	
	obj = sffs_FindObjectByName(oldDir,oldName);
	if(obj && obj->renameAllowed)
	{
		return sffs_ChangeObjectName(obj,newDir,newName,force);
	}
	return SFFS_FAIL;
}



static int sffs_CheckObjectHashSanity(sffs_Device *dev)
{
	// Scan the buckets and check that the lists 
	// have as many members as the count says there are
	int bucket;
	int countEm;
	struct list_head *j;
	int ok = SFFS_OK;
	
	for(bucket = 0; bucket < SFFS_NOBJECT_BUCKETS; bucket++)
	{
		countEm = 0;
		
		list_for_each(j,&dev->objectBucket[bucket].list)
		{
			countEm++;
		}
		
		if(countEm != dev->objectBucket[bucket].count)
		{
			T(SFFS_TRACE_ERROR,(TSTR("Inode hash inconsistency" TENDSTR)));
			ok = SFFS_FAIL;
		}
	}

	return ok;
}

#if 0
void sffs_ObjectTest(sffs_Device *dev)
{
	sffs_Object *in[1000];
	int inNo[1000];
	sffs_Object *inold[1000];
	int i;
	int j;
	
	memset(in,0,1000*sizeof(sffs_Object *));
	memset(inold,0,1000*sizeof(sffs_Object *));
	
	sffs_CheckObjectHashSanity(dev);
	
	for(j = 0; j < 10; j++)
	{
		
		for(i = 0; i < 1000; i++)
		{
			in[i] = sffs_CreateNewObject(dev,-1,SFFS_OBJECT_TYPE_FILE);
			if(!in[i])
			{
				YINFO("No more inodes");
			}
			else
			{
				inNo[i] = in[i]->objectId;
			}
		}
		
		for(i = 0; i < 1000; i++)
		{
			if(sffs_FindObjectByNumber(dev,inNo[i]) != in[i])
			{
			}
			else
			{
			}
		}
		
		sffs_CheckObjectHashSanity(dev);
	
		for(i = 0; i < 1000; i+=3)
		{
			sffs_FreeObject(in[i]);	
			in[i] = NULL;
		}
		
	
		sffs_CheckObjectHashSanity(dev);
	}
		
}

#endif

/////////////////////////// Block Management and Page Allocation ///////////////////


static int sffs_InitialiseBlocks(sffs_Device *dev,int nBlocks)
{
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_InitialiseBlocks\n"));            	    		    								    
	dev->allocationBlock = -1; // force it to get a new one
	//Todo we're assuming the malloc will pass.
	dev->blockInfo = YMALLOC(nBlocks * sizeof(sffs_BlockInfo));
	// Set up dynamic blockinfo stuff.
	dev->chunkBitmapStride = (dev->nChunksPerBlock+7)/8;
	dev->chunkBits = YMALLOC(dev->chunkBitmapStride * nBlocks);
	if(dev->blockInfo && dev->chunkBits)
	{
		memset(dev->blockInfo,0,nBlocks * sizeof(sffs_BlockInfo));
		memset(dev->chunkBits,0,dev->chunkBitmapStride * nBlocks);
		return SFFS_OK;
	}
	
	return SFFS_FAIL;
	
}

static void sffs_DeinitialiseBlocks(sffs_Device *dev)
{
	YFREE(dev->blockInfo);
	dev->blockInfo = NULL;
	YFREE(dev->chunkBits);
	dev->chunkBits = NULL;
}

// FindDiretiestBlock is used to select the dirtiest block (or close enough)
// for garbage collection.

static int sffs_FindDirtiestBlock(sffs_Device *dev,int aggressive)
{
	
	int b = dev->m_pBlockManager->bmStartData;
	
	int i;
	int iterations;
	int dirtiest = -1;
	int pagesInUse; 
	sffs_BlockInfo *bi;
	sffs_BlockInfo *ScanA;
//	sffs_BlockInfo *ScanB;
	int blockNum = 0;
	int Minium=100000000;
//	int sum = 0;
//	int average;
	int distributionSum = 0;
//	int distributionCheck = 0;
	int distributionAverage;
//    int wholeCnt;
	
  


	#if DEBUG
	#endif
/*
	for(blockNum = dev->startBlock; blockNum < dev->endBlock;blockNum++)
	{
		ScanB = sffs_GetBlockInfo(dev,blockNum);
		if(ScanB->blockState != SFFS_BLOCK_STATE_DEAD)
			sum += ScanB->nErased;
	}

	wholeCnt = (dev->endBlock - dev->startBlock + 1);

	average = sum / wholeCnt;
*/
     	#if DEBUG
		
	//printk(KERN_INFO"--------------------------------------\n");
	//printk(KERN_INFO" nEradsed average : %d \n",average);
	//printk(KERN_INFO"--------------------------------------\n");

	#endif


/*
	for(blockNum = dev->startBlock; blockNum < dev->endBlock;blockNum++)
	{
		ScanB = sffs_GetBlockInfo(dev,blockNum);
		distributionCheck = (ScanB->nErased- average)*10;
		if(distributionCheck < 0) distributionCheck = (distributionCheck*(-1));
		distributionSum += (distributionCheck*distributionCheck);
	}
*/
	distributionAverage = 10000*distributionSum / (dev->endBlock - dev->startBlock + 1);

       #if DEBUG
	//printk(KERN_INFO"+++++++++++++++++++++++++++++++++++++++++++++\n");
	//printk(KERN_INFO"Distribution Average : %d \n", distributionAverage);
	//printk(KERN_INFO"+++++++++++++++++++++++++++++++++++++++++++++\n");
	
	//printk(KERN_INFO"----------------------------------------------------------\n");
	//printk(KERN_INFO"yaffs: dev->currentDirtyChecker : %d\n",dev->currentDirtyChecker);
	//printk(KERN_INFO"----------------------------------------------------------\n");

	#endif

	// If we're doing aggressive GC then we are happy to take a less-dirty block, and
	// search further.


	pagesInUse = (aggressive)? dev->nChunksPerBlock : SFFS_PASSIVE_GC_CHUNKS + 1;

	  
	for(blockNum = dev->m_pBlockManager->bmStartData; blockNum <= dev->m_pBlockManager->bmEndData;blockNum++)
	{
	      
		ScanA = sffs_GetBlockInfo(dev,blockNum);
		if( Minium > ScanA->nErased)
			{
                Minium = ScanA->nErased;
				/*if(ScanA->blockState == SFFS_BLOCK_STATE_FULL &&
					   (ScanA->pagesInUse - ScanA->softDeletions )< pagesInUse)
					{
						if (ScanA->blockState != SFFS_BLOCK_STATE_DEAD &&
							ScanA->blockState != SFFS_BLOCK_STATE_DIRTY)
							
					}*/
			}
	}
	


	#if DEBUG
	#endif
    
	if(aggressive)
	{
		iterations = dev->m_pBlockManager->bmEndData - dev->m_pBlockManager->bmStartData + 1;
	}
	else
	{
		iterations = dev->m_pBlockManager->bmEndData - dev->m_pBlockManager->bmStartData + 1;
	}




	for(i = 0; i <= iterations /*&& pagesInUse > 0*/ ; i++)
	{
		//b++;
		if ( b < dev->m_pBlockManager->bmStartData || b > dev->m_pBlockManager->bmEndData)
		{
			b =  dev->m_pBlockManager->bmStartData;
		}

		if(b < dev->m_pBlockManager->bmStartData || b > dev->m_pBlockManager->bmEndData)
		{
			T(SFFS_TRACE_ERROR,(TSTR("**>> Block %d is not valid" TENDSTR),b));
			YBUG();
		}
		
		bi = sffs_GetBlockInfo(dev,b);

		if(bi->blockState == SFFS_BLOCK_STATE_FULL/* && bi->blockState != SFFS_BLOCK_STATE_DIRTY*/ &&
		   (bi->pagesInUse - bi->softDeletions )< pagesInUse)
		{
            if(bi->nErased == Minium || bi->blockState == SFFS_BLOCK_STATE_ALLOCATING
                || bi->blockState == SFFS_BLOCK_STATE_DEAD||bi->blockState == SFFS_BLOCK_STATE_DIRTY
                || bi->blockState == SFFS_BLOCK_STATE_EMPTY)
            {
    			dirtiest = b;
	    		pagesInUse = (bi->pagesInUse - bi->softDeletions );
            }
		}
		b++;
	}
	
	/*dev->currentDirtyChecker = b;*/
	
	return dirtiest;
}


static void sffs_BlockBecameDirty(sffs_Device *dev,int blockNo)
{
	sffs_BlockInfo *bi = sffs_GetBlockInfo(dev,blockNo);
	
	int erasedOk = 0;
	
	// If the block is still healthy erase it and mark as clean.
	// If the block has had a data failure, then retire it.
	bi->blockState = SFFS_BLOCK_STATE_DIRTY;

	if(!bi->needsRetiring)
	{
		erasedOk = sffs_EraseBlockInNAND(dev,blockNo);
		if(!erasedOk)
		{
			T(SFFS_TRACE_ERROR | SFFS_TRACE_BAD_BLOCKS,(TSTR("**>> Erasure failed %d" TENDSTR),blockNo));
		}
	}
	
	if( erasedOk )
	{
		// Clean it up...
		bi->blockState = SFFS_BLOCK_STATE_EMPTY;
		dev->nErasedBlocks++;
		bi->pagesInUse = 0;
		bi->softDeletions = 0;
		sffs_ClearChunkBits(dev,blockNo);
	
		T(SFFS_TRACE_ERASE,(TSTR("Erased block %d" TENDSTR),blockNo));
	}
	else
	{
		sffs_RetireBlock(dev,blockNo);
		T(SFFS_TRACE_ERROR | SFFS_TRACE_BAD_BLOCKS,(TSTR("**>> Block %d retired" TENDSTR),blockNo));
	}
}


static int sffs_FindBlockForAllocation(sffs_Device *dev)
{
	int i;
	
	sffs_BlockInfo *bi;
				T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_FindBlockForAllocation\n"));            	    		    								    	
	if(dev->nErasedBlocks < 1)
	{
		// Hoosterman we've got a problem.
		// Can't get space to gc
		T(SFFS_TRACE_ERROR, (TSTR("sffs tragedy: no space during gc" TENDSTR)));

		return -1;
	}
	
	// Find an empty block.
	
	for(i = dev->m_pBlockManager->bmStartData; i <= dev->m_pBlockManager->bmEndData; i++)
	{
		dev->allocationBlockFinder++;
		if(dev->allocationBlockFinder <dev->startBlock || dev->allocationBlockFinder> dev->endBlock) 
		{
			dev->allocationBlockFinder = dev->startBlock;
		}
		
		bi = sffs_GetBlockInfo(dev,dev->allocationBlockFinder);

		if(bi->blockState == SFFS_BLOCK_STATE_EMPTY)
		{
			bi->blockState = SFFS_BLOCK_STATE_ALLOCATING;
			dev->nErasedBlocks--;			
			return dev->allocationBlockFinder;
		}
	}
	
	return -1;	
}



static int sffs_AllocateChunk(sffs_Device *dev,int useReserve)
{
	int retVal;
	sffs_BlockInfo *bi;

	
	if(dev->allocationBlock < 0)
	{
		// Get next block to allocate off
		dev->allocationBlock = sffs_FindBlockForAllocation(dev);
		dev->allocationPage = 0;
	}
	
	if(!useReserve &&  dev->nErasedBlocks <= dev->nReservedBlocks)
	{
		// Not enough space to allocate unless we're allowed to use the reserve.
		return -1;
	}
	
	// Next page please....
	if(dev->allocationBlock >= 0)
	{
		bi = sffs_GetBlockInfo(dev,dev->allocationBlock);
		
		retVal = (dev->allocationBlock * dev->nChunksPerBlock) + 
			  	  dev->allocationPage;
		bi->pagesInUse++;
		sffs_SetChunkBit(dev,dev->allocationBlock,dev->allocationPage);

		dev->allocationPage++;
		
		dev->nFreeChunks--;
		
		// If the block is full set the state to full
		if(dev->allocationPage >= dev->nChunksPerBlock)
		{
			bi->blockState = SFFS_BLOCK_STATE_FULL;
			dev->allocationBlock = -1;
		}


		return retVal;
		
	}
	T(SFFS_TRACE_ERROR,(TSTR("!!!!!!!!! Allocator out !!!!!!!!!!!!!!!!!" TENDSTR)));

	return -1;	
}

// To determine if we have enough space we just look at the 
// number of erased blocks.
// The cache is allowed to use reserved blocks.

static int sffs_CheckSpaceForChunkCache(sffs_Device *dev)
{
	return (dev->nErasedBlocks >= dev->nReservedBlocks);
}


static int  sffs_GarbageCollectBlock(sffs_Device *dev,int block)
{
	int oldChunk;
	int newChunk;
	int chunkInBlock;
	int markNAND;
	
	
	sffs_Spare spare;
	sffs_Tags  tags;
	__u8  buffer[SFFS_BYTES_PER_CHUNK];
	
//	sffs_BlockInfo *bi = sffs_GetBlockInfo(dev,block);
	
	sffs_Object *object;

    
	T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_GarbageCollectBlock\n"));

	
	for(chunkInBlock = 0,oldChunk = block * dev->nChunksPerBlock; 
	    chunkInBlock < dev->nChunksPerBlock && sffs_StillSomeChunkBits(dev,block);
	    chunkInBlock++, oldChunk++ )
	{
		if(sffs_CheckChunkBit(dev,block,chunkInBlock))
		{
            
			// This page is in use and might need to be copied off
			
			markNAND = 1;
			
			//T(("copying page %x from %d to %d\n",mask,oldChunk,newChunk));
			
			sffs_ReadChunkFromNAND(dev,oldChunk,buffer, &spare,1);
			
			sffs_GetTagsFromSpare(dev,&spare,&tags);

			object = sffs_FindObjectByNumber(dev,tags.objectId);
			
			if(object && object->deleted && tags.chunkId != 0)
			{
				// Data chunk in a deleted file, throw it away
				// It's a deleted data chunk,
				// No need to copy this, just forget about it and fix up the
				// object.
				
				//sffs_PutChunkIntoFile(object, tags.chunkId, 0,0); 
				object->nDataChunks--;
				
				if(object->nDataChunks <= 0)
				{
                    
					// Time to delete the file too
					sffs_FreeTnode(object->myDev,object->variant.fileVariant.top);
					object->variant.fileVariant.top = NULL;
					T(SFFS_TRACE_TRACING,(TSTR("sffs: About to finally delete object %d" TENDSTR),object->objectId));
					sffs_DoGenericObjectDeletion(object);					
 				}
				markNAND = 0;
			}
			else if( 0 /* Todo object && object->deleted && object->nDataChunks == 0 */)
			{
				// Deleted object header with no data chunks.
				// Can be discarded and the file deleted.
				object->chunkId = 0;
				sffs_FreeTnode(object->myDev,object->variant.fileVariant.top);
				object->variant.fileVariant.top = NULL;
				sffs_DoGenericObjectDeletion(object);
				
			}
			else if(object)
			{
				// It's either a data chunk in a live file or
				// an ObjectHeader, so we're interested in it.
				// NB Need to keep the ObjectHeaders of deleted files
				// until the whole file has been deleted off
				tags.serialNumber++;
				sffs_LoadTagsIntoSpare(&spare,&tags);

				dev->nGCCopies++;

				newChunk = sffs_WriteNewChunkToNAND(dev, buffer, &spare,1);
			
				if(newChunk < 0)
				{
					return SFFS_FAIL;
				}
			
				// Ok, now fix up the Tnodes etc.
			
				if(tags.chunkId == 0)
				{
					// It's a header
					object->chunkId = newChunk;
				}
				else
				{
					// It's a data chunk
					sffs_PutChunkIntoFile(object, tags.chunkId, newChunk,0);
				}
			}
			sffs_DeleteChunk(dev,oldChunk,markNAND);			
			
		}
	}
	return SFFS_OK;
}


static sffs_Object *sffs_FindDeletedUnlinkedFile(sffs_Device *dev)
{
	// find a file to delete
	struct list_head *i;	
	sffs_Object *l;


	//Scan the unlinked files looking for one to delete
	list_for_each(i,&dev->unlinkedDir->variant.directoryVariant.children)
	{
		if(i)
		{
			l = list_entry(i, sffs_Object,siblings);
			if(l->deleted)
			{
				return l;			
			}
		}
	}	
	return NULL;
}



static void sffs_DoUnlinkedFileDeletion(sffs_Device *dev)
{
	// This does background deletion on unlinked files.. only deleted ones.
	// If we don't have a file we're working on then find one
	if(!dev->unlinkedDeletion && dev->nDeletedFiles > 0)
	{
		dev->unlinkedDeletion = sffs_FindDeletedUnlinkedFile(dev);
	}
	
	// OK, we're working on a file...
	if(dev->unlinkedDeletion)
	{
		sffs_Object *obj = dev->unlinkedDeletion;
		int delresult;
		int limit; // Number of chunks to delete in a file.
				   // NB this can be exceeded, but not by much.
				   
		limit = -1;

		delresult = sffs_DeleteWorker(obj, obj->variant.fileVariant.top, obj->variant.fileVariant.topLevel, 0,&limit);
		
		if(obj->nDataChunks == 0)
		{
			// Done all the deleting of data chunks.
			// Now dump the header and clean up
			sffs_FreeTnode(dev,obj->variant.fileVariant.top);
			obj->variant.fileVariant.top = NULL;
			sffs_DoGenericObjectDeletion(obj);
			dev->nDeletedFiles--;
			dev->nUnlinkedFiles--;
			dev->nBackgroundDeletions++;
			dev->unlinkedDeletion = NULL;	
		}
	}
}


#if 0
#define SFFS_GARBAGE_COLLECT_LOW_WATER 2
static int sffs_CheckGarbageCollection(sffs_Device *dev)
{
	int block;
	int aggressive=0;
	
	//sffs_DoUnlinkedFileDeletion(dev);
	
	if(dev->nErasedBlocks <= (dev->nReservedBlocks + SFFS_GARBAGE_COLLECT_LOW_WATER))
	{
		aggressive = 1;
	}		
	
	if(aggressive)
	{
		block = sffs_FindDirtiestBlock(dev,aggressive);
		
		if(block >= 0)
		{
			dev->garbageCollections++;
			return sffs_GarbageCollectBlock(dev,block);
		}	
		else
		{
			return SFFS_FAIL;
		}
	}

	return SFFS_OK;
}
#endif

// New garbage collector
// If we're very low on erased blocks then we do aggressive garbage collection
// otherwise we do "passive" garbage collection.
// Aggressive gc looks further (whole array) and will accept dirtier blocks.
// Passive gc only inspects smaller areas and will only accept cleaner blocks.
//
// The idea is to help clear out space in a more spread-out manner.
// Dunno if it really does anything useful.
//
static int sffs_CheckGarbageCollection(sffs_Device *dev)
{
	int block;
	int aggressive=0;

	if(dev->nErasedBlocks <= (dev->nReservedBlocks + 1))
	{
		aggressive = 1;
	}		
	
	block = sffs_FindDirtiestBlock(dev,aggressive);
    
	if(block >= 0)
	{
		dev->garbageCollections++;
		if(!aggressive)
		{
			dev->passiveGarbageCollections++;
		}

		T(SFFS_TRACE_GC,(TSTR("sffs: GC erasedBlocks %d aggressive %d" TENDSTR),dev->nErasedBlocks,aggressive));

		return sffs_GarbageCollectBlock(dev,block);
	}	

	return aggressive ? SFFS_FAIL : SFFS_OK;
}


//////////////////////////// TAGS ///////////////////////////////////////

static void sffs_LoadTagsIntoSpare(sffs_Spare *sparePtr, sffs_Tags *tagsPtr)
{
	sffs_TagsUnion *tu = (sffs_TagsUnion *)tagsPtr;
					T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_LoadTagsIntoSpare\n"));            	    		    								    		
	sffs_CalcTagsECC(tagsPtr);
	
	sparePtr->tagByte0 = tu->asBytes[0];
	sparePtr->tagByte1 = tu->asBytes[1];
	sparePtr->tagByte2 = tu->asBytes[2];
	sparePtr->tagByte3 = tu->asBytes[3];
	sparePtr->tagByte4 = tu->asBytes[4];
	sparePtr->tagByte5 = tu->asBytes[5];
	sparePtr->tagByte6 = tu->asBytes[6];
	sparePtr->tagByte7 = tu->asBytes[7];
}

static void sffs_GetTagsFromSpare(sffs_Device *dev, sffs_Spare *sparePtr,sffs_Tags *tagsPtr)
{
	sffs_TagsUnion *tu = (sffs_TagsUnion *)tagsPtr;
	int result;
					T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_GetTagsFromSpare\n"));            	    		    								    		
	tu->asBytes[0]= sparePtr->tagByte0;
	tu->asBytes[1]= sparePtr->tagByte1;
	tu->asBytes[2]= sparePtr->tagByte2;
	tu->asBytes[3]= sparePtr->tagByte3;
	tu->asBytes[4]= sparePtr->tagByte4;
	tu->asBytes[5]= sparePtr->tagByte5;
	tu->asBytes[6]= sparePtr->tagByte6;
	tu->asBytes[7]= sparePtr->tagByte7;
	
	result =  sffs_CheckECCOnTags(tagsPtr);
	if(result> 0)
	{
		dev->tagsEccFixed++;
	}
	else if(result <0)
	{
		dev->tagsEccUnfixed++;
	}
}

static void sffs_SpareInitialise(sffs_Spare *spare)
{
	memset(spare,0xFF,sizeof(sffs_Spare));
}

static int sffs_ReadChunkTagsFromNAND(sffs_Device *dev,int chunkInNAND, sffs_Tags *tags, int *chunkDeleted)
{
					T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_ReadChunkTagsFromNAND\n"));            	    		    								    		    
	if(tags)
	{
		sffs_Spare spare;
		if(sffs_ReadChunkFromNAND(dev,chunkInNAND,NULL,&spare,1) == SFFS_OK)
		{
			*chunkDeleted = (sffs_CountBits(spare.pageStatus) < 7) ? 1 : 0;
			sffs_GetTagsFromSpare(dev,&spare,tags);
			return SFFS_OK;
		}
		else
		{
			return SFFS_FAIL;
		}
	}
	
	return SFFS_OK;
}

#if 0
static int sffs_WriteChunkWithTagsToNAND(sffs_Device *dev,int chunkInNAND, const __u8 *buffer, sffs_Tags *tags)
{
	// NB There must be tags, data is optional
	// If there is data, then an ECC is calculated on it.
	
	sffs_Spare spare;
	
	if(!tags)
	{
		return SFFS_FAIL;
	}
	
	sffs_SpareInitialise(&spare);
	
	if(!dev->useNANDECC && buffer)
	{
		sffs_CalcECC(buffer,&spare);
	}
	
	sffs_LoadTagsIntoSpare(&spare,tags);
	
	return sffs_WriteChunkToNAND(dev,chunkInNAND,buffer,&spare);
	
}
#endif


static int sffs_WriteNewChunkWithTagsToNAND(sffs_Device *dev, const __u8 *buffer, sffs_Tags *tags, int useReserve)
{
	// NB There must be tags, data is optional
	// If there is data, then an ECC is calculated on it.
	
	sffs_Spare spare;

  	T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_WriteNewChunkWithTagsToNAND\n"));
            
	if(!tags)
	{
		return SFFS_FAIL;
	}
	
	sffs_SpareInitialise(&spare);
	
	if(!dev->useNANDECC && buffer)
	{
		sffs_CalcECC(buffer,&spare);
	}
	
	sffs_LoadTagsIntoSpare(&spare,tags);
	
	return sffs_WriteNewChunkToNAND(dev,buffer,&spare,useReserve);
	
}

static int sffs_TagsMatch(const sffs_Tags *tags, int objectId, int chunkInObject, int chunkDeleted)
{
	return 	(  tags->chunkId == chunkInObject &&
			   tags->objectId == objectId &&
			   !chunkDeleted) ? 1 : 0;
	
}



int sffs_FindChunkInFile(sffs_Object *in,int chunkInInode,sffs_Tags *tags)
{
	//Get the Tnode, then get the level 0 offset chunk offset
    sffs_Tnode *tn;     
    int theChunk = -1;
    sffs_Tags localTags;
    int i;
    int found = 0;
    int chunkDeleted;
    
    sffs_Device *dev = in->myDev;
    
					T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_FindChunkInFile\n"));            	    		    								    		        
    if(!tags)
    {
    	// Passed a NULL, so use our own tags space
    	tags = &localTags;
    }
    
    tn = sffs_FindLevel0Tnode(dev,&in->variant.fileVariant, chunkInInode);
    
    if(tn)
    {
		theChunk = tn->level0[chunkInInode & SFFS_TNODES_LEVEL0_MASK] << dev->chunkGroupBits;

		// Now we need to do the shifting etc and search for it
		for(i = 0,found = 0; theChunk && i < dev->chunkGroupSize && !found; i++)
		{
			sffs_ReadChunkTagsFromNAND(dev,theChunk,tags,&chunkDeleted);
			if(sffs_TagsMatch(tags,in->objectId,chunkInInode,chunkDeleted))
			{
				// found it;
				found = 1;
			}
			else
			{
				theChunk++;
			}
		}
    }
    return found ? theChunk : -1;
}

int sffs_FindAndDeleteChunkInFile(sffs_Object *in,int chunkInInode,sffs_Tags *tags)
{
	//Get the Tnode, then get the level 0 offset chunk offset
    sffs_Tnode *tn;     
    int theChunk = -1;
    sffs_Tags localTags;
    int i;
    int found = 0;
    sffs_Device *dev = in->myDev;
    int chunkDeleted;
    int cnt;
    
    if(!tags)
    {
    	// Passed a NULL, so use our own tags space
    	tags = &localTags;
    }
    
    tn = sffs_FindLevel0Tnode(dev,&in->variant.fileVariant, chunkInInode); 
    
    if(tn)
    {
    
		theChunk = tn->level0[chunkInInode & SFFS_TNODES_LEVEL0_MASK] << dev->chunkGroupBits;
    
		// Now we need to do the shifting etc and search for it
		for(i = 0,found = 0; theChunk && i < dev->chunkGroupSize && !found; i++)
		{
			sffs_ReadChunkTagsFromNAND(dev,theChunk,tags,&chunkDeleted);
			if(sffs_TagsMatch(tags,in->objectId,chunkInInode,chunkDeleted))
			{
				// found it;
				found = 1;
			}
			else
			{
				theChunk++;
			}
		}

		// Delete the entry in the filestructure
		if(found)
		{

			tn->level0[chunkInInode & SFFS_TNODES_LEVEL0_MASK] = 0;
		}
    }
    else
    {

    }
    
    if(!found)
    {

    }
    return found ? theChunk : -1;
}


#ifdef SFFS_PARANOID

static int sffs_CheckFileSanity(sffs_Object *in)
{
	int chunk;
	int nChunks;
	int fSize;
	int failed = 0;
	int objId;
	sffs_Tnode *tn;
    sffs_Tags localTags;
    sffs_Tags *tags = &localTags;
    int theChunk;
    int chunkDeleted;
    

	if(in->variantType != SFFS_OBJECT_TYPE_FILE)
	{
		return SFFS_FAIL;
	}
	
	objId = in->objectId;
	fSize  = in->variant.fileVariant.fileSize;
	nChunks = (fSize + in->myDev->nBytesPerChunk -1)/in->myDev->nBytesPerChunk;

   
	for(chunk = 1; chunk <= nChunks; chunk++)
	{
		tn = sffs_FindLevel0Tnode(in->myDev,&in->variant.fileVariant, chunk);
    
		if(tn)
		{
    
			theChunk = tn->level0[chunk & SFFS_TNODES_LEVEL0_MASK] << in->myDev->chunkGroupBits;
    
				sffs_ReadChunkTagsFromNAND(in->myDev,theChunk,tags,&chunkDeleted);
				if(sffs_TagsMatch(tags,in->objectId,chunk,chunkDeleted))
				{
				
				}
				else
				{
					//T(("File problem file [%d,%d] NAND %d  tags[%d,%d]\n",
					//		objId,chunk,theChunk,tags->chunkId,tags->objectId);

					failed = 1;
								
				}
    
		}
		else
		{
			//T(("No level 0 found for %d\n", chunk));
		}
	}

	return failed ? SFFS_FAIL : SFFS_OK;
}

#endif

static int sffs_PutChunkIntoFile(sffs_Object *in,int chunkInInode, int chunkInNAND, int inScan)
{
	sffs_Tnode *tn,*exTn;
	sffs_Device *dev = in->myDev;
	int existingChunk;
	sffs_Tags existingTags;
	sffs_Tags newTags;
	unsigned existingSerial, newSerial;
	
	int newChunkDeleted;
	//js
//	__u8 data[512];
//    sffs_Spare spare;
//    sffs_Tags tags;
//    int i;
    

	
	
	tn = sffs_AddOrFindLevel0Tnode(dev,&in->variant.fileVariant, chunkInInode);
	if(!tn)
	{
		return SFFS_FAIL;
	}

    
	existingChunk = tn->level0[chunkInInode & SFFS_TNODES_LEVEL0_MASK];		

 
	if(inScan)
	{
		if(existingChunk != 0)
		{           
			sffs_ReadChunkTagsFromNAND(dev,chunkInNAND, &newTags,&newChunkDeleted);
			
			existingChunk = sffs_FindChunkInFile(in,chunkInInode, &existingTags);

			if(existingChunk <=0)
			{
				T(SFFS_TRACE_ERROR, (TSTR("sffs tragedy: existing chunk < 0 in scan" TENDSTR)));
			}

 			newSerial = newTags.serialNumber;
			existingSerial = existingTags.serialNumber;



                printk("INNER\n");

                printk("g_nMount : %d / existingChunk : %d / newSerial : %d / g_nIsCrash : %d / existingSerial : %d\n",
                    g_nMount,existingChunk,newSerial,g_nIsCrash,existingSerial);
            
    			//if( existingChunk <= 0 ||
    			if( existingChunk < 0 ||
    			    ((existingSerial+1) & 3) == newSerial || g_nMount==M_FAIL )
    			{
                    //ο  .
                    if(g_nMount==M_FAIL && g_nIsCrash==1)
                    {
                     	printk("CASE  : chunkInNAND : %d / existingChunk : %d\n",chunkInNAND,existingChunk);

                        sffs_DeleteChunk(dev,chunkInNAND,1);
                        printk("MOUNT Fail, CRASH!\n");
                        chunkInNAND=existingChunk;                        
                    }
                    else if(g_nMount==M_FAIL && g_nIsCrash==0)                    
                    {
                        sffs_DeleteChunk(dev,existingChunk,1);  
                    }
 
     			}
    			else
    			{
    				//  				

                    if(g_nMount==M_SUCCESS)
                	{
                        if(existingChunk>0)
                            sffs_DeleteChunk(dev,existingChunk,1);
                            //chunkInNAND=existingChunk;
                    }
                    else
                    {
                        if(existingChunk>0)
                            sffs_DeleteChunk(dev,chunkInNAND,1);
                    }
    				return SFFS_OK;
    			}
            
		}

	}

	if(existingChunk == 0)
	{
		in->nDataChunks++;
	}
    
        
	tn->level0[chunkInInode & SFFS_TNODES_LEVEL0_MASK] = (chunkInNAND >> dev->chunkGroupBits);

	return SFFS_OK;
}



int sffs_ReadChunkDataFromObject(sffs_Object *in,int chunkInInode, __u8 *buffer)
{
    int chunkInNAND = sffs_FindChunkInFile(in,chunkInInode,NULL);

    
    if(chunkInNAND >= 0)
    {
		return sffs_ReadChunkFromNAND(in->myDev,chunkInNAND,buffer,NULL,1);
	}
	else
	{
		return 0;
	}

}

static void js_spare_dump(sffs_Spare *spare)
{
    //printk("-------spare dump-------------\n");
    //printk("blockStatus : %d\n",spare->blockStatus);
    //printk("ecc1 : %d\n",spare->ecc1);
    //printk("ecc2 : %d\n",spare->ecc2);
    //printk("pageStatus : %d\n",spare->pageStatus);
    //printk("tagByte0 : %d\n",spare->tagByte0);
    //printk("tagByte2 : %d\n",spare->tagByte2);
    //printk("tagByte3 : %d\n",spare->tagByte3);
    //printk("tagByte4 : %d\n",spare->tagByte4);
    //printk("tagByte5 : %d\n",spare->tagByte5);    
    //printk("tagByte6 : %d\n",spare->tagByte6);    
    //printk("tagByte7 : %d\n",spare->tagByte7);    
    //printk("-------spare dump end-------------\n");
    

 }

static void sffs_DeleteChunk(sffs_Device *dev,int chunkId,int markNAND)
{
	int block;
	int page;
	sffs_Spare spare;
	sffs_Spare temp_spare;    
	sffs_BlockInfo *bi;

    int cnt;
	//js
	__u8 data[512];
    memset(data,0,sizeof(data));
	
	
	if(chunkId <= 0) return;	
	
	dev->nDeletions++;
	block = chunkId / dev->nChunksPerBlock;
	page = chunkId % dev->nChunksPerBlock;
	
	if(markNAND)
	{
		sffs_SpareInitialise(&spare);
	
		spare.pageStatus = 0; // To mark it as deleted.

   		sffs_SpareInitialise(&temp_spare);        
        sffs_ReadChunkFromNAND(dev,chunkId,&data,&temp_spare,1);

       
		sffs_WriteChunkToNAND(dev,chunkId,NULL,&spare);
		sffs_HandleUpdateChunk(dev,chunkId,&spare);
        //js
   		sffs_SpareInitialise(&temp_spare);        
        sffs_ReadChunkFromNAND(dev,chunkId,NULL,&temp_spare,1);

        
	}
	else
	{
			dev->nUnmarkedDeletions++;
	}	
	
	bi = sffs_GetBlockInfo(dev,block);
			
	
	// Pull out of the management area.
	// If the whole block became dirty, this will kick off an erasure.
	if(	bi->blockState == SFFS_BLOCK_STATE_ALLOCATING ||
	    bi->blockState == SFFS_BLOCK_STATE_FULL)
	{
		dev->nFreeChunks++;

		sffs_ClearChunkBit(dev,block,page);
		bi->pagesInUse--;
		
		if(bi->pagesInUse == 0 &&
	       bi->blockState == SFFS_BLOCK_STATE_FULL)
	    {
	    	sffs_BlockBecameDirty(dev,block);
	    }

	}
	else
	{
	}
}




int sffs_WriteChunkDataToObject(sffs_Object *in,int chunkInInode, const __u8 *buffer,int nBytes,int useReserve)
{
	// Find old chunk Need to do this to get serial number
	// Write new one and patch into tree.
	// Invalidate old tags.

    int prevChunkId;
    sffs_Tags prevTags;
    
    int newChunkId;
    sffs_Tags newTags;
//    sffs_Tags tempTags;

    sffs_Device *dev = in->myDev;    
	//js
	__u8 *buf;
    int chunk,cnt;
//	sffs_Spare spare;
//	sffs_Tags  tags;
//    int chk,chunkDeleted;
    
    buf=YMALLOC(sizeof(__u8));


    
    if(g_nCreateflag==0)
    {

        for(cnt=0;cnt<nBytes;cnt++)
        {
            
            if(buffer[cnt]==buf[cnt])
                continue;
            else
                g_nTranFlag=1;
        }
    }    
    

      {
    	sffs_CheckGarbageCollection(dev);

    	// Get the previous chunk at this location in the file if it exists
        prevChunkId  = sffs_FindChunkInFile(in,chunkInInode,&prevTags);
        
        // Set up new tags
       	newTags.chunkId = chunkInInode;
    	newTags.objectId = in->objectId;
    	newTags.serialNumber = (prevChunkId >= 0) ? prevTags.serialNumber+1  : 1;
       	newTags.byteCount = nBytes;
    	newTags.newType = 0xff;

        
    	sffs_CalcTagsECC(&newTags);


    	newChunkId = sffs_WriteNewChunkWithTagsToNAND(dev,buffer,&newTags,useReserve);
        g_cache.previd=prevChunkId;

       
    	if(newChunkId >= 0)
    	{
    		sffs_PutChunkIntoFile(in,chunkInInode,newChunkId,0);
    		
    		if(prevChunkId >= 0)
    		{
                if(g_nTranFlag==0)
                {
        			sffs_DeleteChunk(dev,prevChunkId,1);                    
                }
                //g_nTranFlag==1̸  .
                else
                {
                }

    	
    		}
       		sffs_CheckFileSanity(in);
       		
    	}
    }

    g_nCacheCnt++;
	return newChunkId;





}


// UpdateObjectHeader updates the header on NAND for an object.
// If name is not NULL, then that new name is used.
//
int sffs_UpdateObjectHeader(sffs_Object *in,const char *name, int force)
{

//    int cnt;
    //js
    int m_nChunk;
	sffs_Device *dev = in->myDev;
	
    int prevChunkId;
    
    int newChunkId;
    sffs_Tags newTags;
    __u8 bufferNew[SFFS_BYTES_PER_CHUNK];
    __u8 bufferOld[SFFS_BYTES_PER_CHUNK];
    
    sffs_ObjectHeader *oh = (sffs_ObjectHeader *)bufferNew;
    sffs_ObjectHeader *ohOld = (sffs_ObjectHeader *)bufferOld;

   
    
    if(!in->fake || force)
    {


		sffs_CheckGarbageCollection(dev);		
    
		memset(bufferNew,0xFF,SFFS_BYTES_PER_CHUNK);
    
		prevChunkId = in->chunkId;

        
		if(prevChunkId >= 0)
		{
			sffs_ReadChunkFromNAND(dev,prevChunkId,bufferOld,NULL,1);    	
		}

		// Header data
		oh->type = in->variantType;
		
		oh->st_mode = in->st_mode;

#ifdef CONFIG_SFFS_WINCE
		oh->win_atime[0] = in->win_atime[0];
		oh->win_ctime[0] = in->win_ctime[0];
		oh->win_mtime[0] = in->win_mtime[0];
		oh->win_atime[1] = in->win_atime[1];
		oh->win_ctime[1] = in->win_ctime[1];
		oh->win_mtime[1] = in->win_mtime[1];
#else
		oh->st_uid = in->st_uid;
		oh->st_gid = in->st_gid;
		oh->st_atime = in->st_atime;
		oh->st_mtime = in->st_mtime;
		oh->st_ctime = in->st_ctime;
		oh->st_rdev = in->st_rdev;
        
        
#endif	
		if(in->parent)
		{
			oh->parentObjectId = in->parent->objectId;
		}
		else
		{
			oh->parentObjectId = 0;
		}
		
		//oh->sum = in->sum;
		if(name && *name)
		{
			memset(oh->name,0,SFFS_MAX_NAME_LENGTH + 1);
			strncpy(oh->name,name,SFFS_MAX_NAME_LENGTH);
		}
		else if(prevChunkId)
		{
			memcpy(oh->name, ohOld->name,SFFS_MAX_NAME_LENGTH + 1);
		}
		else
		{
			memset(oh->name,0,SFFS_MAX_NAME_LENGTH + 1);	
		}
	
		switch(in->variantType)
		{
			case SFFS_OBJECT_TYPE_UNKNOWN: 	
				// Should not happen
				break;
			case SFFS_OBJECT_TYPE_FILE:
				oh->fileSize = in->variant.fileVariant.fileSize;
				break;
			case SFFS_OBJECT_TYPE_HARDLINK:
				oh->equivalentObjectId = in->variant.hardLinkVariant.equivalentObjectId;
				break;
			case SFFS_OBJECT_TYPE_SPECIAL:	
				// Do nothing
				break;
			case SFFS_OBJECT_TYPE_DIRECTORY:	
				// Do nothing
				break;
			case SFFS_OBJECT_TYPE_SYMLINK:
				strncpy(oh->alias,in->variant.symLinkVariant.alias,SFFS_MAX_ALIAS_LENGTH);
				oh->alias[SFFS_MAX_ALIAS_LENGTH] = 0;
				break;
		}

		// Tags
		in->serial++;
		newTags.chunkId = 0;
		newTags.objectId = in->objectId;
		newTags.serialNumber = in->serial;
		newTags.byteCount =   0xFFFFFFFF;
		newTags.newType = 0xFFFFFFFF;

    
		sffs_CalcTagsECC(&newTags);
	

		// Create new chunk in NAND
		newChunkId = sffs_WriteNewChunkWithTagsToNAND(dev,bufferNew,&newTags, (prevChunkId >= 0) ? 1 : 0 );

        if(name==NULL)
        {
        }
        else
        {
            if(g_nCreateflag==1)
            {
                g_nFileCreate=newChunkId;
                sprintf(g_strCreate,"%s",name);
            }
        }
    
		if(newChunkId >= 0)
		{

			in->chunkId = newChunkId;		
		
			if(prevChunkId >= 0)
			{

				sffs_DeleteChunk(dev,prevChunkId,1);
			}
		
			in->dirty = 0;
		}
        
        if(g_nCrashStart==1 && name==NULL)
        {
            g_nCrashEnd=1;
            g_nCrashStart=0;
            sffs_TransactionEnd(dev);
//            do_gettimeofday(&g_WriteEndTime);        
 //           g_ltemp= (g_WriteEndTime.tv_sec * 1000000 + g_WriteEndTime.tv_usec) - (g_WriteSartTime.tv_sec * 1000000 + g_WriteSartTime.tv_usec);

        }

		return newChunkId;

    }
    return 0;
}


/////////////////////// Short Operations Cache ////////////////////////////////
//	In many siturations where there is no high level buffering (eg WinCE) a lot of
//	reads might be short sequential reads, and a lot of writes may be short 
//  sequential writes. eg. scanning/writing a jpeg file.
//	In these cases, a short read/write cache can provide a huge perfomance benefit 
//  with dumb-as-a-rock code.
//  There are a limited number (~10) of cache chunks per device so that we don't
//  need a very intelligent search.





static void sffs_FlushFilesChunkCache(sffs_Object *obj)
{
	sffs_Device *dev = obj->myDev;
	int lowest;
	int i;
	sffs_ChunkCache *cache;
	int chunkWritten;
	int nBytes;
	int nCaches = obj->myDev->nShortOpCaches;
					T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_FlushFilesChunkCache\n"));            	    		    								    		        	
	if  (nCaches > 0)
	{
		do{
			cache = NULL;
		
			// Find the dirty cache for this object with the lowest chunk id.
			for(i = 0; i < nCaches; i++)
			{
				if(dev->srCache[i].object == obj &&
				dev->srCache[i].dirty)
				{
					if(!cache ||  dev->srCache[i].chunkId < lowest)
					{
						cache = &dev->srCache[i];
						lowest = cache->chunkId;
					}
				}
			}
		
			if(cache)
			{
				//Write it out

#if 0
				nBytes = cache->object->variant.fileVariant.fileSize - ((cache->chunkId -1) * SFFS_BYTES_PER_CHUNK);
			
				if(nBytes > SFFS_BYTES_PER_CHUNK)
				{
					nBytes= SFFS_BYTES_PER_CHUNK;
				}
#endif			
				chunkWritten = sffs_WriteChunkDataToObject(cache->object,
															cache->chunkId,
															cache->data,
															cache->nBytes,1);

				cache->dirty = 0;
				cache->object = NULL;
			}
		
		} while(cache && chunkWritten > 0);
	
		if(cache)
		{
			//Hoosterman, disk full while writing cache out.
			T(SFFS_TRACE_ERROR, (TSTR("sffs tragedy: no space during cache write" TENDSTR)));

		}
	}	
		
}


// Grab us a chunk for use.
// First look for an empty one. 
// Then look for the least recently used non-dirty one.
// Then look for the least recently used dirty one...., flush and look again.
static sffs_ChunkCache *sffs_GrabChunkCacheWorker(sffs_Device *dev)
{
	int i;
	int usage;
	int theOne;
					T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_GrabChunkCacheWorker\n"));            	    		    								    		        		
	if(dev->nShortOpCaches > 0)
	{
		for(i = 0; i < dev->nShortOpCaches; i++)
		{
			if(!dev->srCache[i].object)
			{
				//T(("Grabbing empty %d\n",i));
				
				//printf("Grabbing empty %d\n",i);
			
				return &dev->srCache[i];
			}
		}
		
		return NULL;
	
		theOne = -1; 
		usage = 0; // just to stop the compiler grizzling
	
		for(i = 0; i < dev->nShortOpCaches; i++)
		{
			if(!dev->srCache[i].dirty &&
			((dev->srCache[i].lastUse < usage  && theOne >= 0)|| 
				theOne < 0))
			{
				usage = dev->srCache[i].lastUse;
				theOne = i;
			}
		}
	
		//T(("Grabbing non-empty %d\n",theOne));
		
		//if(theOne >= 0) printf("Grabbed non-empty cache %d\n",theOne);
		
		return  theOne >= 0 ?  &dev->srCache[theOne] : NULL;
	}
	else
	{
		return NULL;
	}
	
}


static sffs_ChunkCache *sffs_GrabChunkCache(sffs_Device *dev)
{
	sffs_ChunkCache *cache;
	sffs_Object *theObj;
	int usage;
	int i;
	int pushout;
	
	if(dev->nShortOpCaches > 0)
	{
		// Try find a non-dirty one...
	
		cache = sffs_GrabChunkCacheWorker(dev);
	
		if(!cache)
		{
			// They were all dirty, find the last recently used object and flush
			// its cache, then  find again.
			// NB what's here is not very accurate, we actually flush the object
			// the last recently used page.
		
			theObj = dev->srCache[0].object;
			usage = dev->srCache[0].lastUse;
			cache = &dev->srCache[0];
			pushout = 0;
	
			for(i = 1; i < dev->nShortOpCaches; i++)
			{
				if( dev->srCache[i].object && 
					dev->srCache[i].lastUse < usage)
				{
					usage  = dev->srCache[i].lastUse;
					theObj = dev->srCache[i].object;
					cache = &dev->srCache[i];
					pushout = i;
				}
			}
		
			if(!cache || cache->dirty)
			{
			
				sffs_FlushFilesChunkCache(theObj);
		
				// Try again
				cache = sffs_GrabChunkCacheWorker(dev);
			}
			else
			{
			}
			
		}

		return cache;
	}
	else
		return NULL;

}


// Find a cached chunk
static sffs_ChunkCache *sffs_FindChunkCache(const sffs_Object *obj, int chunkId)
{
	sffs_Device *dev = obj->myDev;
	int i;
	if(dev->nShortOpCaches > 0)
	{
		for(i = 0; i < dev->nShortOpCaches; i++)
		{
			if(dev->srCache[i].object == obj && 
			dev->srCache[i].chunkId == chunkId)
			{
				dev->cacheHits++;
			
				return &dev->srCache[i];
			}
		}
	}
	return NULL;
}

// Mark the chunk for the least recently used algorithym
static void sffs_UseChunkCache(sffs_Device *dev, sffs_ChunkCache *cache, int isAWrite)
{

	if(dev->nShortOpCaches > 0)
	{
		if( dev->srLastUse < 0 || 
			dev->srLastUse > 100000000)
		{
			// Reset the cache usages
			int i;
			for(i = 1; i < dev->nShortOpCaches; i++)
			{
				dev->srCache[i].lastUse = 0;
			}
			dev->srLastUse = 0;
		}

		dev->srLastUse++;
	
		cache->lastUse = dev->srLastUse;

		if(isAWrite)
		{
			cache->dirty = 1;
		}
	}
}

// Invalidate a single cache page.
// Do this when a whole page gets written,
// ie the short cache for this page is no longer valid.
static void sffs_InvalidateChunkCache(sffs_Object *object, int chunkId)
{
	if(object->myDev->nShortOpCaches > 0)
	{
		sffs_ChunkCache *cache = sffs_FindChunkCache(object,chunkId);

		if(cache)
		{
			cache->object = NULL;
		}
	}
}


// Invalidate all the cache pages associated with this object
// Do this whenever ther file is deleted or resized.
static void sffs_InvalidateWholeChunkCache(sffs_Object *in)
{
	int i;
	sffs_Device *dev = in->myDev;
	
	if(dev->nShortOpCaches > 0)
	{ 
		// Now invalidate it.
		for(i = 0; i < dev->nShortOpCaches; i++)
		{
			if(dev->srCache[i].object == in)
			{
				dev->srCache[i].object = NULL;
			}
		}
	}
}





///////////////////////// File read/write ///////////////////////////////
// Read and write have very similar structures.
// In general the read/write has three parts to it
// * An incomplete chunk to start with (if the read/write is not chunk-aligned)
// * Some complete chunks
// * An incomplete chunk to end off with
//
// Curve-balls: the first chunk might also be the last chunk.

int sffs_ReadDataFromFile(sffs_Object *in, __u8 * buffer, __u32 offset, int nBytes)
{
	
	
	int chunk;
	int start;
	int nToCopy;
	int n = nBytes;
	int nDone = 0;
    __u8 strTemp[513];
    int size;
	sffs_ChunkCache *cache;
	
	sffs_Device *dev;

    size=in->variant.fileVariant.fileSize;        
    
	dev = in->myDev;

	while(n > 0)
	{
		chunk = offset / SFFS_BYTES_PER_CHUNK + 1; // The first chunk is 1
		start = offset % SFFS_BYTES_PER_CHUNK;

		// OK now check for the curveball where the start and end are in
		// the same chunk.	
		if(	(start + n) < SFFS_BYTES_PER_CHUNK)
		{
			nToCopy = n;
		}
		else
		{
			nToCopy = SFFS_BYTES_PER_CHUNK - start;
		}
	
		cache = sffs_FindChunkCache(in,chunk);

        if(cache!=NULL)
        {
            strncpy(strTemp,cache->data,SFFS_BYTES_PER_CHUNK);
            strTemp[SFFS_BYTES_PER_CHUNK+1]='\0';
        }
        else
         {
        }
        
		
		// If the chunk is already in the cache or it is less than a whole chunk
		// then use the cache (if there is caching)
		// else bypass the cache.
		if( cache || nToCopy != SFFS_BYTES_PER_CHUNK)
		{
			if(dev->nShortOpCaches > 0)
			{
				
				// If we can't find the data in the cache, then load it up.
				
				if(!cache)
				{
					cache = sffs_GrabChunkCache(in->myDev);
					cache->object = in;
					cache->chunkId = chunk;
					cache->dirty = 0;
					sffs_ReadChunkDataFromObject(in,chunk,cache->data);
					cache->nBytes = 0;	
				}
			
				sffs_UseChunkCache(dev,cache,0);

				memcpy(buffer,&cache->data[start],nToCopy);
			}
			else
			{
				// Read into the local buffer then copy...
				sffs_ReadChunkDataFromObject(in,chunk,dev->localBuffer);		
				memcpy(buffer,&dev->localBuffer[start],nToCopy);
			}

		}
		else
		{
#ifdef CONFIG_SFFS_WINCE
			
			// Under WinCE can't do direct transfer. Need to use a local buffer.
			// This is because we otherwise screw up WinCE's memory mapper
			sffs_ReadChunkDataFromObject(in,chunk,dev->localBuffer);
			memcpy(buffer,dev->localBuffer,SFFS_BYTES_PER_CHUNK);
#else
			// A full chunk. Read directly into the supplied buffer.
			sffs_ReadChunkDataFromObject(in,chunk,buffer);
#endif
		}
		
		n -= nToCopy;
		offset += nToCopy;
		buffer += nToCopy;
		nDone += nToCopy;
	}
	
	return nDone;
}
static int sffs_ChunkCacheCp(sffs_ChunkCache *src, sffs_ChunkCache *dest)
{
    dest->chunkId=src->chunkId;
    memcpy(dest->data,src->data,sizeof(src->data));
    dest->dirty=src->dirty;
    dest->lastUse=src->lastUse;
    dest->nBytes=src->nBytes;
    dest->object=src->object;
}


//  writeϴ κ.
int sffs_WriteDataToFile(sffs_Object *in,const __u8 * buffer, __u32 offset, int nBytes)
{	
	
	int chunk;
	int start;
	int nToCopy;
	int n = nBytes;
	int nDone = 0;
	int nToWriteBack;
	int startOfWrite = offset;
	int chunkWritten = 0;
	int nBytesRead;
	
	sffs_Device *dev;

    //js
    int chk,i;
    sffs_Tnode *tn;

    tn=YMALLOC(sizeof(sffs_Tnode));
	dev = in->myDev;
	
	T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_WriteDataToFile\n"));
	while(n > 0 && chunkWritten >= 0)
	{
		chunk = offset / SFFS_BYTES_PER_CHUNK + 1;
		start = offset % SFFS_BYTES_PER_CHUNK;

        
		if(	(start + n) < SFFS_BYTES_PER_CHUNK)
		{
			nToCopy = n;
			
			// Now folks, to calculate how many bytes to write back....
			// If we're overwriting and not writing to then end of file then
			// we need to write back as much as was there before.
			nBytesRead = in->variant.fileVariant.fileSize - ((chunk -1) * SFFS_BYTES_PER_CHUNK);
			
			if(nBytesRead > SFFS_BYTES_PER_CHUNK)
			{
				nBytesRead = SFFS_BYTES_PER_CHUNK;
			}
			
			nToWriteBack = (nBytesRead > (start + n)) ? nBytesRead : (start +n);
		}
		else
		{
			nToCopy = SFFS_BYTES_PER_CHUNK - start;
			nToWriteBack = SFFS_BYTES_PER_CHUNK;
		}

		if(nToCopy != SFFS_BYTES_PER_CHUNK)
		{
            
			// An incomplete start or end chunk (or maybe both start and end chunk)
			if(dev->nShortOpCaches > 0)
			{
				sffs_ChunkCache *cache;
				// If we can't find the data in the cache, then load it up.
				cache = sffs_FindChunkCache(in,chunk);
				if(!cache && sffs_CheckSpaceForChunkCache(in->myDev))
				{
					cache = sffs_GrabChunkCache(in->myDev);
					cache->object = in;
					cache->chunkId = chunk;
					cache->dirty = 0;
					sffs_ReadChunkDataFromObject(in,chunk,cache->data);


				}
				if(cache)
				{	
					sffs_UseChunkCache(dev,cache,1);
					memcpy(&cache->data[start],buffer,nToCopy);
					cache->nBytes = nToWriteBack;


				}
				else
				{
					chunkWritten = -1; // fail the write
				}
			}
			else
			{
				// An incomplete start or end chunk (or maybe both start and end chunk)
				// Read into the local buffer then copy, then copy over and write back.
                //printk(KERN_INFO"else !! nToCopy != SFFS_BYTES_PER_CHUNK\n");
				sffs_ReadChunkDataFromObject(in,chunk,dev->localBuffer);
				
				memcpy(&dev->localBuffer[start],buffer,nToCopy);

                //js!!!!
				chunkWritten = sffs_WriteChunkDataToObject(in,chunk,dev->localBuffer,nToWriteBack,0);
			
			}
			
		}
		else
		{
			
#ifdef CONFIG_SFFS_WINCE
			// Under WinCE can't do direct transfer. Need to use a local buffer.
			// This is because we otherwise screw up WinCE's memory mapper
			memcpy(dev->localBuffer,buffer,SFFS_BYTES_PER_CHUNK);
			chunkWritten = sffs_WriteChunkDataToObject(in,chunk,dev->localBuffer,SFFS_BYTES_PER_CHUNK,0);
#else
			// A full chunk. Write directly from the supplied buffer.
			chunkWritten = sffs_WriteChunkDataToObject(in,chunk,buffer,SFFS_BYTES_PER_CHUNK,0);
#endif
			// Since we've overwritten the cached data, we better invalidate it.
			//printk("FUNC BFORE sffs_InvalidateChunkCache\n");
			sffs_InvalidateChunkCache(in,chunk);
		}

		if(chunkWritten >= 0)
		{
			n -= nToCopy;
			offset += nToCopy;
			buffer += nToCopy;
			nDone += nToCopy;
		}
		
	}
	
	// Update file object

	if((startOfWrite + nDone) > in->variant.fileVariant.fileSize)
	{
		in->variant.fileVariant.fileSize = (startOfWrite + nDone);
	}
	
	in->dirty = 1;

    if(in->variant.fileVariant.fileSize%512==0)       
        g_cache.idInChunk=in->variant.fileVariant.fileSize/512+g_nCacheCnt;
    else
        g_cache.idInChunk=in->variant.fileVariant.fileSize/512+1+g_nCacheCnt;
    
    g_nCacheCnt++;
    T(SFFS_TRACE_OS,(KERN_DEBUG"------------end sffs_WriteDataToFile-----------\n"));			
	return nDone;
}


int sffs_ResizeFile(sffs_Object *in, int newSize)
{
	int i;
	int chunkId;
	int oldFileSize = in->variant.fileVariant.fileSize;
	int sizeOfPartialChunk = newSize % SFFS_BYTES_PER_CHUNK;
	
	sffs_Device *dev = in->myDev;
	
	sffs_FlushFilesChunkCache(in);	
	sffs_InvalidateWholeChunkCache(in);
	
	if(in->variantType != SFFS_OBJECT_TYPE_FILE)
	{
		return sffs_GetFileSize(in);
	}
	
	if(newSize < oldFileSize)
	{
		
		int lastDel = 1 + (oldFileSize-1)/SFFS_BYTES_PER_CHUNK;
		
		int startDel = 1 + (newSize + SFFS_BYTES_PER_CHUNK - 1)/
							SFFS_BYTES_PER_CHUNK;

		// Delete backwards so that we don't end up with holes if
		// power is lost part-way through the operation.
		for(i = lastDel; i >= startDel; i--)
		{
			// NB this could be optimised somewhat,
			// eg. could retrieve the tags and write them without
			// using sffs_DeleteChunk

			chunkId = sffs_FindAndDeleteChunkInFile(in,i,NULL);
			if(chunkId < (dev->startBlock * 32) || chunkId >= ((dev->endBlock+1) * 32))
			{
				//T(("Found daft chunkId %d for %d\n",chunkId,i));
			}
			else
			{
				//in->nDataChunks--;
                //js 0626
				//sffs_DeleteChunk(dev,chunkId,1);
			}
		}
		
		
		if(sizeOfPartialChunk != 0)
		{
			int lastChunk = 1+ newSize/SFFS_BYTES_PER_CHUNK;
			
			// Got to read and rewrite the last chunk with its new size.
			sffs_ReadChunkDataFromObject(in,lastChunk,dev->localBuffer);
			
			sffs_WriteChunkDataToObject(in,lastChunk,dev->localBuffer,sizeOfPartialChunk,1);
				
		}
		
		in->variant.fileVariant.fileSize = newSize;
		
		sffs_PruneFileStructure(dev,&in->variant.fileVariant);
		
		return newSize;
		
	}
	else
	{
		return oldFileSize;
	}
}


loff_t sffs_GetFileSize(sffs_Object *obj)
{
	obj = sffs_GetEquivalentObject(obj);
	
	switch(obj->variantType)
	{
		case SFFS_OBJECT_TYPE_FILE: 
			return obj->variant.fileVariant.fileSize;
		case SFFS_OBJECT_TYPE_SYMLINK:
			return strlen(obj->variant.symLinkVariant.alias);
		default:
			return 0;
	}
}



// sffs_FlushFile() updates the file's
// objectId in NAND

int sffs_FlushFile(sffs_Object *in, int updateTime)
{
	int retVal;
    
    //printk("sffs_FlushFile\n");
    
	if(in->dirty)
	{
        //printk("sffs_FlushFile : DIRTY\n");        
        T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_FlushFile\n"));
			
		//T(("flushing object header\n"));
		
		sffs_FlushFilesChunkCache(in);
		if(updateTime)
		{
#ifdef CONFIG_SFFS_WINCE
			yfsd_WinFileTimeNow(in->win_mtime);
#else
			in->st_mtime = CURRENT_TIME;
#endif
		}

		retVal = (sffs_UpdateObjectHeader(in,NULL,0) >= 0)? SFFS_OK : SFFS_FAIL;
	}
	else
	{
		retVal = SFFS_OK;
	}
	return retVal;
	
}


static int sffs_DoGenericObjectDeletion(sffs_Object *in)
{

	// First off, invalidate the file's data in the cache, without flushing.
	sffs_InvalidateWholeChunkCache(in);
	
	sffs_RemoveObjectFromDirectory(in);
	sffs_DeleteChunk(in->myDev,in->chunkId,1);
#ifdef __KERNEL__
	if(in->myInode)
	{
		in->myInode->u.generic_ip = NULL;
		in->myInode = 0;
	}
#endif
	sffs_FreeObject(in);
	return SFFS_OK;

}

// sffs_DeleteFile deletes the whole file data
// and the inode associated with the file.
// It does not delete the links associated with the file.
static int sffs_UnlinkFile(sffs_Object *in)
{

#ifdef CONFIG_SFFS_DISABLE_BACKGROUND_DELETION

	// Delete the file data & tnodes

	 sffs_DeleteWorker(in, in->variant.fileVariant.top, in->variant.fileVariant.topLevel, 0,NULL);
	 

	sffs_FreeTnode(in->myDev,in->variant.fileVariant.top);
	
	return  sffs_DoGenericObjectDeletion(in);
#else
	int retVal;
	int immediateDeletion=0;
	retVal = sffs_ChangeObjectName(in, in->myDev->unlinkedDir,NULL,0);
	if(1 || // Ignore the result of the change name. This will only fail
			// if the disk was completely full (an error condition)
			// We ignore the error so we can delete files to recover the disk
	   retVal == SFFS_OK)
	{
		//in->unlinked = 1;
		//in->myDev->nUnlinkedFiles++;
		//in->renameAllowed = 0;
#ifdef __KERNEL__
		if(!in->myInode)
		{
			immediateDeletion = 1;

		}
#else
		if(in->inUse <= 0)
		{
			immediateDeletion = 1;

		}
#endif
		
		if(immediateDeletion)
		{
			T(SFFS_TRACE_TRACING,(TSTR("sffs: immediate deletion of file %d" TENDSTR),in->objectId));
			in->deleted=1;
			in->myDev->nDeletedFiles++;
			sffs_SoftDeleteFile(in);
		}
	
	}
	return retVal;

	
#endif
}

int sffs_DeleteFile(sffs_Object *in)
{
	int retVal = SFFS_OK;

	
	if(in->nDataChunks > 0)
	{
		// Use soft deletion
		if(!in->unlinked)
		{
			retVal = sffs_UnlinkFile(in);
		}
		if(retVal == SFFS_OK && 
		in->unlinked &&
		!in->deleted)
		{
			in->deleted = 1;
			in->myDev->nDeletedFiles++;
			sffs_SoftDeleteFile(in);
		}
		return in->deleted ? SFFS_OK : SFFS_FAIL;	
	}
	else
	{
		// The file has no data chunks so we toss it immediately
		sffs_FreeTnode(in->myDev,in->variant.fileVariant.top);
		in->variant.fileVariant.top = NULL;
		sffs_DoGenericObjectDeletion(in);	
		
		return SFFS_OK;	
	}
}

static int sffs_DeleteDirectory(sffs_Object *in)
{
	//First check that the directory is empty.

	if(list_empty(&in->variant.directoryVariant.children))
	{
		return  sffs_DoGenericObjectDeletion(in);
	}
	
	return SFFS_FAIL;
	
}

static int sffs_DeleteSymLink(sffs_Object *in)
{
	YFREE(in->variant.symLinkVariant.alias);

	return  sffs_DoGenericObjectDeletion(in);
}

static int sffs_DeleteHardLink(sffs_Object *in)
{
	// remove this hardlink from the list assocaited with the equivalent
	// object
	list_del(&in->hardLinks);
	return  sffs_DoGenericObjectDeletion(in);	
}


static void sffs_AbortHalfCreatedObject(sffs_Object *obj)
{
	switch(obj->variantType)
	{
		case SFFS_OBJECT_TYPE_FILE: sffs_DeleteFile(obj); break;
		case SFFS_OBJECT_TYPE_DIRECTORY: sffs_DeleteDirectory(obj); break;
		case SFFS_OBJECT_TYPE_SYMLINK: sffs_DeleteSymLink(obj); break;
		case SFFS_OBJECT_TYPE_HARDLINK: sffs_DeleteHardLink(obj); break;
		case SFFS_OBJECT_TYPE_SPECIAL: sffs_DoGenericObjectDeletion(obj); break;
		case SFFS_OBJECT_TYPE_UNKNOWN: break; // should not happen.
	}
}


static int sffs_UnlinkWorker(sffs_Object *obj)
{

	
	if(obj->variantType == SFFS_OBJECT_TYPE_HARDLINK)
	{
		return  sffs_DeleteHardLink(obj);
	}
	else if(!list_empty(&obj->hardLinks))
	{	
		// Curve ball: We're unlinking an object that has a hardlink.
		//
		//	This problem arises because we are not strictly following
		//  The Linux link/inode model.
		//
		// We can't really delete the object.
		// Instead, we do the following:
		// - Select a hardlink.
		// - Unhook it from the hard links
		// - Unhook it from its parent directory (so that the rename can work)
		// - Rename the object to the hardlink's name.
		// - Delete the hardlink
		
		
		sffs_Object *hl;
		int retVal;
		char name[SFFS_MAX_NAME_LENGTH+1];
		
		hl = list_entry(obj->hardLinks.next,sffs_Object,hardLinks);
		
		list_del_init(&hl->hardLinks);
		list_del_init(&hl->siblings);
		
		sffs_GetObjectName(hl,name,SFFS_MAX_NAME_LENGTH+1);
		
		retVal = sffs_ChangeObjectName(obj, hl->parent, name,0);
		
		if(retVal == SFFS_OK)
		{
			retVal = sffs_DoGenericObjectDeletion(hl);
		}
		return retVal;
				
	}
	else
	{
		switch(obj->variantType)
		{
			case SFFS_OBJECT_TYPE_FILE:
				return sffs_UnlinkFile(obj);
				break;
			case SFFS_OBJECT_TYPE_DIRECTORY:
				return sffs_DeleteDirectory(obj);
				break;
			case SFFS_OBJECT_TYPE_SYMLINK:
				return sffs_DeleteSymLink(obj);
				break;
			case SFFS_OBJECT_TYPE_HARDLINK:
			case SFFS_OBJECT_TYPE_UNKNOWN:
			default:
				return SFFS_FAIL;
		}
	}
}

int sffs_Unlink(sffs_Object *dir, const char *name)
{
	sffs_Object *obj;
	
	 obj = sffs_FindObjectByName(dir,name);
	 
	 if(obj && obj->unlinkAllowed)
	 {
	 	return sffs_UnlinkWorker(obj);
	 }
	 
	 return SFFS_FAIL;
	
}

//////////////// Initialisation Scanning /////////////////



// For now we use the SmartMedia check.
// We look at the blockStatus byte in the first two chunks
// These must be 0xFF to pass as OK.
// todo: this function needs to be modifyable foir different NAND types
// and different chunk sizes.  Suggest make this into a per-device configurable
// function.
static int sffs_IsBlockBad(sffs_Device *dev, int blk)
{
	sffs_Spare spare;
	
	sffs_ReadChunkFromNAND(dev,blk * dev->nChunksPerBlock,NULL,&spare,1);
#if 1
	if(sffs_CountBits(spare.blockStatus) < 7)
	{
		return 1;
	}
#else
	if(spare.blockStatus != 0xFF)
	{
		return 1;
	}
#endif
	sffs_ReadChunkFromNAND(dev,blk * dev->nChunksPerBlock + 1,NULL,&spare,1);

#if 1
	if(sffs_CountBits(spare.blockStatus) < 7)
	{
		return 1;
	}
#else
	if(spare.blockStatus != 0xFF)
	{
		return 1;
	}
#endif
	
	return 0;
	
}

static int sffs_Scan(sffs_Device *dev)
{
	sffs_Spare spare;
	sffs_Tags tags;
	int blk;
	int chunk;
	int c;
	int deleted;
	sffs_BlockState state;
	sffs_Object *hardList = NULL;
	sffs_Object *hl;
	sffs_BlockInfo *bi;

	
	sffs_ObjectHeader *oh;
	sffs_Object *in;
	sffs_Object *parent;

    int nTemp;
	
	__u8 chunkData[SFFS_BYTES_PER_CHUNK];
					T(SFFS_TRACE_OS,(KERN_DEBUG"sffs_Scan\n"));            	    		    								    		        				
	
	// Scan all the blocks...
	
	for(blk = dev->m_pBlockManager->bmStartData; blk <= dev->m_pBlockManager->bmEndData; blk++)
	{
		deleted = 0;
		bi = sffs_GetBlockInfo(dev,blk);
		sffs_ClearChunkBits(dev,blk);
		bi->pagesInUse = 0;
		bi->softDeletions = 0;
		state = SFFS_BLOCK_STATE_SCANNING;
		
		
		if(sffs_IsBlockBad(dev,blk))
		{
			state = SFFS_BLOCK_STATE_DEAD;
			T(SFFS_TRACE_BAD_BLOCKS,(TSTR("block %d is bad" TENDSTR),blk));
		}
		
		// Read each chunk in the block.

        // ûũĵ.
		for(c = 0; c < dev->nChunksPerBlock && 
				   state == SFFS_BLOCK_STATE_SCANNING; c++)
		{
			// Read the spare area and decide what to do
			chunk = blk * dev->nChunksPerBlock + c;

            // ĵ
			sffs_ReadChunkFromNAND(dev,chunk,NULL,&spare,1);

			
			// This block looks ok, now what's in this chunk?
			sffs_GetTagsFromSpare(dev,&spare,&tags);


            nTemp=tags.chunkId;
            if(nTemp==0)
            {
            }

      
			if(sffs_CountBits(spare.pageStatus) < 6)
			{
				// A deleted chunk
				deleted++;
				dev->nFreeChunks ++;
			}

            // ʴ chunk
			else if(tags.objectId == SFFS_UNUSED_OBJECT_ID)
			{
				// An unassigned chunk in the block
				// This means that either the block is empty or 
				// this is the one being allocated from
				
				if(c == 0)
				{
                    // ù ûũ.

					// the block is unused
					state = SFFS_BLOCK_STATE_EMPTY;
					dev->nErasedBlocks++;
				}
				else
				{
					
					T(SFFS_TRACE_SCAN,(TSTR(" Allocating from %d %d" TENDSTR),blk,c));
					state = SFFS_BLOCK_STATE_ALLOCATING;
                    
					dev->allocationBlock = blk;
					dev->allocationPage = c;
				}

				dev->nFreeChunks += (dev->nChunksPerBlock - c);
			}
			else if(tags.chunkId > 0)
			{
				int endpos;
				// A data chunk.
				sffs_SetChunkBit(dev,blk,c);
                //  ûũ Ǵ.
				bi->pagesInUse++;

                //ûũ find  ûũ Ͻ 
				in = sffs_FindOrCreateObjectByNumber(dev,tags.objectId,SFFS_OBJECT_TYPE_FILE);
				// PutChunkIntoFIle checks for a clash (two data chunks with
				// the same chunkId).
				// ûũ ϴ üũ. 
				
				sffs_PutChunkIntoFile(in,tags.chunkId,chunk,1);
                
				endpos = (tags.chunkId - 1)* SFFS_BYTES_PER_CHUNK + tags.byteCount;
                
				if(in->variant.fileVariant.scannedFileSize <endpos)
				{
					in->variant.fileVariant.scannedFileSize = endpos;
#ifndef CONFIG_SFFS_USE_HEADER_FILE_SIZE
						in->variant.fileVariant.fileSize = 	
							in->variant.fileVariant.scannedFileSize;
#endif

				}
			}
			else
			{
				// chunkId == 0, so it is an ObjectHeader.
				// Thus, we read in the object header and make the object

//                in = sffs_FindOrCreateObjectByNumber(dev,tags.objectId,SFFS_OBJECT_TYPE_FILE);
                if(g_nIsCrash==2)
                {

                    
                    if(strstr(g_strCreateMount,oh->name)!=NULL)
                    {
                    }
                }
                    
				sffs_SetChunkBit(dev,blk,c);
				bi->pagesInUse++;
							
				sffs_ReadChunkFromNAND(dev,chunk,chunkData,NULL,1);
				
				oh = (sffs_ObjectHeader *)chunkData;

                
				in = sffs_FindOrCreateObjectByNumber(dev,tags.objectId,oh->type);

                //Ͻý ö󰥽.
				if(in->valid)
				{
					// We have already filled this one. We have a duplicate and need to resolve it.
					
					unsigned existingSerial = in->serial;
					unsigned newSerial = tags.serialNumber;
					
					if(((existingSerial+1) & 3) == newSerial)
					{
						// Use new one - destroy the exisiting one
						//Ż.
						sffs_DeleteChunk(dev,in->chunkId,1);
						in->valid = 0;
					}
					else
					{
						// Use existing - destroy this one.
						sffs_DeleteChunk(dev,chunk,1);
					}
				}
				
				if(!in->valid &&
				   (tags.objectId == SFFS_OBJECTID_ROOT ||
				    tags.objectId == SFFS_OBJECTID_LOSTNFOUND))
				{
					// We only load some info, don't fiddle with directory structure
					in->valid = 1;
					in->variantType = oh->type;
	
					in->st_mode  = oh->st_mode;
#ifdef CONFIG_SFFS_WINCE
					in->win_atime[0] = oh->win_atime[0];
					in->win_ctime[0] = oh->win_ctime[0];
					in->win_mtime[0] = oh->win_mtime[0];
					in->win_atime[1] = oh->win_atime[1];
					in->win_ctime[1] = oh->win_ctime[1];
					in->win_mtime[1] = oh->win_mtime[1];
#else
					in->st_uid   = oh->st_uid;
					in->st_gid   = oh->st_gid;
					in->st_atime = oh->st_atime;
					in->st_mtime = oh->st_mtime;
					in->st_ctime = oh->st_ctime;
					in->st_rdev = oh->st_rdev;
#endif
					in->chunkId  = chunk;

				}
				else if(!in->valid)
				{
					// we need to load this info

//int sffs_ReadDataFromFile(sffs_Object *in, __u8 * buffer, __u32 offset, int nBytes)

					in->valid = 1;
					in->variantType = oh->type;
	
					in->st_mode  = oh->st_mode;
#ifdef CONFIG_SFFS_WINCE
					in->win_atime[0] = oh->win_atime[0];
					in->win_ctime[0] = oh->win_ctime[0];
					in->win_mtime[0] = oh->win_mtime[0];
					in->win_atime[1] = oh->win_atime[1];
					in->win_ctime[1] = oh->win_ctime[1];
					in->win_mtime[1] = oh->win_mtime[1];
#else
					in->st_uid   = oh->st_uid;
					in->st_gid   = oh->st_gid;
					in->st_atime = oh->st_atime;
					in->st_mtime = oh->st_mtime;
					in->st_ctime = oh->st_ctime;
					in->st_rdev = oh->st_rdev;
#endif
					in->chunkId  = chunk;

					sffs_SetObjectName(in,oh->name);
					in->dirty = 0;
							
					// directory stuff...
					// hook up to parent
	
					parent = sffs_FindOrCreateObjectByNumber(dev,oh->parentObjectId,SFFS_OBJECT_TYPE_DIRECTORY);
					if(parent->variantType == SFFS_OBJECT_TYPE_UNKNOWN)
					{
						// Set up as a directory
						parent->variantType = SFFS_OBJECT_TYPE_DIRECTORY;
						INIT_LIST_HEAD(&parent->variant.directoryVariant.children);
					}
					else if(parent->variantType != SFFS_OBJECT_TYPE_DIRECTORY)
					{
						// Hoosterman, another problem....
						// We're trying to use a non-directory as a directory
						// Todo ... handle
						T(SFFS_TRACE_ERROR, (TSTR("sffs tragedy: attempting to use non-directory as a directory in scan" TENDSTR)));

					}
				
					sffs_AddObjectToDirectory(parent,in);
					if(parent == dev->unlinkedDir)
					{
						in->deleted = 1; // If it is unlinked at start up then it wants deleting
						dev->nDeletedFiles++;
					}
				
					// Note re hardlinks.
					// Since we might scan a hardlink before its equivalent object is scanned
					// we put them all in a list.
					// After scanning is complete, we should have all the objects, so we run through this
					// list and fix up all the chains.		
	
					switch(in->variantType)
					{
						case SFFS_OBJECT_TYPE_UNKNOWN: 	// Todo got a problem
							break;
						case SFFS_OBJECT_TYPE_FILE:
#ifdef CONFIG_SFFS_USE_HEADER_FILE_SIZE
							in->variant.fileVariant.fileSize = oh->fileSize;
#endif
							break;
						case SFFS_OBJECT_TYPE_HARDLINK:
							in->variant.hardLinkVariant.equivalentObjectId = oh->equivalentObjectId;
							in->hardLinks.next = (struct list_head *)hardList;
							hardList = in;
							break;
						case SFFS_OBJECT_TYPE_DIRECTORY:	// Do nothing
							break;
						case SFFS_OBJECT_TYPE_SPECIAL:	// Do nothing
							break;
						case SFFS_OBJECT_TYPE_SYMLINK: 	// Do nothing
							in->variant.symLinkVariant.alias = sffs_CloneString(oh->alias);
							break;
					}
				}
			}
		}

        
		if(state == SFFS_BLOCK_STATE_SCANNING)
		{
			// If we got this far while scanning, then the block is fully allocated.
			state = SFFS_BLOCK_STATE_FULL;	
		}
		
		bi->blockState = state;
		
		// Now let's see if it was dirty
		if(	bi->pagesInUse == 0 &&
	        bi->blockState == SFFS_BLOCK_STATE_FULL)
	    {
	    	sffs_BlockBecameDirty(dev,blk);
	    }

	}
	
	// Fix up the hard link chains.
	// We should now have scanned all the objects, now it's time to add these 
	// hardlinks.
	while(hardList)
	{
		hl = hardList;
		hardList = (sffs_Object *)(hardList->hardLinks.next);
		in = sffs_FindObjectByNumber(dev,hl->variant.hardLinkVariant.equivalentObjectId);
		
		if(in)
		{
			// Add the hardlink pointers
			hl->variant.hardLinkVariant.equivalentObject=in;
			list_add(&hl->hardLinks,&in->hardLinks);
		}
		else
		{
			//Todo Need to report/handle this better.
			// Got a problem... hardlink to a non-existant object
			hl->variant.hardLinkVariant.equivalentObject=NULL;
			INIT_LIST_HEAD(&hl->hardLinks);
			
		}
		
	}
	
	{
		struct list_head *i;	
		struct list_head *n;
			
		sffs_Object *l;
		// Soft delete all the unlinked files
		list_for_each_safe(i,n,&dev->unlinkedDir->variant.directoryVariant.children)
		{
			if(i)
			{
				l = list_entry(i, sffs_Object,siblings);
				if(l->deleted)
				{
					sffs_SoftDeleteFile(l);		
				}
			}
		}	
	}
	
	
	return SFFS_OK;
}


////////////////////////// Directory Functions /////////////////////////


static void sffs_AddObjectToDirectory(sffs_Object *directory, sffs_Object *obj)
{

	if(obj->siblings.prev == NULL)
	{
		// Not initialised
		INIT_LIST_HEAD(&obj->siblings);
		
	}
	else if(!list_empty(&obj->siblings))
	{
		// If it is holed up somewhere else, un hook it
		list_del_init(&obj->siblings);
	}
	// Now add it
	list_add(&obj->siblings,&directory->variant.directoryVariant.children);
	obj->parent = directory;
	
	if(directory == obj->myDev->unlinkedDir)
	{
		obj->unlinked = 1;
		obj->myDev->nUnlinkedFiles++;
		obj->renameAllowed = 0;
	}
}

static void sffs_RemoveObjectFromDirectory(sffs_Object *obj)
{
	list_del_init(&obj->siblings);
	obj->parent = NULL;
}

sffs_Object *sffs_FindObjectByName(sffs_Object *directory,const char *name)
{
	int sum;
	
	struct list_head *i;
	char buffer[SFFS_MAX_NAME_LENGTH+1];
	
	sffs_Object *l;
	
	sum = sffs_CalcNameSum(name);
	
	list_for_each(i,&directory->variant.directoryVariant.children)
	{
		if(i)
		{
			l = list_entry(i, sffs_Object,siblings);
		
			// Special case for lost-n-found
			if(l->objectId == SFFS_OBJECTID_LOSTNFOUND)
			{
				if(sffs_strcmp(name,SFFS_LOSTNFOUND_NAME) == 0)
				{
					return l;
				}
			}
			else if(sffs_SumCompare(l->sum, sum))
			{
				// Do a real check
				sffs_GetObjectName(l,buffer,SFFS_MAX_NAME_LENGTH);
				if(sffs_strcmp(name,buffer) == 0)
				{
					return l;
				}
			
			}			
		}
	}
	
	return NULL;
}


int sffs_ApplyToDirectoryChildren(sffs_Object *theDir,int (*fn)(sffs_Object *))
{
	struct list_head *i;	
	sffs_Object *l;
	
	
	list_for_each(i,&theDir->variant.directoryVariant.children)
	{
		if(i)
		{
			l = list_entry(i, sffs_Object,siblings);
			if(l && !fn(l))
			{
				return SFFS_FAIL;
			}
		}
	}
	
	return SFFS_OK;

}


// GetEquivalentObject dereferences any hard links to get to the
// actual object.

sffs_Object *sffs_GetEquivalentObject(sffs_Object *obj)
{
	if(obj && obj->variantType == SFFS_OBJECT_TYPE_HARDLINK)
	{
		// We want the object id of the equivalent object, not this one
		obj = obj->variant.hardLinkVariant.equivalentObject;
	}
	return obj;

}

int sffs_GetObjectName(sffs_Object *obj,char *name,int buffSize)
{
	memset(name,0,buffSize);


	if(obj->objectId == SFFS_OBJECTID_LOSTNFOUND)
	{
		strncpy(name,SFFS_LOSTNFOUND_NAME,buffSize - 1);
	}
	else if(obj->chunkId <= 0)
	{
		char locName[20];
		// make up a name
		sprintf(locName,"%s%d",SFFS_LOSTNFOUND_PREFIX,obj->objectId);
		strncpy(name,locName,buffSize - 1);

	}
#ifdef CONFIG_SFFS_SHORT_NAMES_IN_RAM
	else if(obj->shortName[0])
	{
		strcpy(name,obj->shortName);
	}
#endif
	else
	{
		__u8 buffer[SFFS_BYTES_PER_CHUNK];
		sffs_ObjectHeader *oh = (sffs_ObjectHeader *)buffer;

		memset(buffer,0,SFFS_BYTES_PER_CHUNK);
	
		if(obj->chunkId >= 0)
		{
			sffs_ReadChunkFromNAND(obj->myDev,obj->chunkId,buffer,NULL,1);
		}
		strncpy(name,oh->name,buffSize - 1);
	}
	
	return strlen(name);
}

int sffs_GetObjectFileLength(sffs_Object *obj)
{
	
	// Dereference any hard linking
	obj = sffs_GetEquivalentObject(obj);
	
	if(obj->variantType == SFFS_OBJECT_TYPE_FILE)
	{
		return obj->variant.fileVariant.fileSize;
	}
	if(obj->variantType == SFFS_OBJECT_TYPE_SYMLINK)
	{
		return strlen(obj->variant.symLinkVariant.alias);
	}
	else
	{
		// Only a directory should drop through to here
		return SFFS_BYTES_PER_CHUNK;
	}	
}

int sffs_GetObjectLinkCount(sffs_Object *obj)
{
	int count = 0; 
	struct list_head *i;
	
	if(!obj->unlinked)
	{
		count++;	// the object itself
	}
	list_for_each(i,&obj->hardLinks)
	{
		count++;	// add the hard links;
	}
	return count;
	
}


int sffs_GetObjectInode(sffs_Object *obj)
{
	obj = sffs_GetEquivalentObject(obj);
	
	return obj->objectId;
}

unsigned sffs_GetObjectType(sffs_Object *obj)
{
	obj = sffs_GetEquivalentObject(obj);
	
	switch(obj->variantType)
	{
		case SFFS_OBJECT_TYPE_FILE:		return DT_REG; break;
		case SFFS_OBJECT_TYPE_DIRECTORY:	return DT_DIR; break;
		case SFFS_OBJECT_TYPE_SYMLINK:		return DT_LNK; break;
		case SFFS_OBJECT_TYPE_HARDLINK:	return DT_REG; break;
		case SFFS_OBJECT_TYPE_SPECIAL:		
			if(S_ISFIFO(obj->st_mode)) return DT_FIFO;
			if(S_ISCHR(obj->st_mode)) return DT_CHR;
			if(S_ISBLK(obj->st_mode)) return DT_BLK;
			if(S_ISSOCK(obj->st_mode)) return DT_SOCK;
		default: return DT_REG; break;
	}
}

char *sffs_GetSymlinkAlias(sffs_Object *obj)
{
	obj = sffs_GetEquivalentObject(obj);
	if(obj->variantType == SFFS_OBJECT_TYPE_SYMLINK)
	{
		return sffs_CloneString(obj->variant.symLinkVariant.alias);
	}
	else
	{
		return sffs_CloneString("");
	}
}

#ifndef CONFIG_SFFS_WINCE

int sffs_SetAttributes(sffs_Object *obj, struct iattr *attr)
{
	unsigned int valid = attr->ia_valid;
	
	if(valid & ATTR_MODE) obj->st_mode = attr->ia_mode;
	if(valid & ATTR_UID) obj->st_uid = attr->ia_uid;
	if(valid & ATTR_GID) obj->st_gid = attr->ia_gid;
	
	if(valid & ATTR_ATIME) obj->st_atime = attr->ia_atime;
	if(valid & ATTR_CTIME) obj->st_ctime = attr->ia_ctime;
	if(valid & ATTR_MTIME) obj->st_mtime = attr->ia_mtime;
	
	if(valid & ATTR_SIZE) sffs_ResizeFile(obj,attr->ia_size);
	
	sffs_UpdateObjectHeader(obj,NULL,1);
	
	return SFFS_OK;
	
}
int sffs_GetAttributes(sffs_Object *obj, struct iattr *attr)
{
	unsigned int valid = 0;
	
	attr->ia_mode = obj->st_mode;	valid |= ATTR_MODE;
	attr->ia_uid = obj->st_uid;		valid |= ATTR_UID;
	attr->ia_gid = obj->st_gid;		valid |= ATTR_GID;
	
	attr->ia_atime = obj->st_atime;	valid |= ATTR_ATIME;
	attr->ia_ctime = obj->st_ctime;	valid |= ATTR_CTIME;
	attr->ia_mtime = obj->st_mtime;	valid |= ATTR_MTIME;
	
	attr->ia_size = sffs_GetFileSize(obj); valid |= ATTR_SIZE;
	
	attr->ia_valid = valid;
	
	return SFFS_OK;
	
}

#endif

int sffs_DumpObject(sffs_Object *obj)
{
//	__u8 buffer[SFFS_BYTES_PER_CHUNK];
	char name[257];
//	sffs_ObjectHeader *oh = (sffs_ObjectHeader *)buffer;

//	memset(buffer,0,SFFS_BYTES_PER_CHUNK);
	
//	if(obj->chunkId >= 0)
//	{
//		sffs_ReadChunkFromNAND(obj->myDev,obj->chunkId,buffer,NULL);
//	}
	
	sffs_GetObjectName(obj,name,256);
	
	T(SFFS_TRACE_ALWAYS,(TSTR("Object %d, inode %d \"%s\"\n dirty %d valid %d serial %d sum %d chunk %d type %d size %d\n" TENDSTR),
			obj->objectId,sffs_GetObjectInode(obj), name, obj->dirty, obj->valid, obj->serial, 
			obj->sum, obj->chunkId, sffs_GetObjectType(obj), sffs_GetObjectFileLength(obj)));

#if 0
	YPRINTF(("Object %d \"%s\"\n dirty %d valid %d serial %d sum %d chunk %d\n",
			obj->objectId, oh->name, obj->dirty, obj->valid, obj->serial, 
			obj->sum, obj->chunkId));
		switch(obj->variantType)
	{
		case SFFS_OBJECT_TYPE_FILE: 
			YPRINTF((" FILE length %d\n",obj->variant.fileVariant.fileSize));
			break;
		case SFFS_OBJECT_TYPE_DIRECTORY:
			YPRINTF((" DIRECTORY\n"));
			break;
		case SFFS_OBJECT_TYPE_HARDLINK: //todo
		case SFFS_OBJECT_TYPE_SYMLINK:
		case SFFS_OBJECT_TYPE_UNKNOWN:
		default:
	}
#endif
	
	return SFFS_OK;
}

int sffs_EraseBlockInNAND(struct sffs_DeviceStruct *dev,int blockInNAND)
{
	dev->nBlockErasures++;
    
    
	return dev->eraseBlockInNAND(dev,blockInNAND - dev->blockOffset);
}

static int sffs_WriteChunkToNAND(struct sffs_DeviceStruct *dev,int chunkInNAND, const __u8 *data, sffs_Spare *spare)
{
	int block_number=0;
	sffs_BlockInfo *bi;
	
	if(chunkInNAND < dev->startBlock* dev->nChunksPerBlock)
	{
		T(SFFS_TRACE_ERROR,(TSTR("**>> sffs chunk %d is not valid" TENDSTR),chunkInNAND));
		return SFFS_FAIL;
	}

	dev->nPageWrites++;

	block_number=chunkInNAND / SFFS_CHUNKS_PER_BLOCK;
	bi=sffs_GetBlockInfo(dev,block_number);
	bi->isUsed= CHUNK_USED;

	return dev->writeChunkToNAND(dev,chunkInNAND - dev->chunkOffset,data,spare);
}



int sffs_ReadChunkFromNAND(struct sffs_DeviceStruct *dev,
		int chunkInNAND, 
		__u8 *data, 
		sffs_Spare *spare, 
		int doErrorCorrection)
{
	int retVal;
	sffs_Spare localSpare;

	dev->nPageReads++;




	if(!spare && data)
	{
		// If we don't have a real spare, then we use a local one.
		// Need this for the calculation of the ecc
		spare = &localSpare;
	}

    
	if(!dev->useNANDECC)
	{
    	retVal  = dev->readChunkFromNAND(dev,chunkInNAND - dev->chunkOffset,data,spare);
		if(data && doErrorCorrection)
		{
			// Do ECC correction
			//Todo handle any errors
			int eccResult1,eccResult2;
			__u8 calcEcc[3];

			sffs_ECCCalculate(data,calcEcc);
			eccResult1 = sffs_ECCCorrect (data,spare->ecc1, calcEcc);
			sffs_ECCCalculate(&data[256],calcEcc);
			eccResult2 = sffs_ECCCorrect(&data[256],spare->ecc2, calcEcc);

			if(eccResult1>0)
			{
				T(SFFS_TRACE_ERROR, (TSTR("**>>ecc error fix performed on chunk %d:0" TENDSTR),chunkInNAND - dev->chunkOffset));
				dev->eccFixed++;
			}
			else if(eccResult1<0)
			{
				T(SFFS_TRACE_ERROR,(TSTR("**>>ecc error unfixed on chunk %d:0" TENDSTR),chunkInNAND - dev->chunkOffset));
				dev->eccUnfixed++;
			}

			if(eccResult2>0)
			{
				T(SFFS_TRACE_ERROR,(TSTR("**>>ecc error fix performed on chunk %d:1" TENDSTR),chunkInNAND - dev->chunkOffset));
				dev->eccFixed++;
			}
			else if(eccResult2<0)
			{
				T(SFFS_TRACE_ERROR,(TSTR("**>>ecc error unfixed on chunk %d:1" TENDSTR),chunkInNAND - dev->chunkOffset));
				dev->eccUnfixed++;
			}

			if(eccResult1 || eccResult2)
			{
				// Hoosterman, we had a data problem on this page
				sffs_HandleReadDataError(dev,chunkInNAND);
			}
		}
	}
	else
	{
		// Must allocate enough memory for spare+2*sizeof(int) for ecc results from device.
		struct sffs_NANDSpare nspare;

		retVal  = dev->readChunkFromNAND(dev,chunkInNAND - dev->chunkOffset,data,(sffs_Spare*)&nspare);
		memcpy (spare, &nspare, sizeof(sffs_Spare));
		if(data && doErrorCorrection)
		{
			if(nspare.eccres1>0)
			{
				T(SFFS_TRACE_ERROR,(TSTR("**>>ecc error fix performed on chunk %d:0" TENDSTR),chunkInNAND - dev->chunkOffset));
			}
			else if(nspare.eccres1<0)
			{
				T(SFFS_TRACE_ERROR,(TSTR("**>>ecc error unfixed on chunk %d:0" TENDSTR),chunkInNAND - dev->chunkOffset));
			}

			if(nspare.eccres2>0)
			{
				T(SFFS_TRACE_ERROR,(TSTR("**>>ecc error fix performed on chunk %d:1" TENDSTR),chunkInNAND - dev->chunkOffset));
			}
			else if(nspare.eccres2<0)
			{
				T(SFFS_TRACE_ERROR,(TSTR("**>>ecc error unfixed on chunk %d:1" TENDSTR),chunkInNAND - dev->chunkOffset));
			}

			if(nspare.eccres1 || nspare.eccres2)
			{
				// Hoosterman, we had a data problem on this page
				sffs_HandleReadDataError(dev,chunkInNAND);
			}

		}
	}
	return retVal;
}


static int sffs_CheckChunkErased(struct sffs_DeviceStruct *dev,int chunkInNAND)
{

	static int init = 0;
	static __u8 cmpbuf[SFFS_BYTES_PER_CHUNK];
	static __u8 data[SFFS_BYTES_PER_CHUNK];
	// Might as well always allocate the larger size for dev->useNANDECC == true;
	static __u8 spare[sizeof(struct sffs_NANDSpare)];	

	dev->readChunkFromNAND(dev,chunkInNAND - dev->chunkOffset,data,(sffs_Spare *)spare);

	if(!init)
	{
		memset(cmpbuf,0xff,SFFS_BYTES_PER_CHUNK);
		init = 1;
	}

	if(memcmp(cmpbuf,data,SFFS_BYTES_PER_CHUNK)) return  SFFS_FAIL;
	if(memcmp(cmpbuf,spare,16)) return SFFS_FAIL;


	return SFFS_OK;

}



///////////////////////// Initialisation code ///////////////////////////



int sffs_GutsInitialise(sffs_Device *dev)
{
	unsigned x;
	int bits;
	int extraBits;
	int nBlocks;

	if(	dev->nBytesPerChunk != SFFS_BYTES_PER_CHUNK ||
	
		dev->nChunksPerBlock < 2 ||
		dev->nReservedBlocks < 2 ||
		dev->startBlock <= 0 ||
		dev->endBlock <= 0 ||
		dev->endBlock <= (dev->startBlock + dev->nReservedBlocks)
	  )
	{
		//these parameters must be set before stating sffs
		// Other parameters startBlock,
		return SFFS_FAIL;
	}


	
	if(!sffs_CheckStructures())
	{
		T(SFFS_TRACE_ALWAYS,(TSTR("sffs_CheckStructures failed\n" TENDSTR)));
		return SFFS_FAIL;
	}

	if(dev->isMounted)
	{
		T(SFFS_TRACE_ALWAYS,(TSTR("sffs: device already mounted\n" TENDSTR)));
		return SFFS_FAIL;
	}

	dev->isMounted = 1;
	
	if(dev->startBlock <= 0 ||
	   (dev->endBlock - dev->startBlock) < 10)
	{
		T(SFFS_TRACE_ALWAYS,(TSTR("startBlock %d or endBlock %d invalid\n" TENDSTR),
				dev->startBlock, dev->endBlock));
		return SFFS_FAIL;
	}

	dev->blockOffset = 0;
	dev->chunkOffset = 0;
       
	if(dev->startBlock == 0)
	{
		dev->blockOffset++;
		dev->chunkOffset = dev->nChunksPerBlock;		
	}
    
	nBlocks = dev->endBlock - dev->startBlock + 1;
	

		
	// OK now calculate a few things for the device
	// Calculate chunkGroupBits. 
	// We need to find the next power of 2 > than endBlock
	
	x = dev->nChunksPerBlock * (dev->endBlock+1);
	
	for(bits = extraBits = 0; x > 1; bits++)
	{
		if(x & 1) extraBits++;
		x >>= 1;
	}

	if(extraBits > 0) bits++;
	
	
	// Level0 Tnodes are 16 bits, so if the bitwidth of the
	// chunk range we're using is greater than 16 we need 
	// to figure out chunk shift and chunkGroupSize
	if(bits <= 16) 
	{
		dev->chunkGroupBits = 0;
	}
	else
	{
		dev->chunkGroupBits = bits - 16;
	}
	dev->chunkGroupSize = 1 << dev->chunkGroupBits;
	
	
	
	// More device initialisation
	dev->garbageCollections = 0;
	dev->passiveGarbageCollections = 0;
	dev->currentDirtyChecker = 0;
	dev->bufferedBlock = -1;
	dev->doingBufferedBlockRewrite = 0;
	dev->nDeletedFiles = 0;
	dev->nBackgroundDeletions=0;
	dev->nUnlinkedFiles = 0;
	dev->eccFixed=0;
	dev->eccUnfixed=0;
	dev->tagsEccFixed=0;
	dev->tagsEccUnfixed=0;
	
	dev->localBuffer = YMALLOC(dev->nBytesPerChunk);
	

	
	
	sffs_InitialiseBlocks(dev,nBlocks);
	
	sffs_InitialiseTnodes(dev);

	sffs_InitialiseObjects(dev);
	
	if(dev->nShortOpCaches > 0)
	{ 
		int i;
		
		if(dev->nShortOpCaches >  SFFS_MAX_SHORT_OP_CACHES)
		{
			dev->nShortOpCaches = SFFS_MAX_SHORT_OP_CACHES;
		}
		
		dev->srCache = YMALLOC( dev->nShortOpCaches * sizeof(sffs_ChunkCache));
		
		for(i=0; i < dev->nShortOpCaches; i++)
		{
			dev->srCache[i].object = NULL;
			dev->srCache[i].lastUse = 0;
			dev->srCache[i].dirty = 0;
		}
		dev->srLastUse = 0;
	}

	dev->cacheHits = 0;
	
	
	// Initialise the unlinked, root and lost and found directories
	dev->lostNFoundDir = dev->rootDir = dev->unlinkedDir = NULL;
	
	dev->unlinkedDir = sffs_CreateFakeDirectory(dev,SFFS_OBJECTID_UNLINKED, S_IFDIR);

	dev->rootDir = sffs_CreateFakeDirectory(dev,SFFS_OBJECTID_ROOT,SFFS_ROOT_MODE | S_IFDIR);
	dev->lostNFoundDir = sffs_CreateFakeDirectory(dev,SFFS_OBJECTID_LOSTNFOUND,SFFS_LOSTNFOUND_MODE | S_IFDIR);
	sffs_AddObjectToDirectory(dev->rootDir,dev->lostNFoundDir);
	
		
    sffs_InitFileSystem(dev);
    
	
	// Zero out stats
	dev->nPageReads = 0;
    dev->nPageWrites =  0;
	dev->nBlockErasures = 0;
	dev->nGCCopies = 0;
	dev->nRetriedWrites = 0;
	dev->nRetiredBlocks = 0;

	
	return SFFS_OK;
		
}

static int sffs_pow(int num)
{
    int temp,result;
    result=1;
    if(num==4)
        return 1000;
    else if(num==3)
        return 100;
    else if(num==2)
        return 10;
    else if(num==1)
        return 1;
}

static int sffs_IsCrash(sffs_Device *dev)
{
	sffs_Spare spare;
	sffs_Tags tags;
	int blk;
	int chunk;
	int c;
	int deleted;
    int start,end;
    int nTemp;
    sffs_BlockInfo *bi;
    __u8 data[512],temp[10];
    __u8 *str;
    int i,t,cnt;
    sffs_Object*theObject;

    memset(data,0,sizeof(data));
    

    start=dev->m_pBlockManager->bmStartTran* dev->nChunksPerBlock;
    end=(dev->m_pBlockManager->bmEndTran+1)* dev->nChunksPerBlock-1;

	for(blk = dev->m_pBlockManager->bmStartTran; blk <= dev->m_pBlockManager->bmEndTran; blk++)
	{
		deleted = 0;
		bi = sffs_GetBlockInfo(dev,blk);
		
		

        // ûũĵ.
		for(c = 0; c < dev->nChunksPerBlock ; c++)
		{
			// Read the spare area and decide what to do
			chunk = blk * dev->nChunksPerBlock + c;

            // ĵ ?
			sffs_ReadChunkFromNAND(dev,chunk,NULL,&spare,1);

			
			// This block looks ok, now what's in this chunk?
			sffs_GetTagsFromSpare(dev,&spare,&tags);



			if(sffs_CountBits(spare.pageStatus) < 6)
			{
				// A deleted chunk
				deleted++;
				dev->nFreeChunks ++;
                sffs_ReadChunkFromNAND(dev,chunk,&data,NULL,1);
				
			}
            else
            {
                sffs_ReadChunkFromNAND(dev,chunk,&data,NULL,1);
                //ڿ "FILE" ԵǴ.
//                printk("sffs_IsCrash : %s\n",data);
                if(strstr(data,"WRITE")!=NULL)
                {
                    g_nIsCrash=1;
                }
                else if(strstr(data,"CREATE")!=NULL)
                {
                    g_nIsCrash=2;
                    memset(temp,0,sizeof(temp));
                    str=strstr(data,"Q");
                    str++;

                    cnt=i=t=0;
                    while(str[i]!='Q')
                    {
                        temp[i]=str[i];
                        i++;
                    }
                    temp[i]='\0';
                    t=i;   
                    i=0;                    
                    while(temp[i]!='\0')
                    {
                        cnt+=(temp[i]-'0')*sffs_pow(t);
                        i++;                        
                        t--;                        
                    }
                    g_nFileCreate=cnt;
                    sffs_DeleteChunk(dev,cnt,1);
                    sprintf(g_strCreateMount,"%s","test");

                }
                    
                else
                {
                    g_nIsCrash=0;
                }
               	sffs_EraseBlockInNAND(dev, dev->m_pBlockManager->bmStartTran);    
               	sffs_EraseBlockInNAND(dev, dev->m_pBlockManager->bmEndTran);        
                
                return 1;
            }
		}
        
    }
    //  κ ֱ.
   	sffs_EraseBlockInNAND(dev, dev->m_pBlockManager->bmStartTran);    
   	sffs_EraseBlockInNAND(dev, dev->m_pBlockManager->bmEndTran);        
    
    
}



static int sffs_TranactionCreateStart(sffs_Device *dev,int m_nChunk)
{
    sffs_BlockInfo *theBlock ;
    int MappingChunk;
    __u8 data[512];

    memset(data,0,sizeof(data));

    if(g_nTranCnt==32)
        g_nTranCnt=0;
    
	theBlock =	sffs_GetBlockInfo(dev,dev->m_pBlockManager->bmStartTran);

    MappingChunk=dev->m_pBlockManager->bmStartTran* dev->nChunksPerBlock;
    g_nTransBlock=0;

    if(theBlock->pagesInUse>=30)
    {
        sffs_EraseBlockInNAND(dev, dev->m_pBlockManager->bmStartTran);        
        theBlock =	sffs_GetBlockInfo(dev,dev->m_pBlockManager->bmEndTran);
        MappingChunk=dev->m_pBlockManager->bmEndTran* dev->nChunksPerBlock;
        g_nTransBlock=1;
    }
    
    MappingChunk+=g_nTranCnt;
    sprintf(data,"%s %s%d%s","FILE CREATE","objectidQ",m_nChunk,"QTHIS");
    
   	sffs_WriteChunkToNAND(dev, MappingChunk, &data, NULL);

    memset(data,0,sizeof(data));
    sffs_ReadChunkFromNAND(dev, MappingChunk, &data, NULL, 1);
    
}


static int sffs_TransactionStart(sffs_Device *dev,int m_nChunk)
{
    //Ķ : file name
    //  ã üũ(켱)  ؼ   ã    write
    // writeϱ ҽ chunk a.txt  write
    // writeϸ  chunk .


    sffs_BlockInfo *theBlock ;
    int MappingChunk;
    __u8 data[512];

    memset(data,0,sizeof(data));
    if(g_nTranCnt==32)
        g_nTranCnt=0;
    
	theBlock =	sffs_GetBlockInfo(dev,dev->m_pBlockManager->bmStartTran);

    MappingChunk=dev->m_pBlockManager->bmStartTran* dev->nChunksPerBlock;
    g_nTransBlock=0;
    if(theBlock->pagesInUse>=30)
    {
        sffs_EraseBlockInNAND(dev, dev->m_pBlockManager->bmStartTran);        
        theBlock =	sffs_GetBlockInfo(dev,dev->m_pBlockManager->bmEndTran);
        MappingChunk=dev->m_pBlockManager->bmEndTran* dev->nChunksPerBlock;
        g_nTransBlock=1;
    }
    
    MappingChunk+=g_nTranCnt;
	sprintf(data,"%s %s%d%s","FILE WRITE","objectid:",m_nChunk,":THIS");
   	sffs_WriteChunkToNAND(dev, MappingChunk, &data, NULL);

    memset(data,0,sizeof(data));
    sffs_ReadChunkFromNAND(dev, MappingChunk, &data, NULL, 1);
   
}

static int sffs_TransactionEnd(sffs_Device *dev)
{
    int MappingChunk;
    __u8 data[512];
    memset(data,0,sizeof(data));
    

    if(g_nTransBlock==0)    //start
   	{
        MappingChunk=dev->m_pBlockManager->bmStartTran* dev->nChunksPerBlock+g_nTranCnt;
    }
    else 
    {
        MappingChunk=dev->m_pBlockManager->bmEndTran* dev->nChunksPerBlock+g_nTranCnt;        
    }
    sffs_ReadChunkFromNAND(dev, MappingChunk, &data, NULL, 1);

    sffs_DeleteChunk(dev,MappingChunk,1);

    g_nTranCnt++;
    
}


static int sffs_InitMountReigon(sffs_Device *dev)
{
	sffs_Spare spare;
 	__u8 *bytes= (__u8 *)(dev->m_pBlockManager);


	sffs_BlockManager *ismount=YMALLOC(512);
	__u8 *ismount_byte=(__u8 *)(ismount);

    __u32 metaCnt=0;
    __u32 currCnt=0;

    __u8 *data=YMALLOC(512);
    __u8 *temp=YMALLOC(SFFS_BYTES_PER_CHUNK);
    int MappingChunk;
    int cnt,rvalue,nBlocks;
    

 	dev->m_pBlockManager = YMALLOC(SFFS_BYTES_PER_CHUNK);
	memset(dev->m_pBlockManager, 0, SFFS_BYTES_PER_CHUNK);
        
/////////////////////////////////////////////////////////////////////////////////////
    dev->m_pBlockManager->bmStartMountInfo=dev->startBlock+1;
    currCnt=dev->m_pBlockManager->bmEndMountInfo=dev->m_pBlockManager->bmStartMountInfo+1;


    nBlocks=dev->endBlock-dev->startBlock+1;

    dev->m_pBlockManager->bmStartMetaInfo=dev->m_pBlockManager->bmEndMountInfo+1;
	dev->m_pBlockManager->bmEndMetaInfo=((nBlocks * SFFS_CHUNKS_PER_BLOCK * sizeof(sffs_Spare)) / SFFS_BYTES_PER_BLOCK)+dev->m_pBlockManager->bmStartMetaInfo;

    dev->m_pBlockManager->bmStartLogInfo=dev->m_pBlockManager->bmEndMetaInfo+1;
    dev->m_pBlockManager->bmEndLogInfo=dev->m_pBlockManager->bmStartLogInfo+1;

    dev->m_pBlockManager->bmStartTran=dev->m_pBlockManager->bmEndLogInfo+1;
    dev->m_pBlockManager->bmEndTran=dev->m_pBlockManager->bmStartTran+1;

    dev->m_pBlockManager->bmStartWear=dev->m_pBlockManager->bmEndTran+1;
    dev->m_pBlockManager->bmEndWear=dev->m_pBlockManager->bmStartWear+1;

    dev->m_pBlockManager->bmStartData=dev->m_pBlockManager->bmEndWear+1;
    dev->m_pBlockManager->bmEndData=dev->endBlock;

/////////////////////////////////////////////////////////////////////////////////////

	MappingChunk=dev->m_pBlockManager->bmEndMountInfo* dev->nChunksPerBlock;


    sffs_ReadChunkFromNAND(dev, MappingChunk, temp, NULL, 1);
    

    
    // unmount Ǵ.
    if(strstr(temp,"unmountd")!=NULL)
    {
        dev->m_pBlockManager->bmMount= MOUNT_SUCCESS;
    }
    //crash 
    else if(strstr(temp,"mount")!=NULL || strstr(temp,"mkunt")!=NULL)
    {
        dev->m_pBlockManager->bmMount= MOUNT_FAIL;
    }
    //ó ҽ.
    else
    {
        dev->m_pBlockManager->bmMount= MOUNT_WELCOME;        
    }

	memset(ismount_byte,0, SFFS_BYTES_PER_CHUNK);
    
   	sffs_EraseBlockInNAND(dev, dev->m_pBlockManager->bmEndMountInfo);    
    memset(temp,0,sizeof(temp));
    sprintf(temp,"%s","mount");
	sffs_WriteChunkToNAND(dev, MappingChunk, temp, NULL);


	return dev->m_pBlockManager->bmMount;
}

static int sffs_WriteWearLeveling(sffs_Device *dev)
{
	int blk,blockNum;
	int chunk;
	int c;
	int RetVal;
	sffs_BlockInfo *bi;
    int nBlock,i,j,cnt;
    __u8 data[512];

   	nBlock = dev->endBlock- dev->startBlock+ 1;
    j=i=cnt=0;
    memset(data,0,sizeof(data));

    blk=dev->m_pBlockManager->bmStartWear;

	for(blockNum = dev->startBlock; blockNum  <= dev->endBlock;blockNum++)
	{
        bi = sffs_GetBlockInfo(dev, blockNum);

        data[i]=bi->nErased;
        i++;
        if(blockNum==(dev->endBlock))
        {
            data[i]='*';
        }

        if(blockNum%512==0 ||blockNum==(dev->endBlock) )
        {
            i=0;
            chunk = blk * dev->nChunksPerBlock + j;
            j++;
          	sffs_WriteChunkToNAND(dev, chunk, &data, NULL);
            memset(data,0,sizeof(data));            

        }     
        
	}	

}


static int sffs_ReadWearLeveling(sffs_Device *dev)
{

	__u8 *bytes;
	int blk,blockNum;
	int chunk;
	int c;
	int RetVal;
	sffs_BlockInfo *bi;
    int nBlock,i,j,cnt;
    __u8 data[512];

   	nBlock = dev->endBlock- dev->startBlock+ 1;

    blk=dev->m_pBlockManager->bmStartWear;

    cnt=dev->startBlock;
    
    for(j=0;j<5;j++)
    {
        chunk = blk * dev->nChunksPerBlock + j;
        
        memset(data,0,sizeof(data));
        RetVal = sffs_ReadChunkFromNAND(dev,chunk,&data,NULL,1);

        for(i=0;i<512;i++)
        {
            if(data[i]=='*')
            {

                break;
            }
     		bi = sffs_GetBlockInfo(dev,cnt);
            bi->nErased=data[i];
            cnt++;
            if(cnt>=dev->endBlock)
                break;
        }
    }

   	sffs_EraseBlockInNAND(dev, dev->m_pBlockManager->bmStartWear);            
	
	
}

static int sffs_ReadLoginfo(sffs_Device *dev)
{
}

void rwTest(sffs_Device *dev)
{
    __u8 *data=YMALLOC(512);
    memset(data,0,sizeof(data));
    
	sffs_EraseBlockInNAND(dev,3);    
	sffs_EraseBlockInNAND(dev, 2);
    sprintf(data,"%s","test");
	sffs_WriteChunkToNAND(dev, 80, data, NULL);    
    
    memset(data,0,sizeof(data));    
	sffs_ReadChunkFromNAND(dev, 80, data, NULL, 1);   
    
    memset(data,0,sizeof(data));        
   	sffs_EraseBlockInNAND(dev,3);    
	sffs_EraseBlockInNAND(dev, 2);
    sprintf(data,"%s","media234");
	sffs_WriteChunkToNAND(dev, 80, data, NULL);    
        
    memset(data,0,sizeof(data));    
	sffs_ReadChunkFromNAND(dev, 80, data, NULL, 1);   
    

}


static int sffs_InitFileSystem(sffs_Device *dev)
{
    int rvalue,MappingChunk;
    __u8 temp[512];

    
    rvalue=sffs_InitMountReigon(dev);
/////////////////////////////////////////////////////////////////////////////////
    struct timeval s_timeval, e_timeval;
    unsigned long ltemp;


//    do_gettimeofday(&s_timeval);    
//////////////////////////////////////////////////////////////////////////////////////

    if(rvalue==MOUNT_SUCCESS)
    {

        g_nMount=M_SUCCESS;

        sffs_Scan(dev);

//////////////////////////////////////////////////////////////////////////////////////
  //      do_gettimeofday(&e_timeval);        
    //    ltemp = (e_timeval.tv_sec * 1000000 + e_timeval.tv_usec) - (s_timeval.tv_sec * 1000000 + s_timeval.tv_usec);

//////////////////////////////////////////////////////////////////////////////////////

        sffs_ReadWearLeveling(dev);
        
    }
    else if(rvalue==MOUNT_FAIL)
    {
       g_nMount=M_FAIL;
       sffs_IsCrash(dev);
        sffs_Scan(dev);
      
       
    }
    else
    {
        g_nMount=M_WELCOME;
        sffs_Scan(dev);
//        do_gettimeofday(&e_timeval);        
//        ltemp = (e_timeval.tv_sec * 1000000 + e_timeval.tv_usec) - (s_timeval.tv_sec * 1000000 + s_timeval.tv_usec);
    }

}

int sffs_WriteMountRegion(sffs_Device *dev)
{
	int val;

	sffs_Spare spare;
    __u8 *data=YMALLOC(512);
    int MappingChunk;

   	memset(data, 0, 512);
    sprintf(data,"%s","unmountd");
    

	MappingChunk=dev->m_pBlockManager->bmEndMountInfo* dev->nChunksPerBlock;
    val=sffs_EraseBlockInNAND(dev, dev->m_pBlockManager->bmEndMountInfo);

    
    val=sffs_WriteChunkToNAND(dev, MappingChunk, data, NULL);

	memset(data, 0, 512);
    val=sffs_ReadChunkFromNAND(dev, MappingChunk, data, NULL,1);
    
}



int sffs_WriteMetaRegion(sffs_Device *dev)
{
	__u8 *spare_data,data[SFFS_BYTES_PER_CHUNK];
	int blk;
	int chunk;
	int MappingChunk;
	int c,i;
	int RetVal;

	sffs_BlockInfo *bi;
	sffs_Spare spare;
	sffs_Tags Tags;
	sffs_Spare spare1[32];

	spare_data = (__u8 *)data;


	for(blk=dev->m_pBlockManager->bmStartData;blk<=dev->m_pBlockManager->bmEndData;blk++)
	{
		bi=sffs_GetBlockInfo(dev, blk);
		if(bi->isUsed != CHUNK_USED)
		{
			continue;
		}
		
				
		for(c=0;c<dev->nChunksPerBlock;c++)
		{
			chunk = blk * dev->nChunksPerBlock + c;

			sffs_ReadChunkFromNAND(dev,chunk,NULL,&spare,1);
			sffs_GetTagsFromSpare(dev,&spare,&Tags);
			memcpy(spare_data+(c*SFFS_BYTES_PER_SPARE), &spare, sizeof(spare));
		}
				
		MappingChunk=(dev->m_pBlockManager->bmStartMetaInfo* dev->nChunksPerBlock)+blk-1;

		sffs_WriteChunkToNAND(dev,MappingChunk, spare_data, NULL);

		memset(data,0,SFFS_BYTES_PER_CHUNK);

		sffs_ReadChunkFromNAND(dev,MappingChunk,data,NULL, 1);
		memcpy(spare1,data,SFFS_BYTES_PER_CHUNK);


		for(i=0;i<32;i++)
		{
			sffs_GetTagsFromSpare(dev, spare1+i, &Tags);
		}

	}
}



static int sffs_WriteLogRegion(sffs_Device *dev)
{

    
}

static int sffs_unMountScanFree(sffs_Device *dev)
{
}



static int sffs_UnMount(sffs_Device *dev)
{	
    

    sffs_WriteMountRegion(dev);   
    sffs_WriteWearLeveling(dev);
   
}




void sffs_Deinitialise(sffs_Device *dev)
{
    int i;
    
    sffs_UnMount(dev);

	if(dev->isMounted)
	{
		sffs_DeinitialiseBlocks(dev);
		sffs_DeinitialiseTnodes(dev);
		sffs_DeinitialiseObjects(dev);
		YFREE(dev->localBuffer);
	}
	
}

#if 0

int  sffs_GetNumberOfFreeChunks(sffs_Device *dev)
{
	int nFree = dev->nFreeChunks - (dev->nChunksPerBlock * SFFS_RESERVED_BLOCKS);
	
	struct list_head *i;	
	sffs_Object *l;
	
	
	// To the free chunks add the chunks that are in the deleted unlinked files.
	list_for_each(i,&dev->unlinkedDir->variant.directoryVariant.children)
	{
		l = list_entry(i, sffs_Object,siblings);
		if(l->deleted)
		{
			nFree++;
			nFree += l->nDataChunks;
		}
	}
	
	
	if(nFree < 0) nFree = 0;

	return nFree;	
	
}

#endif

int  sffs_GetNumberOfFreeChunks(sffs_Device *dev)
{
	int nFree;
	int pending;
	int b;
	int nDirtyCacheChunks=0;
	
	sffs_BlockInfo *blk;
	
	struct list_head *i;	
	sffs_Object *l;
	
	for(nFree = 0, b = dev->startBlock; b <= dev->endBlock; b++)
	{
		blk = sffs_GetBlockInfo(dev,b);
		
		switch(blk->blockState)
		{
			case SFFS_BLOCK_STATE_EMPTY:
			case SFFS_BLOCK_STATE_ALLOCATING: 
			case SFFS_BLOCK_STATE_FULL: nFree += (dev->nChunksPerBlock - blk->pagesInUse); break;
			default: break;
		}
	}
	
	
	pending = 0;
	
	// To the free chunks add the chunks that are in the deleted unlinked files.
	list_for_each(i,&dev->unlinkedDir->variant.directoryVariant.children)
	{
		if(i)
		{
			l = list_entry(i, sffs_Object,siblings);
			if(l->deleted)
			{
				pending++;
				pending += l->nDataChunks;
			}
		}
	}
	
	
	
	if(nFree != dev->nFreeChunks) 
	{
	}

	nFree += pending;
	
	// Now count the number of dirty chunks in the cache and subtract those
	
	{
		int i;
		for(i = 0; i < dev->nShortOpCaches; i++)
		{
			if(dev->srCache[i].dirty) nDirtyCacheChunks++;
		}
	}
	
	nFree -= nDirtyCacheChunks;
	
	nFree -= ((dev->nReservedBlocks + 1) * dev->nChunksPerBlock);
	
	if(nFree < 0) nFree = 0;

	return nFree;	
	
}



/////////////////// SFFS test code //////////////////////////////////

#define sffs_CheckStruct(structure,syze, name) \
           if(sizeof(structure) != syze) \
	       { \
	         T(SFFS_TRACE_ALWAYS,(TSTR("%s should be %d but is %d\n" TENDSTR),name,syze,sizeof(structure))); \
	         return SFFS_FAIL; \
		   }
		 
		 
static int sffs_CheckStructures(void)
{
	sffs_CheckStruct(sffs_Tags,8,"sffs_Tags")
	sffs_CheckStruct(sffs_TagsUnion,8,"sffs_TagsUnion")
	sffs_CheckStruct(sffs_Spare,16,"sffs_Spare")
#ifndef CONFIG_SFFS_TNODE_LIST_DEBUG
	sffs_CheckStruct(sffs_Tnode,2* SFFS_NTNODES_LEVEL0,"sffs_Tnode")
#endif
	sffs_CheckStruct(sffs_ObjectHeader,512,"sffs_ObjectHeader")
	
	
	return SFFS_OK;
}

#if 0
void sffs_GutsTest(sffs_Device *dev)
{
	
	if(sffs_CheckStructures() != SFFS_OK)
	{
		T(SFFS_TRACE_ALWAYS,(TSTR("One or more structures malformed-- aborting\n" TENDSTR)));
		return;
	}
	
	sffs_TnodeTest(dev);
	sffs_ObjectTest(dev);	
}
#endif

void debug_obect(sffs_Object *obj)
{
        //printk(KERN_INFO"-------------sffs object structure--------------\n");
        //printk(KERN_INFO"serial : %d\n",obj->serial);        
        //printk(KERN_INFO"sum : %d\n",obj->sum);                
        //printk(KERN_INFO"chunkId : %d\n",obj->chunkId);        
        //printk(KERN_INFO"objectId : %d\n",obj->objectId);        
        //printk(KERN_INFO"st_uid : %d\n",obj->st_uid);        
        //printk(KERN_INFO"st_gid : %d\n",obj->st_gid);                
        //printk(KERN_INFO"st_atime : %d\n",obj->st_atime);                

        //printk(KERN_INFO"-------------sffs device structure--------------\n");
        //printk(KERN_INFO"startBlock : %d\n",obj->myDev->startBlock);        
        //printk(KERN_INFO"endBlock : %d\n",obj->myDev->endBlock);        
        //printk(KERN_INFO"nTnodesCreated : %d\n",obj->myDev->nTnodesCreated);        
        //printk(KERN_INFO"nFreeTnodes : %d\n",obj->myDev->nFreeTnodes);        
        //printk(KERN_INFO"nFreeChunks : %d\n",obj->myDev->nFreeChunks);        
        //printk(KERN_INFO"currentDirtyChecker : %d\n",obj->myDev->currentDirtyChecker);        
        //printk(KERN_INFO"nPageWrites : %d\n",obj->myDev->nPageWrites);        
        //printk(KERN_INFO"nPageReads : %d\n",obj->myDev->nPageReads);        
        //printk(KERN_INFO"nBlockErasures : %d\n",obj->myDev->nBlockErasures);        
        //printk(KERN_INFO"nGCCopies : %d\n",obj->myDev->nGCCopies);        
        
        //printk(KERN_INFO"++++++++++++++++sffs object structure+++++++++++++\n");        



    
}

