/*
 * SFFS: Yet another FFS. A NAND-flash specific file system. 
 * sffs_guts.h: Configuration etc for sffs_guts
 *
 * Copyright (C) 2002 Aleph One Ltd.
 *   for Toby Churchill Ltd and Brightstar Engineering
 *
 * Created by Charles Manning <charles@aleph1.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 *
 * Note: Only SFFS headers are LGPL, SFFS C code is covered by GPL.
 *
 * $Id: sffs_guts.h,v 1.16 2003/01/31 03:30:33 charles Exp $
 */

#ifndef __SFFS_GUTS_H__
#define __SFFS_GUTS_H__

#include "devextras.h"

//////////////////////////////////SFFS////////////////////////////////////////////////
struct timeval g_WriteSartTime, g_WriteEndTime;
struct timeval g_ReadStartTime, g_ReadEndTime;
unsigned long g_ltemp;
int g_nStartRead,g_nEndRead;
int g_nTranFlag;
int g_nCreateflag;
int g_nCacheCnt;
int g_nMount;
int g_nTranCnt;
int g_nTransBlock;  //0 == start, 1==ends
int g_nCrashStart; //0   unmount, 1  ũ.
int g_nCrashEnd;    // .
int g_nIsCrash; //0̸   writeȰŰ 1̸   .
                // 2̸  ϴ crash߻.
__u32  g_nFileCreate;  
__u8 g_strCreate[32];
__u8 g_strCreateMount[32];

#define M_SUCCESS 1
#define M_FAIL 2
#define M_WELCOME 3

#define MOUNT_STATE_UNMOUNT 1
#define MOUNT_STATE_CRASH 2

#define MOUNT_SUCCESS 1
#define MOUNT_FAIL 0
#define MOUNT_WELCOME 2

#define CHUNK_USED 1
#define CHUNK_UNUSED 0

#define BLOCK_OFFSET 1

typedef struct
{
    __u32 bmMount;
    __u32 bmStartMountInfo;
    __u32 bmEndMountInfo;
    __u32 bmStartMetaInfo;
    __u32 bmEndMetaInfo;
    __u32 bmStartLogInfo;
    __u32 bmEndLogInfo;
    __u32 bmStartData;
    __u32 bmEndData;
    __u32 bmStartTran;
    __u32 bmEndTran;
    __u32 bmStartWear;
    __u32 bmEndWear;
    
}sffs_BlockManager;




//////////////////////////////////////////////////////////////////////////////////////


#define SFFS_OK	1
#define SFFS_FAIL  0

// Give us a Y=0x59, 
// Give us an A=0x41, 
// Give us an FF=0xFF 
// Give us an S=0x53
// And what have we got... 
#define SFFS_MAGIC					0x5941FF53

#define SFFS_NTNODES_LEVEL0	  	16
#define SFFS_TNODES_LEVEL0_BITS	4
#define SFFS_TNODES_LEVEL0_MASK	0xf

#define SFFS_NTNODES_INTERNAL 		(SFFS_NTNODES_LEVEL0 / 2)
#define SFFS_TNODES_INTERNAL_BITS 	(SFFS_TNODES_LEVEL0_BITS - 1)
#define SFFS_TNODES_INTERNAL_MASK	0x7
#define SFFS_TNODES_MAX_LEVEL		6
		
#define SFFS_BYTES_PER_SPARE		16

#define SFFS_BYTES_PER_CHUNK		512
//#define SFFS_CHUNK_SIZE_SHIFT		9


#define SFFS_CHUNKS_PER_BLOCK		32

//32*512
#define SFFS_BYTES_PER_BLOCK		(SFFS_CHUNKS_PER_BLOCK*SFFS_BYTES_PER_CHUNK)

#define SFFS_MAX_CHUNK_ID			0x000FFFFF

#define SFFS_UNUSED_OBJECT_ID		0x0003FFFF

#define SFFS_ALLOCATION_NOBJECTS	100
#define SFFS_ALLOCATION_NTNODES	100
#define SFFS_ALLOCATION_NLINKS		100

#define SFFS_NOBJECT_BUCKETS		256


#define SFFS_OBJECT_SPACE			0x40000
#define SFFS_MAX_NAME_LENGTH		255
#define SFFS_SHORT_NAME_LENGTH		15

#define SFFS_MAX_ALIAS_LENGTH		159

#define SFFS_OBJECTID_ROOT			1
#define SFFS_OBJECTID_LOSTNFOUND	2
#define SFFS_OBJECTID_UNLINKED		3

#define SFFS_MAX_SHORT_OP_CACHES	20


// ChunkCache is used for short read/write operations.
typedef struct
{
	struct sffs_ObjectStruct *object;
	int chunkId;
	int lastUse;
	int dirty;	
	int nBytes;	// Only valid if the cache is dirty
	__u8 data[SFFS_BYTES_PER_CHUNK];
    //js
    __u16 previd;
    __u16 idInChunk;
} sffs_ChunkCache;

//js
sffs_ChunkCache g_cache;



// Tags structures in RAM
// NB This uses bitfield. Bitfields should not straddle a u32 boundary otherwise
// the structure size will get blown out.

typedef struct
{
    unsigned chunkId:20;
    unsigned serialNumber:2;
    unsigned byteCount:10;
    unsigned objectId:18;
    unsigned ecc:12;
    unsigned newType:2; //old : 11 / new : 00
} sffs_Tags;

typedef union
{
    sffs_Tags asTags;
    __u8       asBytes[8];
} sffs_TagsUnion;


// Spare structure
typedef struct
{
    __u8  tagByte0;
    __u8  tagByte1;
    __u8  tagByte2;
    __u8  tagByte3;
    __u8  pageStatus; 	// set to 0 to delete the chunk
    __u8  blockStatus;
    __u8  tagByte4;
    __u8  tagByte5;
    __u8  ecc1[3];
    __u8  tagByte6;
    __u8  tagByte7;
    __u8  ecc2[3];
} sffs_Spare;

//Special structure for passing through to mtd
struct sffs_NANDSpare {
	sffs_Spare	spare;
	int		eccres1;
	int		eccres2;
};

// Block data in RAM

typedef enum {
	SFFS_BLOCK_STATE_UNKNOWN	= 0,
	SFFS_BLOCK_STATE_SCANNING,		// Used while the block is being scanned.
									// NB Don't erase blocks while they're being scanned
	
	SFFS_BLOCK_STATE_EMPTY,		// This block is empty
	
	SFFS_BLOCK_STATE_ALLOCATING,	// This block is partially allocated. 
									// This is the one currently being used for page
									// allocation. Should never be more than one of these
							

	SFFS_BLOCK_STATE_FULL,			// All the pages in this block have been allocated.
									// At least one page holds valid data.
							 
	SFFS_BLOCK_STATE_DIRTY,		// All pages have been allocated and deleted. 
									// Erase me, reuse me.
							
	SFFS_BLOCK_STATE_DEAD 			// This block has failed and is not in use

} sffs_BlockState;




typedef struct
{
#ifndef CONFIG_SFFS_NO_SFFS2
	__u32 sequenceNumber;	// block sequence number for sffs2
#endif
	int   softDeletions:8;  // number of soft deleted pages
    int   pagesInUse:8;		// number of pages in use
    __u32 blockState:4; 	// One of the above block states
    __u32 needsRetiring:1;	// Data has failed on this block, need to get valid data off
    						// and retire the block.
    int isUsed;  
    int nErased;
} sffs_BlockInfo;


//////////////////// Object structure ///////////////////////////
// This is the object structure as stored on NAND

typedef enum
{
	SFFS_OBJECT_TYPE_UNKNOWN,
	SFFS_OBJECT_TYPE_FILE,
	SFFS_OBJECT_TYPE_SYMLINK,
	SFFS_OBJECT_TYPE_DIRECTORY,
	SFFS_OBJECT_TYPE_HARDLINK,
	SFFS_OBJECT_TYPE_SPECIAL
} sffs_ObjectType;

typedef struct
{
	sffs_ObjectType type;

	// Apply to everything	
	int   parentObjectId;
	__u16 sum__NoLongerUsed;	// checksum of name. Calc this off the name to prevent inconsistencies
	char  name[SFFS_MAX_NAME_LENGTH + 1];

	// Thes following apply to directories, files, symlinks - not hard links
	__u32 st_mode;  // protection

#ifdef CONFIG_SFFS_WINCE
	__u32 notForWinCE[5];
#else
	__u32 st_uid;   // user ID of owner
	__u32 st_gid;    // group ID of owner 
	__u32 st_atime; // time of last access
	__u32 st_mtime; // time of last modification
	__u32 st_ctime; // time of last change
#endif

	// File size  applies to files only
	int fileSize; 
		
	// Equivalent object id applies to hard links only.
	int  equivalentObjectId;
	
	// Alias is for symlinks only.
	char alias[SFFS_MAX_ALIAS_LENGTH + 1];
	
	__u32 st_rdev;  // device stuff for block and char devices (maj/min)
	
#ifdef CONFIG_SFFS_WINCE
	__u32 win_ctime[2];
	__u32 win_atime[2];
	__u32 win_mtime[2];
	__u32 roomToGrow[6];
#else
	__u32 roomToGrow[12];
#endif
	
} sffs_ObjectHeader;



////////////////////  Tnode ///////////////////////////

union sffs_Tnode_union
{
#ifdef CONFIG_SFFS_TNODE_LIST_DEBUG
	union sffs_Tnode_union *internal[SFFS_NTNODES_INTERNAL+1];
#else
	union sffs_Tnode_union *internal[SFFS_NTNODES_INTERNAL];
#endif
	__u16 level0[SFFS_NTNODES_LEVEL0];
	
};

typedef union sffs_Tnode_union sffs_Tnode;
//js
sffs_Tnode *g_tNode;

struct sffs_TnodeList_struct
{
	struct sffs_TnodeList_struct *next;
	sffs_Tnode *tnodes;
};

typedef struct sffs_TnodeList_struct sffs_TnodeList;



///////////////////  Object ////////////////////////////////
// An object can be one of:
// - a directory (no data, has children links
// - a regular file (data.... not prunes :->).
// - a symlink [symbolic link] (the alias).
// - a hard link


typedef struct 
{
	__u32 fileSize;
	__u32 scannedFileSize;
	int   topLevel;
	sffs_Tnode *top;
} sffs_FileStructure;

typedef struct
{
	struct list_head children; // list of child links
} sffs_DirectoryStructure;

typedef struct
{
	char *alias;
} sffs_SymLinkStructure;

typedef struct
{
	struct sffs_ObjectStruct *equivalentObject;
	__u32	equivalentObjectId;
} sffs_HardLinkStructure;

typedef union
{
	sffs_FileStructure fileVariant;
	sffs_DirectoryStructure directoryVariant;
	sffs_SymLinkStructure symLinkVariant;
	sffs_HardLinkStructure hardLinkVariant;
} sffs_ObjectVariant;


struct  sffs_ObjectStruct
{
	__u8 deleted: 1;		// This should only apply to unlinked files.
	__u8 softDeleted: 1;	// it has also been soft deleted
	__u8 unlinked: 1;		// An unlinked file. The file should be in the unlinked pseudo directory.
	__u8 fake:1;			// A fake object has no presence on NAND.
	__u8 renameAllowed:1;
	__u8 unlinkAllowed:1;
	__u8 dirty:1;			// the object needs to be written to flash
	__u8 valid:1;			// When the file system is being loaded up, this 
							// object might be created before the data
							// is available (ie. file data records appear before the header).
	__u8 serial;			// serial number of chunk in NAND. Store here so we don't have to
							// read back the old one to update.
	__u16 sum;				// sum of the name to speed searching
	
	struct sffs_DeviceStruct *myDev; // The device I'm on
	
								
	struct list_head hashLink;	// list of objects in this hash bucket
							

	struct list_head hardLinks; // all the equivalent hard linked objects
								// live on this list
	// directory structure stuff
	struct sffs_ObjectStruct  *parent;	//my parent directory
	struct list_head siblings;	// siblings in a directory
								// also used for linking up the free list
		
	// Where's my data in NAND?
	int chunkId;		// where it lives

	int nDataChunks;	
	
	__u32 objectId;		// the object id value
	
	
	__u32 st_mode;  	// protection

#ifdef CONFIG_SFFS_SHORT_NAMES_IN_RAM
	char shortName[SFFS_SHORT_NAME_LENGTH+1];
#endif

#ifndef __KERNEL__
	__u32 inUse;
#endif

#ifdef CONFIG_SFFS_WINCE
	__u32 win_ctime[2];
	__u32 win_mtime[2];
	__u32 win_atime[2];
#else
	__u32 st_uid;   	// user ID of owner
	__u32 st_gid;    	// group ID of owner 
	__u32 st_atime; 	// time of last access
	__u32 st_mtime; 	// time of last modification
	__u32 st_ctime; 	// time of last change
#endif

	__u32 st_rdev; 	    // device stuff for block and char devices



#ifdef __KERNEL__
	struct inode *myInode;
#endif


	
	sffs_ObjectType variantType;
	
	sffs_ObjectVariant variant;
	
};



typedef struct sffs_ObjectStruct sffs_Object;


struct sffs_ObjectList_struct
{
	sffs_Object *objects;
	struct sffs_ObjectList_struct *next;
};

typedef struct sffs_ObjectList_struct sffs_ObjectList;

typedef struct
{
	struct list_head list;
	int count;
} sffs_ObjectBucket;


//////////////////// Device ////////////////////////////////

struct sffs_DeviceStruct
{
    //js
    sffs_BlockManager *m_pBlockManager;
	// Entry parameters set up way early. Yaffs sets up the rest.
	int   nBytesPerChunk; 	 // Should be a power of 2 >= 512
	int	  nChunksPerBlock;	 // does not need to be a power of 2
	int   startBlock;		 // Start block we're allowed to use
	int   endBlock;			 // End block we're allowed to use
	int   nReservedBlocks;	 // We want this tuneable so that we can reduce
							 // reserved blocks on NOR and RAM.
	
	int   useNANDECC;		// Flag to decide whether or not to use NANDECC
	int   nShortOpCaches;	// If <= 0, then short op caching is disabled, else
							// the number of short op caches (don't use too many).
	

	// ̽ ؽƮ Ȯ ǹ?
	void *genericDevice; // Pointer to device context
						 // On an mtd this holds the mtd pointer.

	// NAND access functions (Must be set before calling SFFS)
	//sffs_mtduf .
	int (*writeChunkToNAND)(struct sffs_DeviceStruct *dev,int chunkInNAND, const __u8 *data, sffs_Spare *spare);
	int (*readChunkFromNAND)(struct sffs_DeviceStruct *dev,int chunkInNAND, __u8 *data, sffs_Spare *spare);
	int (*eraseBlockInNAND)(struct sffs_DeviceStruct *dev,int blockInNAND);	
	int (*initialiseNAND)(struct sffs_DeviceStruct *dev);

	// Runtime parameters. Set up by SFFS.
	
	__u16 chunkGroupBits; // 0 for devices <= 32MB. else log2(nchunks) - 16
	__u16 chunkGroupSize; // == 2^^chunkGroupBits
	
#ifdef __KERNEL__

	struct semaphore sem;// Semaphore for waiting on erasure.
	struct semaphore grossLock; // Gross locking semaphore

#endif	
#ifdef __KERNEL__
	void (*putSuperFunc)(struct super_block *sb);
#endif

	int isMounted;
	
	// Block Info
	sffs_BlockInfo *blockInfo;
	__u8 *chunkBits;   // bitmap of chunks in use
	int   chunkBitmapStride; // Number of bytes of chunkBits per block. 
							 //	Must be consistent with nChunksPerBlock.


	int   nErasedBlocks;
	int   allocationBlock;			// Current block being allocated off
	__u32 allocationPage;
	int   allocationBlockFinder;	// Used to search for next allocation block
	
	// Runtime state
	int   nTnodesCreated;	
	sffs_Tnode *freeTnodes;
	int  nFreeTnodes;
	sffs_TnodeList *allocatedTnodeList;


	int   nObjectsCreated;
	sffs_Object *freeObjects;
	int   nFreeObjects;

	sffs_ObjectList *allocatedObjectList;

	sffs_ObjectBucket objectBucket[SFFS_NOBJECT_BUCKETS];

	int	  nFreeChunks;
		
	int   currentDirtyChecker;	// Used to find current dirtiest block
	
	
	// Operations since mount
	int nPageWrites;
	int nPageReads;
	int nBlockErasures;
	int nGCCopies;
	int garbageCollections;
	int passiveGarbageCollections;
	int nRetriedWrites;
	int nRetiredBlocks;
	int eccFixed;
	int eccUnfixed;
	int tagsEccFixed;
	int tagsEccUnfixed;
	int nDeletions;
	int nUnmarkedDeletions;

    //js
	int   blockOffset;
	int   chunkOffset;    
	
	sffs_Object *rootDir;
	sffs_Object *lostNFoundDir;
	
	// Buffer areas for storing data to recover from write failures
//	__u8 		bufferedData[SFFS_CHUNKS_PER_BLOCK][SFFS_BYTES_PER_CHUNK];
//	sffs_Spare bufferedSpare[SFFS_CHUNKS_PER_BLOCK];
	int bufferedBlock;	// Which block is buffered here?
	int doingBufferedBlockRewrite;

	sffs_ChunkCache *srCache;
	int srLastUse;

	int cacheHits;

	// Stuff for background deletion and unlinked files.
	sffs_Object *unlinkedDir;		// Directory where unlinked and deleted files live.
	sffs_Object *unlinkedDeletion;	// Current file being background deleted.
	int nDeletedFiles;				// Count of files awaiting deletion;
	int nUnlinkedFiles;				// Count of unlinked files. 
	int nBackgroundDeletions;			// Count of background deletions.	
	
	__u8 *localBuffer;
	
};

typedef struct sffs_DeviceStruct sffs_Device;



//debug dunc
void debug_obect(sffs_Object *obj);

//////////// SFFS Functions //////////////////

int sffs_GutsInitialise(sffs_Device *dev);
void sffs_Deinitialise(sffs_Device *dev);

int sffs_GetNumberOfFreeChunks(sffs_Device *dev);


// Rename
int sffs_RenameObject(sffs_Object *oldDir, const char *oldName, sffs_Object *newDir, const char *newName);

// generic Object functions
int sffs_Unlink(sffs_Object *dir, const char *name);
int sffs_DeleteFile(sffs_Object *obj);

// Object access functions.
int sffs_GetObjectName(sffs_Object *obj,char *name,int buffSize);
int sffs_GetObjectFileLength(sffs_Object *obj);
int sffs_GetObjectInode(sffs_Object *obj);
unsigned sffs_GetObjectType(sffs_Object *obj);
int sffs_GetObjectLinkCount(sffs_Object *obj);

// Change inode attributes
int sffs_SetAttributes(sffs_Object *obj, struct iattr *attr);
int sffs_GetAttributes(sffs_Object *obj, struct iattr *attr);

// File operations
int sffs_ReadDataFromFile(sffs_Object *obj, __u8 *buffer, __u32 offset, int nBytes);
int sffs_WriteDataToFile(sffs_Object *obj, const __u8 *buffer, __u32 offset, int nBytes);
int sffs_ResizeFile(sffs_Object *obj, int newSize);

sffs_Object *sffs_MknodFile(sffs_Object *parent,const char *name, __u32 mode, __u32 uid, __u32 gid);
int sffs_FlushFile(sffs_Object *obj,int updateTime);


// Directory operations
sffs_Object *sffs_MknodDirectory(sffs_Object *parent,const char *name, __u32 mode, __u32 uid, __u32 gid);
sffs_Object *sffs_FindObjectByName(sffs_Object *theDir,const char *name);
int sffs_ApplyToDirectoryChildren(sffs_Object *theDir,int (*fn)(sffs_Object *));

sffs_Object *sffs_FindObjectByNumber(sffs_Device *dev,__u32 number);

// Link operations
sffs_Object *sffs_Link(sffs_Object *parent, const char *name, sffs_Object *equivalentObject);

sffs_Object *sffs_GetEquivalentObject(sffs_Object *obj);

// Symlink operations
sffs_Object *sffs_MknodSymLink(sffs_Object *parent, const char *name, __u32 mode,  __u32 uid, __u32 gid, const char *alias);
char *sffs_GetSymlinkAlias(sffs_Object *obj);

// Special inodes (fifos, sockets and devices)
sffs_Object *sffs_MknodSpecial(sffs_Object *parent,const char *name, __u32 mode, __u32 uid, __u32 gid,__u32 rdev);


// Special directories
sffs_Object *sffs_Root(sffs_Device *dev);
sffs_Object *sffs_LostNFound(sffs_Device *dev);

#ifdef CONFIG_SFFS_WINCE
// CONFIG_SFFS_WINCE special stuff
void  yfsd_WinFileTimeNow(__u32 target[2]);
#endif


// Debug dump 
int sffs_DumpObject(sffs_Object *obj);


void sffs_GutsTest(sffs_Device *dev);

static  __inline__ sffs_BlockInfo* sffs_GetBlockInfo(sffs_Device *dev, int blk)
{
	if(blk < dev->startBlock || blk > dev->endBlock)
	{
		T(SFFS_TRACE_ERROR,(TSTR("**>> sffs: getBlockInfo block %d is not valid" TENDSTR),blk));
		YBUG();
	}
	return &dev->blockInfo[blk - dev->startBlock];
}


#endif

