

const char *sffs_mtdif_c_version = "$Id: sffs_mtdif.c,SFFS 2007/11/17 $";

#ifdef CONFIG_SFFS_MTD_ENABLED
 
#include "yportenv.h"

#include "sffs_mtdif.h"
#include "linux/mtd/mtd.h"
#include "linux/types.h"
//#include "sffs_guts.h"
#ifndef	CONFIG_SFFS_USE_OLD_MTD
#include "linux/mtd/nand.h"
#endif

struct nand_oobinfo sffs_oobinfo = {
	useecc: 1,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,8))
// this is for versions of mtd nand driver in kernel 2.6.8 and later
	eccbytes: 6,
#endif
	eccpos: {8, 9, 10, 13, 14, 15}
};

struct nand_oobinfo sffs_noeccinfo = {
	useecc: 0,
};
static void sffs_GetTagsFromSpare(sffs_Device *dev, sffs_Spare *sparePtr,sffs_Tags *tagsPtr)
{
	sffs_TagsUnion *tu = (sffs_TagsUnion *)tagsPtr;
	int result;

	tu->asBytes[0]= sparePtr->tagByte0;
	tu->asBytes[1]= sparePtr->tagByte1;
	tu->asBytes[2]= sparePtr->tagByte2;
	tu->asBytes[3]= sparePtr->tagByte3;
	tu->asBytes[4]= sparePtr->tagByte4;
	tu->asBytes[5]= sparePtr->tagByte5;
	tu->asBytes[6]= sparePtr->tagByte6;
	tu->asBytes[7]= sparePtr->tagByte7;

	
}

int nandmtd_WriteChunkToNAND(sffs_Device *dev,int chunkInNAND,const __u8 *data, sffs_Spare *spare)
{

    int i;
	struct mtd_info *mtd = (struct mtd_info *)(dev->genericDevice);
	size_t dummy;
    int retval = 0;
	
	loff_t addr = ((loff_t)chunkInNAND) * dev->nBytesPerChunk;
	
	__u8 *spareAsBytes = (__u8 *)spare;


    printk("\n\nMTD::WRITE\n\n");
      for(i=0;i<100000000;i++);                

#ifndef	CONFIG_SFFS_USE_OLD_MTD
	if(data && spare)
	{
		if(dev->useNANDECC)
			mtd->write_ecc(mtd,addr,dev->nBytesPerChunk,&dummy,data,spareAsBytes,&sffs_oobinfo);
		else
			mtd->write_ecc(mtd,addr,dev->nBytesPerChunk,&dummy,data,spareAsBytes,&sffs_noeccinfo);
	}
	else
	{
#endif
	if(data)
	{
            retval = mtd->write(mtd,addr,dev->nBytesPerChunk,&dummy,data);
	}
    
	if(spare)
	{
		retval = mtd->write_oob(mtd,addr,SFFS_BYTES_PER_SPARE,&dummy,spareAsBytes);
	}
#ifndef	CONFIG_SFFS_USE_OLD_MTD
	}
#endif

    if (retval == 0)
    	return SFFS_OK;
    else
        return SFFS_FAIL;
}

int nandmtd_ReadChunkFromNAND(sffs_Device *dev,int chunkInNAND, __u8 *data, sffs_Spare *spare)
{

	struct mtd_info *mtd = (struct mtd_info *)(dev->genericDevice);
	size_t dummy;
    int retval = 0;
	
	loff_t addr = ((loff_t)chunkInNAND) * dev->nBytesPerChunk;
	
	__u8 *spareAsBytes = (__u8 *)spare;

#ifndef	CONFIG_SFFS_USE_OLD_MTD
	if(data && spare)
	{
		if(dev->useNANDECC)
		{   // Careful, this call adds 2 ints to the end of the spare data.  Calling function should
            // allocate enough memory for spare, i.e. [SFFS_BYTES_PER_SPARE+2*sizeof(int)].
    		retval = mtd->read_ecc(mtd,addr,dev->nBytesPerChunk,&dummy,data,spareAsBytes,&sffs_oobinfo);            
		}
		else
		{
			retval = mtd->read_ecc(mtd,addr,dev->nBytesPerChunk,&dummy,data,spareAsBytes,&sffs_noeccinfo);
		}
	}
	else
	{
#endif
	if(data)
		retval = mtd->read(mtd,addr,dev->nBytesPerChunk,&dummy,data);
	if(spare)
		retval = mtd->read_oob(mtd,addr,SFFS_BYTES_PER_SPARE,&dummy,spareAsBytes);
#ifndef	CONFIG_SFFS_USE_OLD_MTD
	}
#endif

    if (retval == 0)
    	return SFFS_OK;
    else
        return SFFS_FAIL;
}

// Callback not needed for NAND
#if 0
static void nandmtd_EraseCallback(struct erase_info *ei)
{
	sffs_Device *dev = (sffs_Device *)ei->priv;	
	up(&dev->sem);
}
#endif


int nandmtd_EraseBlockInNAND(sffs_Device *dev, int blockNumber)
{
	struct mtd_info *mtd = (struct mtd_info *)(dev->genericDevice);
	__u32 addr = ((loff_t) blockNumber) * dev->nBytesPerChunk * dev->nChunksPerBlock;
	struct erase_info ei;
    int retval = 0;
	int blockNum;
	sffs_BlockInfo * ScanA;
	
	ei.mtd = mtd;
	ei.addr = addr;
	ei.len = dev->nBytesPerChunk * dev->nChunksPerBlock;
	ei.time = 1000;
	ei.retries = 2;
	ei.callback = NULL;
	ei.priv = (u_long)dev;
	ScanA = sffs_GetBlockInfo(dev, blockNumber);

	
	ScanA ->nErased++;
	
	// Todo finish off the ei if required
	
	sema_init(&dev->sem,0);

	retval = mtd->erase(mtd,&ei);	

    //printk("nandmtd_EraseBlockInNAND : %d / blockNumber : %d\n",retval,blockNumber);
	//No need for callback 
	// down(&dev->sem); // Wait for the erasure to complete

    if (retval == 0)	
    	return SFFS_OK;
    else
        return SFFS_FAIL;
}

int nandmtd_InitialiseNAND(sffs_Device *dev)
{
	return SFFS_OK;
}

#endif // CONFIG_SFFS_MTD_ENABLED

 
