
#include "sffscfg.h"
#include "sffsfs.h"
#include <errno.h>

unsigned sffs_traceMask = 0xFFFFFFFF;


void sffsfs_SetError(int err)
{
	//Do whatever to set error
	errno = err;
}

void sffsfs_Lock(void)
{
}

void sffsfs_Unlock(void)
{
}

__u32 sffsfs_CurrentTime(void)
{
	return 0;
}

void sffsfs_LocalInitialisation(void)
{
	// Define locking semaphore.
}

// Configuration for:
// /ram  2MB ramdisk
// /boot 2MB boot disk (flash)
// /flash 14MB flash disk (flash)
// NB Though /boot and /flash occupy the same physical device they
// are still disticnt "sffs_Devices. You may think of these as "partitions"
// using non-overlapping areas in the same device.
// 

#include "sffs_ramdisk.h"
#include "sffs_flashif.h"

static sffs_Device ramDev;
static sffs_Device bootDev;
static sffs_Device flashDev;

static sffsfs_DeviceConfiguration sffsfs_config[] = {

	{ "/ram", &ramDev},
	{ "/boot", &bootDev},
	{ "/flash", &flashDev},
	{(void *)0,(void *)0}
};


int sffs_StartUp(void)
{
	// Stuff to configure SFFS
	// Stuff to initialise anything special (eg lock semaphore).
	sffsfs_LocalInitialisation();
	
	// Set up devices

	// /ram
	ramDev.nBytesPerChunk = SFFS_BYTES_PER_CHUNK;
	ramDev.nChunksPerBlock = SFFS_CHUNKS_PER_BLOCK;
	ramDev.nReservedBlocks = 2; // Set this smaller for RAM
	ramDev.startBlock = 1; // Can't use block 0
	ramDev.endBlock = 127; // Last block in 2MB.	
	ramDev.useNANDECC = 1;
	ramDev.nShortOpCaches = 0;	// Disable caching on this device.
	ramDev.genericDevice = (void *) 0;	// Used to identify the device in fstat.
	ramDev.writeChunkToNAND = yramdisk_WriteChunkToNAND;
	ramDev.readChunkFromNAND = yramdisk_ReadChunkFromNAND;
	ramDev.eraseBlockInNAND = yramdisk_EraseBlockInNAND;
	ramDev.initialiseNAND = yramdisk_InitialiseNAND;

	// /boot
	bootDev.nBytesPerChunk = SFFS_BYTES_PER_CHUNK;
	bootDev.nChunksPerBlock = SFFS_CHUNKS_PER_BLOCK;
	bootDev.nReservedBlocks = 5;
	bootDev.startBlock = 1; // Can't use block 0
	bootDev.endBlock = 127; // Last block in 2MB.	
	bootDev.useNANDECC = 0; // use SFFS's ECC
	bootDev.nShortOpCaches = 10; // Use caches
	bootDev.genericDevice = (void *) 1;	// Used to identify the device in fstat.
	bootDev.writeChunkToNAND = yflash_WriteChunkToNAND;
	bootDev.readChunkFromNAND = yflash_ReadChunkFromNAND;
	bootDev.eraseBlockInNAND = yflash_EraseBlockInNAND;
	bootDev.initialiseNAND = yflash_InitialiseNAND;

		// /flash
	flashDev.nBytesPerChunk = SFFS_BYTES_PER_CHUNK;
	flashDev.nChunksPerBlock = SFFS_CHUNKS_PER_BLOCK;
	flashDev.nReservedBlocks = 5;
	flashDev.startBlock = 128; // First block after 2MB
	flashDev.endBlock = 1023; // Last block in 16MB
	flashDev.useNANDECC = 0; // use SFFS's ECC
	flashDev.nShortOpCaches = 10; // Use caches
	flashDev.genericDevice = (void *) 2;	// Used to identify the device in fstat.
	flashDev.writeChunkToNAND = yflash_WriteChunkToNAND;
	flashDev.readChunkFromNAND = yflash_ReadChunkFromNAND;
	flashDev.eraseBlockInNAND = yflash_EraseBlockInNAND;
	flashDev.initialiseNAND = yflash_InitialiseNAND;

	sffs_initialise(sffsfs_config);
	
	return 0;
}




