/*
 * SFFS: Yet another FFS. A NAND-flash specific file system. 
 * yportenv.h: Portable services used by sffs. This is done to allow
 * simple migration from kernel space into app space for testing.
 *
 * Copyright (C) 2002 Aleph One Ltd.
 *   for Toby Churchill Ltd and Brightstar Engineering
 *
 * Created by Charles Manning <charles@aleph1.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 *
 * Note: Only SFFS headers are LGPL, SFFS C code is covered by GPL.
 *
 * $Id: yportenv.h,v 1.10 2003/01/31 03:30:33 charles Exp $
 *
 */
 
#ifndef __YPORTENV_H__
#define __YPORTENV_H__


#if defined CONFIG_SFFS_WINCE

// CONFIG_SFFS_WINCE
#include "stdlib.h"
#include "stdio.h"
#include "string.h"

#define CONFIG_SFFS_CASE_INSENSITIVE

#define YMALLOC(x) malloc(x)
#define YFREE(x)   free(x)


#define YINFO(s) YPRINTF(( __FILE__ " %d %s\n",__LINE__,s))
#define YALERT(s) YINFO(s)

#include <windows.h>

#define SFFS_LOSTNFOUND_NAME		"Lost Clusters"
#define SFFS_LOSTNFOUND_PREFIX		"OBJ"

#define YPRINTF(x)	RETAILMSG(1,x)

// Always pass the sum compare to overcome the case insensitivity issue
#define sffs_SumCompare(x,y) ((x) == (y))
#define sffs_strcmp(a,b) _stricmp(a,b)


#define u_char unsigned char
#define loff_t int
#define S_IFDIR						04000

#define S_ISFIFO(x) 0
#define S_ISCHR(x) 0
#define S_ISBLK(x) 0
#define S_ISSOCK(x) 0

extern unsigned yfsd_U32FileTimeNow(void);

#define CURRENT_TIME				 yfsd_U32FileTimeNow()
#define SFFS_ROOT_MODE				FILE_ATTRIBUTE_ARCHIVE
#define SFFS_LOSTNFOUND_MODE		FILE_ATTRIBUTE_HIDDEN


#define TENDSTR L"\r\n"
#define TSTR(x) TEXT(x)
#define TOUT(x) RETAILMSG(1, x)


#elif  defined __KERNEL__



// Linux kernel
#include "linux/kernel.h"
#include "linux/mm.h"
#include "linux/string.h"
#include "linux/slab.h"


#define SFFS_LOSTNFOUND_NAME		"lost+found"
#define SFFS_LOSTNFOUND_PREFIX		"obj"

//#define YPRINTF(x) printk x
#define YMALLOC(x) kmalloc(x,GFP_KERNEL)
#define YFREE(x)   kfree(x)

#define SFFS_ROOT_MODE				0666
#define SFFS_LOSTNFOUND_MODE		0666


#define sffs_SumCompare(x,y) ((x) == (y))
#define sffs_strcmp(a,b) strcmp(a,b)

#define TENDSTR "\n"
#define TSTR(x) KERN_DEBUG x
#define TOUT(p) printk p


#elif defined CONFIG_SFFS_DIRECT

// Direct interface

#include "stdlib.h"
#include "stdio.h"
#include "string.h"

#define YMALLOC(x) malloc(x)
#define YFREE(x)   free(x)


//#define YINFO(s) YPRINTF(( __FILE__ " %d %s\n",__LINE__,s))
//#define YALERT(s) YINFO(s)


#define TENDSTR "\n"
#define TSTR(x) x
#define TOUT(p) printf p


#define SFFS_LOSTNFOUND_NAME		"lost+found"
#define SFFS_LOSTNFOUND_PREFIX		"obj"
//#define YPRINTF(x) printf x

#include "sffscfg.h"

#define CURRENT_TIME sffsfs_CurrentTime()
#define SFFS_ROOT_MODE				0666
#define SFFS_LOSTNFOUND_MODE		0666

#define sffs_SumCompare(x,y) ((x) == (y))
#define sffs_strcmp(a,b) strcmp(a,b)

#elif defined CONFIG_SFFS_UTIL

// Stuff for SFFS utilities

#include "stdlib.h"
#include "stdio.h"
#include "string.h"

#define YMALLOC(x) malloc(x)
#define YFREE(x)   free(x)


//#define YINFO(s) YPRINTF(( __FILE__ " %d %s\n",__LINE__,s))
//#define YALERT(s) YINFO(s)


#define TENDSTR "\n"
#define TSTR(x) x
#define TOUT(p) printf p


#define SFFS_LOSTNFOUND_NAME		"lost+found"
#define SFFS_LOSTNFOUND_PREFIX		"obj"
//#define YPRINTF(x) printf x


#define CURRENT_TIME 0
#define SFFS_ROOT_MODE				0666
#define SFFS_LOSTNFOUND_MODE		0666

#define sffs_SumCompare(x,y) ((x) == (y))
#define sffs_strcmp(a,b) strcmp(a,b)

#else
// Should have specified a configuration type
#error Unknown configuration

#endif 


extern unsigned sffs_traceMask;

#define SFFS_TRACE_ERROR		0x0001
#define SFFS_TRACE_OS			0x0002
#define SFFS_TRACE_ALLOCATE	0x0004
#define SFFS_TRACE_SCAN		0x0008
#define SFFS_TRACE_BAD_BLOCKS	0x0010
#define SFFS_TRACE_ERASE		0x0020
#define SFFS_TRACE_GC			0x0040
#define SFFS_TRACE_TRACING		0x0100
#define SFFS_TRACE_ALWAYS		0x0200
#define SFFS_TRACE_BUG			0x8000

//	T(SFFS_TRACE_OS,(KERN_DEBUG"inin test\n"));	
//#define T(mask,p) do{ if((mask) & (sffs_traceMask | SFFS_TRACE_ERROR)) TOUT(p);} while(0) 
#define T(mask,p) do{ if((mask) & (sffs_traceMask | SFFS_TRACE_OS)) TOUT(p);} while(0) 

//#undef YINFO


//#define YINFO(s) YPRINTF((KERN_DEBUG __FILE__ " %d %s\n",__LINE__,s))
//#define YALERT(s) YINFO(s)
#ifdef CONFIG_SFFS_WINCE
#define YBUG() T(SFFS_TRACE_BUG,(TSTR("==>> sffs bug: %s %d" TENDSTR),TEXT(__FILE__),__LINE__))
#else
#define YBUG() T(SFFS_TRACE_BUG,(TSTR("==>> sffs bug: " __FILE__ " %d" TENDSTR),__LINE__))
#endif
#endif
